#pragma once
#include "../../../Utils/FSM/FSMState.h"
#include "Behavior.h"

typedef IFSMState IBehaviorState;

/**
 * @brief A behavior state represents a state in a finite state machine. Each state stores code to execute
 *        when it becomes active, and transitions associated to events to get out of that state.  
 * @tparam DerivedState The most derived base class of this behavior state. Used to allow static polymorphism.
 */
template <class Behavior, class DerivedState>
class BehaviorState : public FSMState<Behavior, DerivedState>
{
public:
  BehaviorState(Behavior& behavior)
    : FSMState(behavior)
  {
  }

  virtual ~BehaviorState()
  {
  }

  /**
   * @return the behavior encapsulating this state, which stores information common to all states.
   * @remarks The actual behavior is stored in derived classes (instead of here as a Behavior*) so that 
   *          they can use it without casting to the derived type.
   */
  Behavior& getBehavior()
  {
    return dynamic_cast<Behavior&>(m_context);
  }
private:
  BehaviorState(const BehaviorState& other);
};
