#include "stdafx.h"
#include "BuildFSM.h"
#include "../Layer1FSM.h"

///////////////////////////////////////////////////
// Goto state

BuildFSM::Goto::Goto(BuildFSM& behavior)
  : BehaviorState(behavior)
  , m_moveAction(*this)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 20, 5)
  , m_transitionToBuildState(
      *this,
      UnitReachedDestination(behavior.m_unit,BWAPI::Position(((m_context.m_buildingTP.x()-1)*TILE_SIZE)-16,((m_context.m_buildingTP.y()-1)*TILE_SIZE)-16),256), 
      NoTransitionAction(), 
      behavior.m_buildState)
  , m_events(m_timerEvent)
  , m_transitions(m_transitionToBuildState)
{
}

void BuildFSM::Goto::onEnterImpl()
{
  BWAPI::Position destination(((m_context.m_buildingTP.x()-1)*TILE_SIZE)-16,((m_context.m_buildingTP.y()-1)*TILE_SIZE)-16);
  destination.makeValid();
  m_moveAction.setDestination(destination);
  Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.m_unit, m_moveAction, Arbitration::ActionImportance(*this));
}

///////////////////////////////////////////////////
// Build state

BuildFSM::Build::Build(BuildFSM& behavior)
  : BehaviorState(behavior)
  , m_buildAction(*this)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToBuildState(
      *this,
      Not<BuildingStarted>(BuildingStarted(behavior.m_buildingTP,behavior.m_buildingType)), 
      NoTransitionAction(), 
      behavior.m_buildState)
  , m_transitionToFinishState(
      *this,
      BuildingStarted(behavior.m_buildingTP,behavior.m_buildingType), 
      NoTransitionAction(), 
      behavior.m_finishState)
  , m_events(m_timerEvent)
  , m_transitions(m_transitionToBuildState,m_transitionToFinishState)
{
}

void BuildFSM::Build::onEnterImpl()
{
  m_buildAction.setDestination(m_context.m_buildingTP);
  m_buildAction.setUnitType(m_context.m_buildingType);
  Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.m_unit, m_buildAction, Arbitration::ActionImportance(*this));
}

///////////////////////////////////////////////////
// Finish state

BuildFSM::Finish::Finish(BuildFSM& behavior)
  : BehaviorState(behavior)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_events(m_timerEvent)
{
}

void BuildFSM::Finish::onEnterImpl()
{

}