#pragma once
#include "../Layer1FSM.h"
#include "../Conditions/ReachedDestination.h"
#include "../Conditions/BuildingStarted.h"
#include "../UnitBehavior.h"
#include "../BehaviorState.h"
#include "../Arbitrator.h"

/**
* @brief Will build, as specified, using an unit, the building at the location.
*/
class BuildFSM : public UnitBehavior<BuildFSM>
{
protected:
  class Goto;
  class Clear;
  class Build;
  class Finish;
  friend class Goto;
  friend class Clear;
  friend class Build;
  friend class Finish;

  class Goto : public BehaviorState<BuildFSM, Goto>
  {
  public:
    Goto(BuildFSM& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_moveAction, Arbitration::ActionImportance(*this));
      m_transitionToBuildState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Goto, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
    const Events m_events;

    typedef boost::tuple<
      Transition<Goto, UnitReachedDestination, NoTransitionAction, Build>&
    > Transitions;
    const Transitions m_transitions;

  protected:
    // Variables

    // Actions
    layer1::MoveAction m_moveAction;

    // Transitions
    TimerEvent m_timerEvent;
    Transition<Goto, UnitReachedDestination, NoTransitionAction, Build> m_transitionToBuildState;
  };

  class Build : public BehaviorState<BuildFSM, Build>
  {
  public:
    Build(BuildFSM& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_transitionToFinishState.testTransition(boost::make_tuple());
      m_transitionToBuildState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Build, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
    const Events m_events;

    typedef boost::tuple<
      Transition<Build, Not<BuildingStarted>, NoTransitionAction, Build>&,
      Transition<Build, BuildingStarted, NoTransitionAction, Finish>&
    > Transitions;
    const Transitions m_transitions;
  protected:
    // Variables

    // Actions
    layer1::BuildAction m_buildAction;

    // Transitions
    TimerEvent m_timerEvent;
    Transition<Build, Not<BuildingStarted>, NoTransitionAction, Build> m_transitionToBuildState;
    Transition<Build, BuildingStarted, NoTransitionAction, Finish> m_transitionToFinishState;
  };

  /*class Clear : public BehaviorState<BuildFSM, Clear>
  {
  public:
  Clear(BuildFSM& behavior);
  void onEnterImpl();
  void onExitImpl() {}

  void onTimer()
  {}
  };*/

  class Finish : public BehaviorState<BuildFSM, Finish>
  {
  public:
    Finish(BuildFSM& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_context.terminate(m_context.Success);
    }

    typedef SCHEDULER_EVENT(Finish, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
    const Events m_events;

  protected:
    // Variables

    // Actions

    // Transitions
    TimerEvent m_timerEvent;
  };

public:
  BuildFSM(const Process* parent, unsigned int initialPriority, BWAPI::Unit* buildingUnit, BWAPI::TilePosition buildingTP, BWAPI::UnitType buildingType)
    : UnitBehavior(parent, initialPriority, buildingUnit),
    IUnitBehavior(parent, initialPriority, buildingUnit),
    IBehavior(parent, initialPriority),
    IFSM(parent),
    Process(parent),
    m_gotoState(*this),    
    m_buildState(*this),
    //m_clearState(*this),
    m_finishState(*this),
    m_buildingTP(buildingTP),
    m_buildingType(buildingType),
    m_states(m_gotoState, m_buildState, m_finishState)
  {
  }

  virtual std::string getName() const
  {
    return "Building " + getObjectDescription(m_buildingType) + " with (" + getObjectDescription(m_unit) + ")";
  }

  typedef boost::tuple<
      Goto&,
      Build&,
      Finish&
    > States;
	const States m_states;

protected:
  virtual void executeImpl()
  {
    UnitBehavior::executeImpl();
    m_gotoState.onEnter();
  }

  // Variables
  BWAPI::TilePosition m_buildingTP;
  BWAPI::UnitType m_buildingType;

  // States
  Goto m_gotoState;  
  Build m_buildState;
  //Clear m_clearState;
  Finish m_finishState;
};