#include "stdafx.h"
#include "DefaultAttack.h"
#include <BWAPI.h>
#include <cassert>

// Engaging state //////////////////////////////////////////////////////////////
DefaultAttackFSM::Engaging::Engaging(DefaultAttackFSM& behavior)
  : BehaviorState(behavior)
  , m_attackAction(*this)
  , m_targetDefeated(*this, NoCondition(), NoTransitionAction(), behavior.m_idlingState)
  //, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_events()
  , m_transitions(m_targetDefeated)
{
}

void DefaultAttackFSM::Engaging::onEnterImpl()
{
  assert(m_context.m_target != NULL);

  m_targetDefeatedId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_context.m_target, &m_targetDefeated, NULL);

  m_attackAction.setUnitToAttack(m_context.m_target);
  
  Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.m_unit, m_attackAction, Arbitration::ActionImportance(*this));
}

void DefaultAttackFSM::Engaging::onExitImpl()
{
  if(m_context.m_target!= NULL)
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_context.m_target, m_targetDefeatedId);
}

//void DefaultAttackFSM::Engaging::onTimer()
//{}

void DefaultAttackFSM::Engaging::targetChanged(BWAPI::Unit* oldTarget)
{
  if(oldTarget != NULL)
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(oldTarget, m_targetDefeatedId);
  if(m_context.m_target != NULL)
    m_targetDefeatedId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_context.m_target, &m_targetDefeated, NULL);
}

// Idling state ////////////////////////////////////////////////////////////////
DefaultAttackFSM::Idling::Idling(DefaultAttackFSM& behavior)
  : BehaviorState(behavior)
  , m_transition(*this, NoCondition(), NoTransitionAction(), behavior.m_engagingState)
  , m_transitions(m_transition)
{
}

void DefaultAttackFSM::Idling::onTargetChanged()
{
  // Transition to Engaging state
  m_transition.testTransition(boost::make_tuple());
}

bool DefaultAttackFSM::selectTarget(const UnitGroup& enemies)
{
    BWAPI::Unit* oldTarget = m_target;
    m_target = selectTargetImpl(enemies);
    if(m_currentState == &m_engagingState)
      m_engagingState.targetChanged(oldTarget);
    if(m_target == NULL)
      return false;
    if (m_currentState == &m_idlingState)
      m_idlingState.onTargetChanged();

    return true;
}

BWAPI::Unit* DefaultAttackFSM::selectTargetImpl(UnitGroup enemies)
{
    int minhp = boost::numeric::bounds<int>::highest();
    int mindist = boost::numeric::bounds<int>::highest();
    bool threatFound = false;
    BWAPI::Unit* target = NULL;

    for(UnitGroup::const_iterator i=enemies.begin(); i!=enemies.end(); ++i)
    {
      BWAPI::Unit* unit = *i;
      //Ignore invisible units, eggs and larvae
      if(!unit->isDetected() 
        || unit->getType().isFlyer()
        || unit->getType() == BWAPI::UnitTypes::Zerg_Larva 
        || unit->getType() == BWAPI::UnitTypes::Zerg_Egg)
        continue;
      int hp = unit->getHitPoints() + unit->getShields();
      int dist = unit->getDistance(m_unit->getPosition());
      //not a threat if the unit is a worker or if the unit is a building that cannot attack
      bool isThreat = !(unit->getType().isWorker() || (unit->getType().isBuilding() && !unit->getType().canAttack()));
      bool inRange = m_unit->isInWeaponRange(unit);

      if( (!threatFound || isThreat)
        && ((!threatFound && isThreat)
          || dist < mindist
          || (inRange && hp < minhp)))
      {
        minhp = hp;
        mindist = dist;
        target = unit;
        if(isThreat)
          threatFound = true;
      }
    }
    if(target != NULL)
      std::string temp = target->getType().getName();
    return target;
}
