#pragma once
#include "../../UnitBehavior.h"
#include "../../BehaviorState.h"
#include "../../Layer1FSM.h"
#include "../../Actions/AttackUnitAction.h"
#include "../../Conditions/CompositeCondition.h"
#include "../../Conditions/HasCriticalHealth.h"
#include "../../Conditions/HasRangedAttack.h"
#include "../../Conditions/IsTargeted.h"
#include "../../Conditions/IsWorthFleeing.h"
#include "../../../../../UnitManager.h"
#include "../../../../../../Utils/ReferenceCounted.h"
#include "UnitAttackFSM.h"
#include <BWAPI.h>

//TODO: Needs to be cleaned!!!

class DefaultAttackFSM : public UnitAttackFSM<DefaultAttackFSM>
{
  class Engaging;
  class Idling;

public:
  DefaultAttackFSM(const Process* parent, unsigned int initialPriority, BWAPI::Unit* unit)
    : UnitAttackFSM(parent, initialPriority, unit)
    , IUnitAttackFSM(parent, initialPriority, unit)
    , IUnitBehavior(parent, initialPriority, unit)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
    , m_engagingState(*this)
    , m_idlingState(*this)
    , m_states(m_engagingState, m_idlingState)
  {
  }

  virtual std::string getName() const
  {
    return "DefaultAttackFSM (" + getObjectDescription(m_unit) + ")";
  }

  bool selectTarget(const UnitGroup& enemies);

protected:
  virtual void executeImpl()
  {
    UnitBehavior::executeImpl();
    m_idlingState.onEnter();
  }

private:
  BWAPI::Unit* selectTargetImpl(UnitGroup enemies);

  BWAPI::Unit* m_target;

  // Engaging state ////////////////////////////////////////////////////////////
  class Engaging : public BehaviorState<DefaultAttackFSM, Engaging>
  {
  public:
    Engaging(DefaultAttackFSM& behavior);
    void onEnterImpl();
    void onExitImpl();

    //void onTimer();

    void targetChanged(BWAPI::Unit* oldTarget);
  private:

    layer1::AttackUnitAction m_attackAction;

    typedef Transition<Engaging, NoCondition, NoTransitionAction, Idling> TargetDefeatedTransition;
    TargetDefeatedTransition m_targetDefeated;

    //typedef SCHEDULER_EVENT(Engaging, onTimer) TimerEvent;
    //TimerEvent m_timerEvent;

    UnitManager::OnIndividualUnitDestroy::SubscriberID m_targetDefeatedId;

  public:
    typedef boost::tuple<> Events;
    const Events m_events;

    typedef boost::tuple<
      TargetDefeatedTransition&
    > Transitions;
    const Transitions m_transitions;
  };

  
  // Idling state //////////////////////////////////////////////////////////////
  class Idling : public BehaviorState<DefaultAttackFSM, Idling>
  {
  public:
    Idling(DefaultAttackFSM& behavior);
    void onEnterImpl() {}
    void onExitImpl() {}

    void onTargetChanged();

  private:
    typedef Transition<Idling, NoCondition, NoTransitionAction, Engaging> TargetAssignedTransition;
    TargetAssignedTransition m_transition;
    
  public:
    typedef boost::tuple<> Events;
    const Events m_events;

    typedef boost::tuple<
      TargetAssignedTransition&
    > Transitions;
    const Transitions m_transitions;
  };

  Engaging m_engagingState;
  Idling m_idlingState;

public:
  typedef boost::tuple<
    Engaging&,
    Idling&
  > States;
  const States m_states;
};