#include "stdafx.h"
#include "DragoonAttack.h"
#include <BWAPI.h>
#include <cassert>

// Engaging state //////////////////////////////////////////////////////////////
DragoonAttackFSM::Engaging::Engaging(DragoonAttackFSM& behavior)
  : BehaviorState(behavior)
  , m_attackAction(*this)
  , m_targetDefeated(*this, NoCondition(), NoTransitionAction(), behavior.m_idlingState)
  , m_lowHealth(
      *this,
      CompositeCondition<HasCriticalHealth,IsWorthFleeing>(
        HasCriticalHealth(behavior.m_unit),
        IsWorthFleeing(behavior.m_unit)),
      NoTransitionAction(),
      behavior.m_fleeingState)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 0, 1)
  , m_events()
  , m_transitions(m_targetDefeated, m_lowHealth)
{
}

void DragoonAttackFSM::Engaging::onEnterImpl()
{
  assert(m_context.m_target != NULL);

  //selectTarget();
 /* if(m_context.m_target!= NULL)
  {*/
    m_targetDefeatedId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_context.m_target, &m_targetDefeated, NULL);

    m_attackAction.setUnitToAttack(m_context.m_target);

    onTimer();
  //}
}

void DragoonAttackFSM::Engaging::onExitImpl()
{
  if(m_context.m_target!= NULL)
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_context.m_target, m_targetDefeatedId);
  m_timerEvent.deactivate();
}

void DragoonAttackFSM::Engaging::onTimer()
{
  if(m_context.m_unit->isAttackFrame())
  {
    m_timerEvent.setWaitingTimeBeforeExecution(10);
    m_timerEvent.activate();
  } else {
   // if(m_context.m_target= NULL)
    //  m_targetDefeated.testTransition(boost::make_tuple());
    if(!m_lowHealth.testTransition(boost::make_tuple()))
    {
      if(m_context.m_target->exists())
        Layer1FSM::getInstance().getArbitrator().
          submitAction(m_context.m_unit, m_attackAction, Arbitration::ActionImportance(*this));
      // Not sure if this is the right time to wait
      size_t nextTick = static_cast<size_t>(m_context.m_unit->getGroundWeaponCooldown()+10);
      if(nextTick <= 1)
        nextTick = static_cast<size_t>(m_context.m_unit->getType().groundWeapon().damageCooldown()+1);

      m_timerEvent.setWaitingTimeBeforeExecution(nextTick);
      m_timerEvent.activate();
    }
  }
}

void DragoonAttackFSM::Engaging::targetChanged(BWAPI::Unit* oldTarget)
{
  if(oldTarget != NULL)
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(oldTarget, m_targetDefeatedId);
  if(m_context.m_target != NULL)
    m_targetDefeatedId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_context.m_target, &m_targetDefeated, NULL);

  m_attackAction.setUnitToAttack(m_context.m_target);
}

// Fleeing state ///////////////////////////////////////////////////////////////
DragoonAttackFSM::Fleeing::Fleeing(DragoonAttackFSM& behavior)
  : BehaviorState(behavior)
  , m_moveAction(*this)
  , m_transition(*this, Not<IsTargeted>(IsTargeted(behavior.m_unit)), NoTransitionAction(), behavior.m_idlingState)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_events(m_timerEvent)
  , m_transitions(m_transition)
{
}

void DragoonAttackFSM::Fleeing::onEnterImpl()
{
  BWAPI::Position position = Spar->getPerceptualState().getSafeRetreatPosition(m_context.m_unit);
  m_moveAction.setDestination(position);

  Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.m_unit, m_moveAction, Arbitration::ActionImportance(*this));
}

void DragoonAttackFSM::Fleeing::onExitImpl()
{
}

void DragoonAttackFSM::Fleeing::onTimer()
{
  m_transition.testTransition(boost::make_tuple());
}

// Idling state ////////////////////////////////////////////////////////////////
DragoonAttackFSM::Idling::Idling(DragoonAttackFSM& behavior)
  : BehaviorState(behavior)
  , m_transition(*this, NoCondition(), NoTransitionAction(), behavior.m_engagingState)
  , m_transitions(m_transition)
{
}

void DragoonAttackFSM::Idling::onTargetChanged()
{
  // Transition to Engaging state
  m_transition.testTransition(boost::make_tuple());
}

bool DragoonAttackFSM::selectTarget(const UnitGroup& enemies)
{
    BWAPI::Unit* oldTarget = m_target;
    m_target = selectTargetImpl(enemies);
    if(m_currentState == &m_engagingState)
      m_engagingState.targetChanged(oldTarget);
    if(m_target == NULL)
      return false;
    if (m_currentState == &m_idlingState)
      m_idlingState.onTargetChanged();

    return true;
}

BWAPI::Unit* DragoonAttackFSM::selectTargetImpl(UnitGroup enemies)
{
    int minhp = boost::numeric::bounds<int>::highest();
    int mindist = boost::numeric::bounds<int>::highest();
    bool threatFound = false;
    BWAPI::Unit* target = NULL;

    for(UnitGroup::const_iterator i=enemies.begin(); i!=enemies.end(); ++i)
    {
      BWAPI::Unit* unit = *i;
      //Ignore invisible units, eggs and larvae
      if(!unit->isDetected() 
        || unit->getType() == BWAPI::UnitTypes::Zerg_Larva 
        || unit->getType() == BWAPI::UnitTypes::Zerg_Egg)
        continue;
      int hp = unit->getHitPoints() + unit->getShields();
      int dist = unit->getDistance(m_unit->getPosition());
      //not a threat if the unit is a worker or if the unit is a building that cannot attack
      bool isThreat = !(unit->getType().isWorker() || (unit->getType().isBuilding() && !unit->getType().canAttack()));
      bool inRange = m_unit->isInWeaponRange(unit);


      if( (!threatFound || isThreat)
        && ((!threatFound && isThreat)
          || dist < mindist
          || (inRange && hp < minhp)))
      {
        minhp = hp;
        mindist = dist;
        target = unit;
        if(isThreat)
          threatFound = true;
      }
    }
    if(target != NULL)
      std::string temp = target->getType().getName();
    return target;
}
