#include "stdafx.h"
#include <BWAPI.h>
#include <functional>
#include "GroupAttack.h"
#include "UnitAttackFSM.h"
#include "../../Behavior.h"
#include "../../BehaviorState.h"
#include "../../../../../PerceptualState/Locations/Dynamic/RectangularGroupDynamicLocation.h"
#include "../../../../../PerceptualState/PerceptualState.h"
#include "../../../../../PersistentUnitGroup.h"
#include "../../../../../SparAIContainer.h"
#include "../../../../../Utils/SafeList.h"
#include "../../../../../Utils/FSM/FSM.h"
#include <boost/numeric/conversion/bounds.hpp>

void GroupAttack::onUnitAddedToGroup(void* data, BWAPI::Unit* unit)
{
  if(getCurrentState() == static_cast<IFSMState*>(&m_goToState))
    m_goToState.unitAdded(unit);
  else if(getCurrentState() == static_cast<IFSMState*>(&m_engagingState))
    m_engagingState.unitAdded(unit);
  else
    assert(false);
}

void GroupAttack::onUnitRemovedFromGroup(void* data, BWAPI::Unit* unit)
{
  stdext::hash_map<BWAPI::Unit*, BehaviorWrapper>::iterator pos = m_subBehaviors.find(unit);
  if(pos != m_subBehaviors.end())
    pos->second.behavior->externalTerminate();
  if(unit->getType() == BWAPI::UnitTypes::Protoss_Observer)
    m_observerGroup.remove(unit);
}

// Global functions ////////////////////////////////////////////////////////////
namespace
{
  /* struct SetTarget : std::unary_function<Process*,SafeListBase::FunctionReturn>
  {
  SetTarget(BWAPI::Unit* target) : m_target(target) {}

  SafeListBase::FunctionReturn operator()(Process* fsm) const
  {
  dynamic_cast<UnitAttackFSM*>(fsm)->setTarget(m_target);
  return SafeListBase::FunctionReturn();
  }

  private:
  BWAPI::Unit* m_target;
  };*/

  /*
  * Fine-tune this function to return the desired target to attack from the set
  */
  BWAPI::Unit* findTarget(const UnitGroup& enemies, const BWAPI::Position& centroid)
  {
    int minhp = boost::numeric::bounds<int>::highest();
    int mindist = boost::numeric::bounds<int>::highest();
    bool threatFound = false;
    BWAPI::Unit* target = NULL;

    for(UnitGroup::const_iterator i=enemies.begin(); i!=enemies.end(); ++i)
    {
      BWAPI::Unit* unit = *i;
      if(!unit->isVisible(BWAPI::Broodwar->self()))
        continue;
      int hp = unit->getHitPoints() + unit->getShields();
      int dist = unit->getDistance(centroid);
      //not a threat if the unit is a worker or if the unit is a building that cannot attack
      bool isThreat = !(unit->getType().isWorker() || (unit->getType().isBuilding() && !unit->getType().canAttack()));
      bool test = unit->getType().isBuilding();

      //  if ((hp  < minhp) ||                        old condition: caused zealots to try and walk around the whole army to snipe units in the back...
      //      (hp == minhp  && dist < mindist))       may require a different target for different unit types?  i.e., zealot attack closest, dragoons snipe weakest
      if(dist < mindist) //TODO: make more precise.  e.g., Attack units before buildings.
      {
        if(isThreat || !threatFound){
          minhp = hp;
          mindist = dist;
          target = unit;
        }
      }
    }
    if(target !=NULL)
      std::string name = target->getType().getName();
    return target;
  }
};

// GoToLocation state //////////////////////////////////////////////////////////
GroupAttack::GoToLocation::GoToLocation(GroupAttack& behaviour, Location* location)
: BehaviorState(behaviour)
, m_location(location)
, m_isAttacking(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 13, 5)
, m_transitionToEngaging(*this,
                         GroupIsAttacking(behaviour.getGroup()), 
                         NoTransitionAction(),
                         behaviour.m_engagingState)
                         , m_transitionToTerminate(*this,
                         GroupReachedDestinationLocation<PersistentUnitGroup>(behaviour.getGroup(),location), 
                         NoTransitionAction(),
                         behaviour.m_terminateState)
                         , m_events(m_isAttacking)
                         , m_transitions(m_transitionToEngaging,m_transitionToTerminate)
{time=0;}

void GroupAttack::GoToLocation::onEnterImpl()
{
  BWAPI::Unit* observerTarget = NULL;
  int smallestDist = boost::numeric::bounds<int>::highest();    
  for(PersistentUnitGroup::const_iterator it=m_context.getGroup()->begin(); it != m_context.getGroup()->end(); ++it)
  {
    if((*it)->getType() != BWAPI::UnitTypes::Protoss_Observer)
    {
      int dist = (*it)->getDistance(m_location->getCurrentCenter());
      if(smallestDist > dist)
      {
        smallestDist = dist;
        observerTarget = (*it);
      }
      assert((*it)->canIssueCommand(BWAPI::UnitCommand::attack((*it), m_location->getCurrentCenter())));
      (*it)->attack(m_location->getCurrentCenter());
    }
    else
    {
      m_context.m_observerGroup.insert(*it);
    }
  }

  if(observerTarget != NULL)
  {
    for(PersistentUnitGroup::const_iterator it = m_context.m_observerGroup.begin(); it != m_context.m_observerGroup.end(); ++it)
    {
      assert((*it)->canIssueCommand(BWAPI::UnitCommand::follow((*it), observerTarget)));
      (*it)->follow(observerTarget);
    }
  }
}

void GroupAttack::GoToLocation::onTimer()
{
  if(time == 3)
  {
    BWAPI::Unit* observerTarget = NULL;
    int smallestDist = boost::numeric::bounds<int>::highest();    
    for(PersistentUnitGroup::const_iterator it=m_context.getGroup()->begin(); it != m_context.getGroup()->end(); ++it)
    {
      if((*it)->getType() != BWAPI::UnitTypes::Protoss_Observer)
      {
        int dist = (*it)->getDistance(m_location->getCurrentCenter());
        if(smallestDist > dist)
        {
          smallestDist = dist;
          observerTarget = (*it);
        }
        assert((*it)->canIssueCommand(BWAPI::UnitCommand::attack((*it), m_location->getCurrentCenter())));
        (*it)->attack(m_location->getCurrentCenter());
      }
    }

    if(observerTarget != NULL)
    {
      for(PersistentUnitGroup::const_iterator it = m_context.m_observerGroup.begin(); it != m_context.m_observerGroup.end(); ++it)
      {
        assert((*it)->canIssueCommand(BWAPI::UnitCommand::follow((*it), observerTarget)));
        (*it)->follow(observerTarget);
      }
    }
    time = 0;
  }
  else
  {
    time++;
  }

  m_transitionToEngaging.testTransition(boost::make_tuple());
  m_transitionToTerminate.testTransition(boost::make_tuple());
}

void GroupAttack::GoToLocation::unitAdded(BWAPI::Unit* unit)
{
  // TODO: if the group only had observer, they will not follow the unit added
  if(unit->getType() == BWAPI::UnitTypes::Protoss_Observer)
  {
    m_context.m_observerGroup.insert(unit);
  }
  else
  {
    assert(unit->canIssueCommand(BWAPI::UnitCommand::attack(unit, m_location->getCurrentCenter())));
    unit->attack(m_location->getCurrentCenter());
  }
}

// Engaging state //////////////////////////////////////////////////////////////

GroupAttack::Engaging::Engaging(GroupAttack& behavior)
: BehaviorState(behavior)
, m_enemySighted(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 17, 5)
, m_onSubBehaviorTerminated(*this)
, m_transitionToGoTo(*this,
                     NoCondition(), 
                     NoTransitionAction(),
                     behavior.m_goToState)
, m_events(m_enemySighted)
, m_transitions(m_transitionToGoTo)
{
}

void GroupAttack::Engaging::onEnterImpl()
{
  for (PersistentUnitGroup::const_iterator i=m_context.getGroup()->begin(); i!=m_context.getGroup()->end(); ++i)
  {
    BWAPI::Unit* unit = *i;
    if(unit->getType() != BWAPI::UnitTypes::Protoss_Observer)
    {
      IUnitAttackFSM* behavior = IUnitAttackFSM::NewAttackFSM(&m_context, m_context.getCurrentPriority() * 2 + 1, unit); //= new UnitAttackFSM(&m_context, m_context.getCurrentPriority() * 2 + 1, unit);
      Process::zombifyOnTerminate(behavior);

      BehaviorWrapper wrapper;
      wrapper.behavior = behavior;
      wrapper.subscriberId = behavior->getTerminateEvent().subscribe(&m_onSubBehaviorTerminated, unit);

      m_context.m_subBehaviors.insert(std::make_pair(unit, wrapper));

      behavior->execute();
    }
  }

  selectTarget();
}

void GroupAttack::Engaging::onTimer()
{
  BWAPI::Unit* observerTarget = NULL;
  int smallestDist = boost::numeric::bounds<int>::highest();    
  for(PersistentUnitGroup::const_iterator it=m_context.getGroup()->begin(); it != m_context.getGroup()->end(); ++it)
  {
    if((*it)->getType() != BWAPI::UnitTypes::Protoss_Observer)
    {
      int dist = (*it)->getDistance(m_context.m_location->getCurrentCenter());
      if(smallestDist > dist)
      {
        smallestDist = dist;
        observerTarget = (*it);
      }
    }
  }

  if(observerTarget != NULL)
  {
    for(PersistentUnitGroup::const_iterator it = m_context.m_observerGroup.begin(); it != m_context.m_observerGroup.end(); ++it)
    {
      assert((*it)->canIssueCommand(BWAPI::UnitCommand::follow((*it), observerTarget)));
      (*it)->follow(observerTarget);
    }
  }

  if(!selectTarget())
    m_transitionToGoTo.testTransition(boost::make_tuple());
}

void GroupAttack::Engaging::unitAdded(BWAPI::Unit* unit)
{
  if(unit->getType() == BWAPI::UnitTypes::Protoss_Observer)
  {
    m_context.m_observerGroup.insert(unit);
  }
  else
  {
    IUnitAttackFSM* behavior = IUnitAttackFSM::NewAttackFSM(&m_context, m_context.getCurrentPriority() * 2 + 1, unit); //= new UnitAttackFSM(&m_context, m_context.getCurrentPriority() * 2 + 1, unit);
    Process::zombifyOnTerminate(behavior);

    BehaviorWrapper wrapper;
    wrapper.behavior = behavior;
    wrapper.subscriberId = behavior->getTerminateEvent().subscribe(&m_onSubBehaviorTerminated, unit);

    m_context.m_subBehaviors.insert(std::make_pair(unit, wrapper));

    behavior->execute();
  }
}

void GroupAttack::Engaging::onSubBehaviorTerminated(void* data)
{
  BWAPI::Unit* unit = reinterpret_cast<BWAPI::Unit*>(data);
  stdext::hash_map<BWAPI::Unit*, BehaviorWrapper>::iterator pos = m_context.m_subBehaviors.find(unit);

  pos->second.behavior->getTerminateEvent().unsubscribe(pos->second.subscriberId);

  m_context.m_subBehaviors.erase(pos);
}

bool GroupAttack::Engaging::selectTarget()
{
  UnitGroup enemies = Spar->getPerceptualState().getEnemyUnits(RectangularGroupDynamicLocation(m_context.getGroup()));
  /*BWAPI::Unit* target = findTarget(enemies, m_context.getGroup()->getCenter());

  if (target != NULL)
  {
  m_context.for_each_sub_behavior(SetTarget(target));
  return true;
  }
  return false;*/

  bool targetFound = false;
  for(stdext::hash_map<BWAPI::Unit*, BehaviorWrapper>::iterator it = m_context.m_subBehaviors.begin(); it != m_context.m_subBehaviors.end(); ++it)
  {
    if(it->second.behavior->selectTarget(enemies))
      targetFound = true;
  }
  return targetFound;

}

GroupAttack::Terminate::Terminate(GroupAttack& behaviour)
: BehaviorState(behaviour)
{}

void GroupAttack::Terminate::onEnterImpl()
{
  m_context.externalTerminate();
}