#include "stdafx.h"
#include "GatherMinerals.h"

const Process::TerminationMode GatherMinerals::NoMoreMinerals;

///////////////////////////////////////////////////
// Mining state
///////////////////////////////////////////////////

GatherMinerals::Mining::Mining(GatherMinerals& behavior)
  : BehaviorState(behavior)
  , m_crystalRemovedEventHandler(*this)
  , m_transitions()
  , m_event(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 19, 5)
  , m_events(m_event)
{
  defence=NULL;
  m_target=NULL;
  isAttacking=false;
}

void GatherMinerals::Mining::onEnterImpl()
{
  std::pair<BaseLocation::crystal_iterator, BaseLocation::crystal_iterator> crystals = m_context.m_baseLocation->getLastKnownCrystalsLeft();

  m_crystalRemovedSubscriberId = m_context.m_baseLocation->getNeutralUnitRemovedEvent().subscribe(&m_crystalRemovedEventHandler, NULL);
  if(crystals.first != crystals.second)
  {
    BaseLocation::crystal_iterator mineralIt = crystals.first;
    // *mineralIt is type "const OwnUnit*"
    // (*mineralIt)->isAccessible() is false if the mineral isn't in the field of view

    m_context.m_lastMineral = (*mineralIt)->getUnit();

    for(PersistentUnitGroup::const_iterator it = m_context.getGroup()->begin(); it != m_context.getGroup()->end(); ++it)
    {
      assert((*it)->getType() == BWAPI::UnitTypes::Protoss_Probe);
      // TODO: use arbitrator
      assert((*it)->canIssueCommand(BWAPI::UnitCommand::gather((*it), (*mineralIt)->getUnit())));
      (*it)->gather((*mineralIt)->getUnit());
      m_context.m_lastMineral = (*mineralIt)->getUnit();
      ++mineralIt;
      if(mineralIt == crystals.second)
        mineralIt = crystals.first;
    }
  } else {
    m_context.terminate(NoMoreMinerals);
  }
}

void GatherMinerals::Mining::onNeutralBuildingRemoved(void* data, NeutralUnit* crystal)
{
  if (crystal->getType().isMineralField())
  {
    std::pair<BaseLocation::crystal_iterator, BaseLocation::crystal_iterator> crystals = m_context.m_baseLocation->getLastKnownCrystalsLeft();
    if(crystals.first == crystals.second)
    {
      m_context.terminate(NoMoreMinerals);
    }
  }
}

void GatherMinerals::Mining::onUnitAddedToGroup(BWAPI::Unit* unit)
{
  std::pair<BaseLocation::crystal_iterator, BaseLocation::crystal_iterator> crystals = m_context.m_baseLocation->getLastKnownCrystalsLeft();

  assert(crystals.first != crystals.second);
  
  BaseLocation::crystal_iterator mineralIt = crystals.first;

  while(mineralIt != crystals.second && (*mineralIt)->getUnit() != m_context.m_lastMineral)
    ++mineralIt;

  if(mineralIt != crystals.second)
    ++mineralIt;
  if(mineralIt == crystals.second)
    mineralIt = crystals.first;
  m_context.m_lastMineral = (*mineralIt)->getUnit();

  // TODO: use arbritrator
  assert(unit->canIssueCommand(BWAPI::UnitCommand::gather(unit, (*mineralIt)->getUnit())));
  unit->gather((*mineralIt)->getUnit());
}

void GatherMinerals::Mining::onExitImpl()
{
  m_context.m_baseLocation->getNeutralUnitRemovedEvent().unsubscribe(m_crystalRemovedSubscriberId);
}