#include "stdafx.h"
#include "LearningCombat.h"
#include "../Layer1FSM.h"

///////////////////////////////////////////////////
// No Action state

BWAPI::Unit* findClosestEnemy(BWAPI::Unit* unit);
BWAPI::Unit* findRandomEnemy(BWAPI::Unit* unit);
BWAPI::Unit* findStrongestEnemy(BWAPI::Unit* unit);
BWAPI::Unit* findDyingEnemy(BWAPI::Unit* unit);

LearningCombat::NoAction::NoAction(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::NoAction::onEnterImpl()
{
  //Do nothing
}

///////////////////////////////////////////////////
// AttackRandom state
LearningCombat::AttackRandom::AttackRandom(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::AttackRandom::onEnterImpl()
{
  for(PersistentUnitGroup::const_iterator it = m_context.getGroup()->begin() ; it != m_context.getGroup()->end() ; ++it)
  {
    std::map<BWAPI::Unit*,BWAPI::Unit*>::iterator target = m_context.mapOfTarget.find((*it));
    target->second = findRandomEnemy(*it);
    if(m_context.fleeingUnit.find(*it)==m_context.fleeingUnit.end() && target->second != NULL)
    {
      (*it)->attack(target->second);
    }
  }
}

  ///////////////////////////////////////////////////
// AttackStrongest state
LearningCombat::AttackStrongest::AttackStrongest(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::AttackStrongest::onEnterImpl()
{
  //Determine the strongest enemy and set all unit to attack it
  for(PersistentUnitGroup::const_iterator it = m_context.getGroup()->begin() ; it != m_context.getGroup()->end() ; ++it)
  {
    std::map<BWAPI::Unit*,BWAPI::Unit*>::iterator target = m_context.mapOfTarget.find((*it));
    target->second = findStrongestEnemy(*it);
    if(m_context.fleeingUnit.find(*it)==m_context.fleeingUnit.end() && target->second != NULL)
    {
      (*it)->attack(target->second);
    }
  }
}

  ///////////////////////////////////////////////////
// AttackDying state
LearningCombat::AttackDying::AttackDying(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::AttackDying::onEnterImpl()
{
  for(PersistentUnitGroup::const_iterator it = m_context.getGroup()->begin() ; it != m_context.getGroup()->end() ; ++it)
  {
    std::map<BWAPI::Unit*,BWAPI::Unit*>::iterator target = m_context.mapOfTarget.find((*it));
    target->second = findDyingEnemy(*it);
    if(m_context.fleeingUnit.find(*it)==m_context.fleeingUnit.end() && target->second != NULL)
    {
      (*it)->attack(target->second);
    }
  }
}

  ///////////////////////////////////////////////////
// AttackClosest state
LearningCombat::AttackClosest::AttackClosest(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::AttackClosest::onEnterImpl()
{
  //for each unit make it attack its closest enemy
  for(PersistentUnitGroup::const_iterator it = m_context.getGroup()->begin() ; it != m_context.getGroup()->end() ; ++it)
  {
    std::map<BWAPI::Unit*,BWAPI::Unit*>::iterator target = m_context.mapOfTarget.find((*it));
    target->second = findClosestEnemy(*it);
    if(m_context.fleeingUnit.find(*it)==m_context.fleeingUnit.end() && target->second != NULL)
    {
      (*it)->attack(target->second);
    }
  }
}

  ///////////////////////////////////////////////////
// Flee state
LearningCombat::Flee::Flee(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_moveAction(*this)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::Flee::onEnterImpl()
{
  //find a safe retreat position and give order to move there
  BWAPI::Position destination = Spar->getPerceptualState().
    getSafeRetreatPosition(*(m_context.getGroup()->begin())); //, m_context.m_enemyUnits);

  m_moveAction.setDestination(destination);
  Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.getGroup(), m_moveAction, Arbitration::ActionImportance(*this));
}

  ///////////////////////////////////////////////////
// RemoveTargetedUnit state
LearningCombat::RemoveTargetedUnit::RemoveTargetedUnit(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::RemoveTargetedUnit::onEnterImpl()
{
  //find targeted unit and make them flee
  for(PersistentUnitGroup::const_iterator it = m_context.getGroup()->begin() ; it != m_context.getGroup()->end() ; ++it)
  {
    if((*it)->isUnderAttack() && 0.30 < ((*it)->getHitPoints())/((*it)->getType().maxHitPoints()))
    {
      m_context.fleeingUnit.insert(*it);
      BWAPI::Position destination = Spar->getPerceptualState().
        getSafeRetreatPosition(*(m_context.getGroup()->begin()));
      (*it)->move(destination);
    }
  }
}

  ///////////////////////////////////////////////////
// PutBackSafeUnit state
LearningCombat::PutBackSafeUnit::PutBackSafeUnit(LearningCombat& behavior)
: BehaviorState(behavior)
, m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToARState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackRandomState)
, m_transitionToASState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackStrongestState)
, m_transitionToADState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackDyingState)
, m_transitionToACState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_attackClosestState)
, m_transitionToFleeState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_fleeState)
, m_transitionToRemoveState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_removeTargetedUnitState)
, m_transitionToPutBackState(
                        *this,
                        NoCondition(), 
                        NoTransitionAction(), 
                        behavior.m_putBackSafeUnitState)
, m_events(m_timerEvent)
, m_transitions(m_transitionToARState,m_transitionToASState,m_transitionToADState,m_transitionToACState,m_transitionToFleeState,m_transitionToRemoveState,m_transitionToPutBackState)
{
}

void LearningCombat::PutBackSafeUnit::onEnterImpl()
{
  //find fleeing unit and make them attack again
  for(std::set<BWAPI::Unit*>::iterator it = m_context.fleeingUnit.begin() ; it!=m_context.fleeingUnit.end() ; ++it)
  {
    if(!(*it)->isUnderAttack())
    {
      std::map<BWAPI::Unit*,BWAPI::Unit*>::iterator target = m_context.mapOfTarget.find((*it));
      assert(target != m_context.mapOfTarget.end());
      if(target->second!=NULL)
      {
        if(target->second->exists()){
          (*it)->attack(target->second);
        }
        else
        {
          target->second = NULL;
        }
      }
      if(target->second==NULL)
      {
        (*it)->stop();
      }
      m_context.fleeingUnit.erase(*it);
    }
  }
}

BWAPI::Unit* findClosestEnemy(BWAPI::Unit* unit)
{
  BWAPI::Unit* target = NULL;
  int range = 256;
  while(target == NULL && range <= 1024)
  {
    UnitGroup units = Spar->getPerceptualState().getEnemiesNearUnit(unit,range);
    int closest = 1025;
    for(UnitGroup::iterator it = units.begin() ; it != units.end() ; ++it)
    {
      int tmpDistance = unit->getDistance(*it);
      if(tmpDistance < closest){
        closest = tmpDistance;
        target = *it;
      }
    }
    if(target!=NULL)
    {
      return target;
    }
    else
    {
      range = range*2;
    }
  }
  return NULL;
}

BWAPI::Unit* findDyingEnemy(BWAPI::Unit* unit)
{
  BWAPI::Unit* target = NULL;
  int range = 256;
  while(target == NULL && range <= 1024)
  {
    UnitGroup units = Spar->getPerceptualState().getEnemiesNearUnit(unit,range);
    int closest = 1025;
    for(UnitGroup::iterator it = units.begin() ; it != units.end() ; ++it)
    {
      int tmpDistance = (*it)->getHitPoints();
      if(tmpDistance < closest){
        closest = tmpDistance;
        target = *it;
      }
    }
    if(target!=NULL)
    {
      return target;
    }
    else
    {
      range = range*2;
    }
  }
  return NULL;
}

BWAPI::Unit* findStrongestEnemy(BWAPI::Unit* unit)
{
  BWAPI::Unit* target = NULL;
  int range = 256;
  while(target == NULL && range <= 1024)
  {
    UnitGroup units = Spar->getPerceptualState().getEnemiesNearUnit(unit,range);
    int closest = -1;
    for(UnitGroup::iterator it = units.begin() ; it != units.end() ; ++it)
    {
      int tmpDistance = (*it)->getType().groundWeapon().damageAmount() + (*it)->getType().airWeapon().damageAmount();
      if(tmpDistance > closest){
        closest = tmpDistance;
        target = *it;
      }
    }
    if(target!=NULL)
    {
      return target;
    }
    else
    {
      range = range*2;
    }
  }
  return NULL;
}

BWAPI::Unit* findRandomEnemy(BWAPI::Unit* unit)
{
  UnitGroup units = Spar->getPerceptualState().getEnemiesNearUnit(unit,1024);
  int nb = rand() % (units.size() + 1);
  UnitGroup::iterator it = units.begin();
  for(int i = 0 ; i<nb ; i++)
  {
    ++it;
  }
  if(it==units.end())
  {
    return NULL;
  }
  else
  {
    return (*it);
  }
}