#include "stdafx.h"
#include "Move.h"

///////////////////////////////////////////////////
// Goto state
///////////////////////////////////////////////////

Move::Goto::Goto(Move& behavior, size_t refreshDelay)
  : BehaviorState(behavior)
  , m_moveAction(*this)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, refreshDelay, refreshDelay)
  , m_events(m_timerEvent)
  , m_transitions()
  , m_tickCounter(0)
{
}

void Move::Goto::onEnterImpl()
{
  m_moveAction.setDestination(m_context.m_destination);
  for(PersistentUnitGroup::const_iterator it = m_context.getGroup()->begin(); it != m_context.getGroup()->end(); ++it)
  {    
    assert(!(*it)->getType().isBuilding());
    //assert((*it)->getType().isFlyer() || BWTA::isConnected((*it)->getTilePosition(), m_context.m_destinationTP));
#ifdef _DEBUG
    if(!(*it)->canIssueCommand(BWAPI::UnitCommand::move((*it), m_moveAction.getDestination())))
      SPAR_LOG(LogTypes::GENERAL_ERROR, "Command move could not be issued for %s (%d)", (*it)->getType().getName().c_str(), (*it)->getID());
#endif

    m_context.unitsInMotion.insert(*it);
    Layer1FSM::getInstance().getArbitrator().submitAction((*it), m_moveAction, Arbitration::ActionImportance(*this));
  }
}

void Move::Goto::onUnitAddedToGroup(BWAPI::Unit* unit)
{
  if(m_context.isExecuting())
  {
    assert(!unit->getType().isBuilding());
    //assert(unit->getType().isFlyer() || BWTA::isConnected(unit->getTilePosition(), m_context.m_destinationTP));
#ifdef _DEBUG
    if(!unit->canIssueCommand(BWAPI::UnitCommand::move(unit, m_moveAction.getDestination())))
      SPAR_LOG(LogTypes::GENERAL_ERROR, "Command move could not be issued for %s (%d)", unit->getType().getName().c_str(), unit->getID());
#endif
    m_context.unitsInMotion.insert(unit);
    Layer1FSM::getInstance().getArbitrator().submitAction(unit, m_moveAction, Arbitration::ActionImportance(*this));
  }
}

void Move::Goto::onUnitRemovedFromGroup(BWAPI::Unit* unit)
{
  if(m_context.isExecuting())
  {
    size_t nbErased = m_context.unitsAtDestination.erase(unit);
    if(nbErased == 0)
      m_context.unitsInMotion.erase(unit);
  }
}