#include "stdafx.h"
#include "ObsScout.h"
#include "../../../../Utils/Logger.h"
#include "../Layer1FSM.h"

///////////////////////////////////////////////////
// GoToLocation State
///////////////////////////////////////////////////
ObsScout::GoToLocation::GoToLocation(ObsScout& behavior, Location* location)
: BehaviorState(behavior)
, m_moveAction(*this)
, m_location(location)
, m_locationReached(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 13)
, m_transitionToReached(*this,
                        UnitReachedDestination(behavior.getUnit(), location->getCurrentCenter(), 64), 
                        RaiseScoutArrived(&getBehavior()),
                        behavior.m_reachedState)
, m_events(m_locationReached)
                        , m_transitions(m_transitionToReached)
{
  cpt=0;
}

void ObsScout::GoToLocation::onEnterImpl()
{
}

void ObsScout::GoToLocation::onTimer()
{
  m_transitionToReached.testTransition(boost::make_tuple());
  //Check for 
  //Recalculate potential field and update current mouvement vector
  BWAPI::Position initPos = m_context.m_unit->getPosition();

  //Positive influance (goal)
  double posX =  m_location->getCurrentCenter().x()-initPos.x();
  double posY =  m_location->getCurrentCenter().y()-initPos.y();

  //Normalise
  double tmp = sqrt((posX * posX) + (posY * posY));
  if(tmp>200)
  {
    posX = posX/tmp;
    posY = posY/tmp;

    //Add a factor
    posX = posX*200;
    posY = posY*200;
  }

  //Negative influance (detector)
  BWAPI::Position topLeft(m_context.m_unit->getPosition().x() - 15*TILE_SIZE/2, m_context.m_unit->getPosition().y() - 15*TILE_SIZE/2);
  BWAPI::Position bottomRight(m_context.m_unit->getPosition().x() + 15*TILE_SIZE/2, m_context.m_unit->getPosition().y() + 15*TILE_SIZE/2);

  UnitGroup units = Spar->getPerceptualState().getEnemyUnits(RectangularLocation(topLeft, bottomRight));
  //std::set<BWAPI::Unit*> units = BWAPI::Broodwar->getUnitsInRadius(m_context.m_unit->getPosition(),15*TILE_SIZE);
  double negX = 0;
  double negY = 0;

  for(UnitGroup::iterator i = units.begin(); i!= units.end() ; ++i)
  {
    if((*i)->getPlayer() == BWAPI::Broodwar->enemy() && (*i)->getType().isDetector())
    {
      if((*i)->getType().isBuilding())
      {
        //sight range is 7
        if(m_context.m_unit->getDistance(*i) < 7.2*TILE_SIZE)
        {
          //invariant
          double tmpX = initPos.x()-m_location->getCurrentCenter().x();
          double tmpY = initPos.y()-m_location->getCurrentCenter().y();

          double tmpL = sqrt((tmpX * tmpX) + (tmpY * tmpY));
          tmpX = tmpX/tmpL;
          tmpY = tmpY/tmpL;

            //Add a factor
          tmpX = tmpX*250;
          tmpY = tmpY*250;

          negX += tmpX;
          negY += tmpY;
        }
        else
        {
          double tmpX = initPos.x()-m_location->getCurrentCenter().x();
          double tmpY = initPos.y()-m_location->getCurrentCenter().y();

          double tmpL = sqrt((tmpX * tmpX) + (tmpY * tmpY));
          tmpX = tmpX/tmpL;
          tmpY = tmpY/tmpL;

          if(tmpL/TILE_SIZE < 8)
          {
            tmpX = tmpX*100;
            tmpY = tmpY*100;
          }

          else if(tmpL/TILE_SIZE < 9)
          {
            tmpX = tmpX*75;
            tmpY = tmpY*75;
          }

          else if(tmpL/TILE_SIZE < 10)
          {
            tmpX = tmpX*50;
            tmpY = tmpY*50;
          }

          else if(tmpL/TILE_SIZE < 11)
          {
            tmpX = tmpX*25;
            tmpY = tmpY*25;
          }

          else
          {
            tmpX = 0;
            tmpY = 0;
          }

          negX += tmpX;
          negY += tmpY;
        }
      }
      else
      {
        //sight range is 11
        if(m_context.m_unit->getDistance(*i) < 11.2*TILE_SIZE)
        {
          //invariant
          double tmpX = initPos.x()-m_location->getCurrentCenter().x();
          double tmpY = initPos.y()-m_location->getCurrentCenter().y();

          double tmpL = sqrt((tmpX * tmpX) + (tmpY * tmpY));
          tmpX = tmpX/tmpL;
          tmpY = tmpY/tmpL;

            //Add a factor
          tmpX = tmpX*250;
          tmpY = tmpY*250;

          negX += tmpX;
          negY += tmpY;
        }
        else
        {
          double tmpX = initPos.x()-m_location->getCurrentCenter().x();
          double tmpY = initPos.y()-m_location->getCurrentCenter().y();

          double tmpL = sqrt((tmpX * tmpX) + (tmpY * tmpY));
          tmpX = tmpX/tmpL;
          tmpY = tmpY/tmpL;

          if(tmpL/TILE_SIZE < 12)
          {
            tmpX = tmpX*75;
            tmpY = tmpY*75;
          }

          else if(tmpL/TILE_SIZE < 13)
          {
            tmpX = tmpX*50;
            tmpY = tmpY*50;
          }

          else if(tmpL/TILE_SIZE < 14)
          {
            tmpX = tmpX*25;
            tmpY = tmpY*25;
          }

          else
          {
            tmpX = 0;
            tmpY = 0;
          }

          negX += tmpX;
          negY += tmpY;
        }
      }
    }
  }

  BWAPI::Position destination = BWAPI::Position(initPos.x() + posX + negX , initPos.y() + posY + negY);

  m_context.m_unit->move(destination);

  if(cpt==10)
  {
    if(lastDistance - m_context.m_unit->getDistance(m_location->getCurrentCenter()) < 50)
    {
      m_context.getScoutArrivedEvent().raise();
      //m_context.externalTerminate();
    }
    cpt=0;
  }
  if(cpt==0)
  {
    lastDistance = m_context.m_unit->getDistance(m_location->getCurrentCenter());
  }
  cpt++;
}

///////////////////////////////////////////////////
// Reached State
///////////////////////////////////////////////////
ObsScout::Reached::Reached(ObsScout& behavior)
: BehaviorState(behavior)
, m_events()
, m_transitions()
{}

void ObsScout::Reached::onEnterImpl()
{
  m_context.externalTerminate();
}
