#include "stdafx.h"
#include "ReaverControl.h"
#include "../../Layer1FSM.h"
#include "../../../../../PerceptualState/Locations/Static/Arbitrary/CircularLocation.h"

///////////////////////////////////////////////////
// IsLoaded state

ReaverControl::IsLoaded::IsLoaded(ReaverControl& behavior)
: BehaviorState(behavior)
, m_unitIsUnloaded(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToBuildState(
                           *this,
                           UnitOnGround(behavior.m_unit), 
                           NoTransitionAction(), 
                           behavior.m_buildState)
, m_events(m_unitIsUnloaded)
, m_transitions(m_transitionToBuildState)
{
}

void ReaverControl::IsLoaded::onEnterImpl()
{
}

///////////////////////////////////////////////////
// Build state

ReaverControl::Build::Build(ReaverControl& behavior)
: BehaviorState(behavior)
, m_buildScarab(*this)
, m_unitIsUnloaded(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 15, 15)
, m_transitionToAttackState(
                            *this,
                            UnitOnGround(behavior.m_unit),
                            NoTransitionAction(), 
                            behavior.m_attackState)
, m_transitionToLoadedState(
                            *this,
                            UnitLoaded(behavior.m_unit),
                            NoTransitionAction(), 
                            behavior.m_isLoadedState)
, m_events(m_unitIsUnloaded)
, m_transitions(m_transitionToAttackState, m_transitionToLoadedState)
{
}

void ReaverControl::Build::onEnterImpl()
{  
  Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_buildScarab, Arbitration::ActionImportance(*this));
}

///////////////////////////////////////////////////
// Attack state

ReaverControl::Attack::Attack(ReaverControl& behavior)
: BehaviorState(behavior)
, m_reaverAttack(*this)
, m_unitIsUnloaded(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 15, 15)
, m_transitionToBuildState(
                           *this,
                           UnitOnGround(behavior.m_unit),
                           NoTransitionAction(), 
                           behavior.m_buildState)
, m_transitionToLoadedState(
                           *this,
                           UnitLoaded(behavior.m_unit),
                           NoTransitionAction(), 
                           behavior.m_isLoadedState)
, m_events(m_unitIsUnloaded)
, m_transitions(m_transitionToBuildState, m_transitionToLoadedState)
{
}

BWAPI::Unit* getReaverTarget(BWAPI::Unit* reaver)
{
  // TODO: hardcoded values?

  
  BWAPI::Unit* target;
  BWAPI::Unit* secondChoice;
  double distance = 100000;
  BWAPI::Position position = reaver->getPosition();
  CircularLocation location = CircularLocation(position,256+32);
  UnitGroup enemyGroup = Spar->getPerceptualState().getEnemyUnits(location);
  bool noTarget = true;
  for (UnitGroup::const_iterator it = enemyGroup.begin(); it != enemyGroup.end(); ++it)
  {
    if((*it)->getType().canAttack() && !(*it)->getType().isFlyer() && !(*it)->getType().isFlyingBuilding()){
      int dist = (*it)->getDistance(reaver);
      if(dist<distance){
        distance = dist;
        target = *it;
      }
      noTarget=false;
    }else if(noTarget && !(*it)->getType().isFlyer()){
      secondChoice = *it;
      noTarget=false;
    }
  }
  if(noTarget){
    return NULL;
  }
  if(distance == 100000){
    return secondChoice;
  }
  return target;

}

void ReaverControl::Attack::onEnterImpl()
{  
  BWAPI::Unit* target = getReaverTarget(m_context.m_unit);
  if(target != NULL){
    m_reaverAttack.setTarget(target);
    Layer1FSM::getInstance().getArbitrator().
      submitAction(m_context.m_unit, m_reaverAttack, Arbitration::ActionImportance(*this));
  }
}