#pragma once
#include "../../UnitBehavior.h"
#include "../../BehaviorState.h"
#include "../../Arbitrator.h"
#include "../../Conditions/IsIdle.h"
#include "../../Conditions/UnitLoaded.h"
#include "../../Conditions/AtLeastOneUnit.h"
#include "../../Actions/ReaverActions/BuildScarab.h"
#include "../../Actions/ReaverActions/ReaverAttack.h"

class ReaverControl : public UnitBehavior<ReaverControl>
{
protected:
	class IsLoaded;
	class Build;
	class Attack;
	friend class IsLoaded;
	friend class Build;
	friend class Attack;

	class IsLoaded : public BehaviorState<ReaverControl, IsLoaded>
	{
	public:
		IsLoaded(ReaverControl& behavior);
		void onEnterImpl();
		void onExitImpl() {}

    void onTimer()
    {
      m_transitionToBuildState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(IsLoaded, onTimer) TimerEvent;

		typedef boost::tuple<
			TimerEvent&
			> Events;
		const Events m_events;
    
		typedef boost::tuple<
			Transition<IsLoaded, UnitOnGround, NoTransitionAction, Build>&
			> Transitions;
		const Transitions m_transitions;

	protected:
		// Variables

		// Actions

	    // Transitions
		TimerEvent m_unitIsUnloaded;
		Transition<IsLoaded, UnitOnGround, NoTransitionAction, Build> m_transitionToBuildState;
	};

	class Build : public BehaviorState<ReaverControl, Build>
	{
	public:
		Build(ReaverControl& behavior);
		void onEnterImpl();
		void onExitImpl() {}

    void onTimer()
    {
      m_transitionToAttackState.testTransition(boost::make_tuple());
      m_transitionToLoadedState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Build, onTimer) TimerEvent;

		typedef boost::tuple<
			TimerEvent&
			> Events;
		const Events m_events;
    
		typedef boost::tuple<
			Transition<Build, UnitOnGround, NoTransitionAction, Attack>&,
			Transition<Build, UnitLoaded, NoTransitionAction, IsLoaded>&
			> Transitions;
		const Transitions m_transitions;

	protected:
		// Variables

		// Actions
		layer1::BuildScarab m_buildScarab;

	    // Transitions
    TimerEvent m_unitIsUnloaded;
		Transition<Build, UnitOnGround, NoTransitionAction, Attack> m_transitionToAttackState;
		Transition<Build, UnitLoaded, NoTransitionAction, IsLoaded> m_transitionToLoadedState;
	};

	class Attack : public BehaviorState<ReaverControl, Attack>
	{
	public:
		Attack(ReaverControl& behavior);
		void onEnterImpl();
		void onExitImpl() {}		

    void onTimer()
    {
      m_transitionToBuildState.testTransition(boost::make_tuple());
      m_transitionToLoadedState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Attack, onTimer) TimerEvent;

		typedef boost::tuple<
			TimerEvent&
			> Events;
		const Events m_events;
    
		typedef boost::tuple<
			Transition<Attack, UnitOnGround, NoTransitionAction, Build>&,
			Transition<Attack, UnitLoaded, NoTransitionAction, IsLoaded>&
			> Transitions;
		const Transitions m_transitions;

	protected:
		// Variables

		// Actions
		layer1::ReaverAttack m_reaverAttack;

	    // Transitions
		TimerEvent m_unitIsUnloaded;
		Transition<Attack, UnitOnGround, NoTransitionAction, Build> m_transitionToBuildState;
		Transition<Attack, UnitLoaded, NoTransitionAction, IsLoaded> m_transitionToLoadedState;
	};

public:
	ReaverControl(const Process* parent, unsigned int initialPriority, BWAPI::Unit* reaver)
		: UnitBehavior(parent, initialPriority, reaver)
    , IUnitBehavior(parent, initialPriority, reaver)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
		, m_isLoadedState(*this), m_buildState(*this),m_attackState(*this)
    , m_states(m_isLoadedState, m_buildState, m_attackState)
	{
	}

	virtual std::string getName() const
	{
		return "ReaverControl (" + getObjectDescription(m_unit) + ")";
	}

  typedef boost::tuple<
      IsLoaded&,
      Build&,
      Attack&
    > States;
	const States m_states;

protected:
  virtual void executeImpl()
  {
    UnitBehavior::executeImpl();
    m_buildState.onEnter();
  }

  // Arguments

  // Variables

  // States
  IsLoaded m_isLoadedState;
  Build m_buildState;
  Attack m_attackState;
};