#include "stdafx.h"
#include "ReaverShuttleGroup.h"
#include "../../Layer1FSM.h"
#include <limits>
#include <BWAPI.h>

///////////////////////////////////////////////////
// SetPairs state
///////////////////////////////////////////////////

ReaverShuttleGroup::SetPairs::SetPairs(ReaverShuttleGroup& behavior)
  : BehaviorState(behavior)
  , m_event(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 1, 1)
  , m_transitionToPickUpState(*this,
                              NoCondition(), 
                              NoTransitionAction(),
                              behavior.m_pickUpState)
  , m_events(m_event)
  , m_transitions(m_transitionToPickUpState)
{
}

void ReaverShuttleGroup::SetPairs::onEnterImpl()
{
  UnitGroup::const_iterator it_shuttles = m_context.m_shuttles->begin();
  for(UnitGroup::const_iterator it_reavers = m_context.m_reavers->begin();
    it_reavers != m_context.m_reavers->end();
    ++it_reavers)
  {
    if(it_shuttles != m_context.m_shuttles->end()){
      m_context.m_pairs.insert(std::pair<BWAPI::Unit*,BWAPI::Unit*>((*it_shuttles),(*it_reavers)));
      it_shuttles++;
    }
  }
}

///////////////////////////////////////////////////
// PickUp state
///////////////////////////////////////////////////

ReaverShuttleGroup::PickUp::PickUp(ReaverShuttleGroup& behavior)
  : BehaviorState(behavior)
  , m_loadUnit(*this)
  , m_event(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToSendToTargetState(*this,
                                    AllPickedUp<UnitGroup>(behavior.m_reavers), 
                                    NoTransitionAction(),
                                    behavior.m_sendToTargetState)
  , m_events(m_event)
  , m_transitions(m_transitionToSendToTargetState)
{
}


void ReaverShuttleGroup::PickUp::onEnterImpl()
{
  //send them to pick reaver
  for(std::map<BWAPI::Unit*,BWAPI::Unit*>::const_iterator it = m_context.m_pairs.begin();it != m_context.m_pairs.end();it++){
    m_loadUnit.setUnit((*it).second);
    Layer1FSM::getInstance().getArbitrator().
      submitAction((*it).first, m_loadUnit, Arbitration::ActionImportance(*this));
  }
}

///////////////////////////////////////////////////
// SendToTarget state
///////////////////////////////////////////////////

ReaverShuttleGroup::SendToTarget::SendToTarget(ReaverShuttleGroup& behavior)
  : BehaviorState(behavior)
  , m_moveAction(*this)
  , m_event(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToAttackPhaseState(*this,
                                   GroupReachedDestination<UnitGroup>(behavior.m_shuttles,behavior.m_enemyBase), 
                                   NoTransitionAction(),
                                   behavior.m_attackPhaseState)
  , m_events(m_event)
  , m_transitions(m_transitionToAttackPhaseState)
{
}

void ReaverShuttleGroup::SendToTarget::onEnterImpl()
{
  for(std::map<BWAPI::Unit*,BWAPI::Unit*>::const_iterator it = m_context.m_pairs.begin();it != m_context.m_pairs.end();it++){
    m_moveAction.setDestination(m_context.m_enemyBase);
    Layer1FSM::getInstance().getArbitrator().
      submitAction((*it).first, m_moveAction, Arbitration::ActionImportance(*this));
  }
}

///////////////////////////////////////////////////
// AttackPhase state
///////////////////////////////////////////////////

ReaverShuttleGroup::AttackPhase::AttackPhase(ReaverShuttleGroup& behavior)
  : BehaviorState(behavior)
  , m_event(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToRemoveUnitState(*this,
                                  PairBroken<UnitGroup>(behavior.m_reavers), 
                                  NoTransitionAction(),
                                  behavior.m_removeUnitState)
  , m_events(m_event)
  , m_transitions(m_transitionToRemoveUnitState)
{
}

void ReaverShuttleGroup::AttackPhase::onEnterImpl()
{
  //spawn ATTACK!~
}

///////////////////////////////////////////////////
// RemoveUnit state
///////////////////////////////////////////////////

ReaverShuttleGroup::RemoveUnit::RemoveUnit(ReaverShuttleGroup& behavior)
  : BehaviorState(behavior)
  , m_event(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToSelf(*this,
                       PairBroken<UnitGroup>(behavior.m_reavers), 
                       NoTransitionAction(),
                       *this)
 , m_events(m_event)
 , m_transitions(m_transitionToSelf)
{
}

void ReaverShuttleGroup::RemoveUnit::onEnterImpl()
{
  //Remove pair
}