#pragma once
#include "../../GroupBehavior.h"
#include "../../BehaviorState.h"
#include "../../Arbitrator.h"
#include "../../../../../PersistentUnitGroup.h"
#include "../../Conditions/AllPickedUp.h"
#include "../../Conditions/ReachedDestination.h"

template <class UnitGroup>
class PairBroken : public Condition
{
public:
  PairBroken(UnitGroup* units)
    : m_units(units)
  {
  }

  bool evaluate() const
  {
    return false; //TODO
  }

protected:
  UnitGroup* m_units;

};

/**
* @brief ReaverShuttleGroup is a reaver/shuttle group behavior
*/
class ReaverShuttleGroup : public GroupBehavior<ReaverShuttleGroup>
{
protected:
  class SetPairs;
  class PickUp;
  class SendToTarget;
  class AttackPhase;
  class RemoveUnit;
  friend class SetPairs;
  friend class PickUp;
  friend class SendToTarget;
  friend class AttackPhase;
  friend class RemoveUnit;

  class SetPairs : public BehaviorState<ReaverShuttleGroup, SetPairs>
  {
  public:
    SetPairs(ReaverShuttleGroup& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_transitionToPickUpState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(SetPairs, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<SetPairs, NoCondition, NoTransitionAction, PickUp>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions

    // Transitions
    TimerEvent m_event;
    Transition<SetPairs, NoCondition, NoTransitionAction, PickUp> m_transitionToPickUpState;
  };

  class PickUp : public BehaviorState<ReaverShuttleGroup, PickUp>
  {
  public:
    PickUp(ReaverShuttleGroup& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_transitionToSendToTargetState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(PickUp, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<PickUp, AllPickedUp<UnitGroup>, NoTransitionAction, SendToTarget>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions
    layer1::LoadUnit m_loadUnit;

    // Transitions
    TimerEvent m_event;
    Transition<PickUp, AllPickedUp<UnitGroup>, NoTransitionAction, SendToTarget> m_transitionToSendToTargetState;
  };

  class SendToTarget : public BehaviorState<ReaverShuttleGroup, SendToTarget>
  {
  public:
    SendToTarget(ReaverShuttleGroup& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_transitionToAttackPhaseState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(SendToTarget, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<SendToTarget, GroupReachedDestination<UnitGroup>, NoTransitionAction, AttackPhase>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions
    layer1::MoveAction m_moveAction;

    // Transitions
    TimerEvent m_event;
    Transition<SendToTarget, GroupReachedDestination<UnitGroup>, NoTransitionAction, AttackPhase> m_transitionToAttackPhaseState;
  };

  class AttackPhase : public BehaviorState<ReaverShuttleGroup, AttackPhase>
  {
  public:
    AttackPhase(ReaverShuttleGroup& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_transitionToRemoveUnitState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(AttackPhase, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<AttackPhase, PairBroken<UnitGroup>, NoTransitionAction, RemoveUnit>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions

    // Transitions
    TimerEvent m_event;
    Transition<AttackPhase, PairBroken<UnitGroup>, NoTransitionAction, RemoveUnit> m_transitionToRemoveUnitState;
  };

  class RemoveUnit : public BehaviorState<ReaverShuttleGroup, RemoveUnit>
  {
  public:
    RemoveUnit(ReaverShuttleGroup& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onTimer()
    {
      m_transitionToSelf.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(RemoveUnit, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<RemoveUnit, PairBroken<UnitGroup>, NoTransitionAction, RemoveUnit>&
    > Transitions;
		const Transitions m_transitions; 

  protected:
    // Variables

    // Actions

    // Transitions
    TimerEvent m_event;
    Transition<RemoveUnit, PairBroken<UnitGroup>, NoTransitionAction, RemoveUnit> m_transitionToSelf; 
  };


public:
  ReaverShuttleGroup(const Process* parent, unsigned int initialPriority, const PersistentUnitGroup* reavers, const PersistentUnitGroup* shuttles, BWAPI::Position enemyBase)
    : GroupBehavior(parent, initialPriority, shuttles)
    , IGroupBehavior(parent, initialPriority, shuttles)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
    , m_setPairsState(*this), m_pickUpState(*this), m_sendToTargetState(*this)
    , m_attackPhaseState(*this), m_removeUnitState(*this)
    , m_enemyBase(enemyBase)
    , m_states(m_setPairsState, m_pickUpState, m_sendToTargetState, m_attackPhaseState, m_removeUnitState)
  {
    m_reavers = &a;
    m_shuttles = &b;

    for(PersistentUnitGroup::const_iterator it = reavers->begin();it!=reavers->end();it++){
      m_reavers->insert((*it));
    }
    for(PersistentUnitGroup::const_iterator it = shuttles->begin();it!=shuttles->end();it++){
      m_shuttles->insert((*it));
    }
    //m_reavers = reavers;
    //m_shuttles = shuttles;
    m_pairs = std::map<BWAPI::Unit*,BWAPI::Unit*>();
  }


  virtual std::string getName() const
  {
    return "ReverShuttleGroup";
  }

  virtual void onUnitAddedToGroup(void* data, BWAPI::Unit* unit)
  {
    // TODO
  }

  virtual void onUnitRemovedFromGroup(void* data, BWAPI::Unit* unit)
  {
    // TODO
  }

protected:
  virtual void executeImpl()
  {
    GroupBehavior::executeImpl();
    m_setPairsState.onEnter();
  }

  // Arguments

  // Variables
  BWAPI::Position m_enemyBase;
  std::map<BWAPI::Unit*,BWAPI::Unit*> m_pairs;
public:
  UnitGroup* m_reavers;
  UnitGroup* m_shuttles;
  UnitGroup a;
  UnitGroup b;

  typedef boost::tuple<
      SetPairs&,
      PickUp&,
      SendToTarget&,
      AttackPhase&,
      RemoveUnit&
    > States;
	const States m_states;

protected:
  // States
  SetPairs m_setPairsState;
  PickUp m_pickUpState;
  SendToTarget m_sendToTargetState;
  AttackPhase m_attackPhaseState;
  RemoveUnit m_removeUnitState;
};