#include "stdafx.h"
#include "ShuttleControl.h"
#include "../../Layer1FSM.h"

///////////////////////////////////////////////////
// Waiting state

ShuttleControl::Waiting::Waiting(ShuttleControl& behavior)
  : BehaviorState(behavior)
  , m_move(*this)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 2, 2)
  , m_transitionToPickUpStateDanger(
      *this,
      InDanger(behavior.m_reaver), 
      NoTransitionAction(), 
      behavior.m_pickUpState)
  , m_transitionToPickUpStateTarget(
      *this,
      NoTarget(behavior.m_reaver), 
      NoTransitionAction(), 
      behavior.m_pickUpState)
  , m_transitionToSelf(
      *this,
      IsDecelerating(behavior.m_unit,behavior.m_reaver), 
      NoTransitionAction(), 
      *this)
  , m_events(m_timerEvent)
  , m_transitions(m_transitionToPickUpStateDanger,m_transitionToPickUpStateTarget,m_transitionToSelf)
{
}

void ShuttleControl::Waiting::onEnterImpl()
{  
  int x = m_context.m_unit->getPosition().x() - m_context.m_reaver->getPosition().x();
  int y = m_context.m_unit->getPosition().y() - m_context.m_reaver->getPosition().y();
  if(x > 0)
    x = m_context.m_reaver->getPosition().x()-200;
  else
    x = m_context.m_reaver->getPosition().x()+200;
  if(y > 0)
    y = m_context.m_reaver->getPosition().y()-200;
  else
    y = m_context.m_reaver->getPosition().y()+200;
  BWAPI::Position pos(x,y);
  m_move.setDestination(pos);
  Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_move, Arbitration::ActionImportance(*this));
}

///////////////////////////////////////////////////
// Drop state

ShuttleControl::Drop::Drop(ShuttleControl& behavior)
  : BehaviorState(behavior)
  , m_unloadAllUnitsAtPosition(*this)
  , m_unitIsUnloaded(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 1, 1)
  , m_transitionToWaitingState(
      *this,
      UnitOnGround(behavior.m_reaver), 
      NoTransitionAction(), 
      behavior.m_getDistanceState)
  , m_events(m_unitIsUnloaded)
  , m_transitions(m_transitionToWaitingState)
{
}

void ShuttleControl::Drop::onEnterImpl()
{
  //BWAPI::Position position = Spar->getPerceptualState().getBorder(m_context.m_unit->getPosition());
  m_unloadAllUnitsAtPosition.setPosition(Spar->getPerceptualState().getBorder(m_context.m_unit->getPosition()));
  Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_unloadAllUnitsAtPosition, Arbitration::ActionImportance(*this));
}

///////////////////////////////////////////////////
// PickUp state

ShuttleControl::PickUp::PickUp(ShuttleControl& behavior)
  : BehaviorState(behavior)
  , m_loadUnit(*this)
  , m_unitIsLoaded(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 1, 1)
  , m_transitionToDropState(
      *this,
      UnitLoaded(behavior.m_reaver), 
      NoTransitionAction(), 
      behavior.m_dropState)
  , m_events(m_unitIsLoaded)
  , m_transitions(m_transitionToDropState)
{
}

void ShuttleControl::PickUp::onEnterImpl()
{
	m_loadUnit.setUnit(m_context.m_reaver);
	Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_loadUnit, Arbitration::ActionImportance(*this));
}

///////////////////////////////////////////////////
// ShieldRecharge state

ShuttleControl::GetDistance::GetDistance(ShuttleControl& behavior)
  : BehaviorState(behavior)
  , m_move(*this)
  , m_moveActionGiven(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 1, 1)
  , m_transitionToWaitingState(
      *this,
      NoCondition(), 
      NoTransitionAction(), 
      behavior.m_waitingState)
  , m_events(m_moveActionGiven)
  , m_transitions(m_transitionToWaitingState)
{
}

void ShuttleControl::GetDistance::onEnterImpl()
{
  int x = m_context.m_reaver->getPosition().x() + 150;
  int y = m_context.m_reaver->getPosition().y() + 150;
  BWAPI::Position pos(x,y);
  m_move.setDestination(pos);
  Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_move, Arbitration::ActionImportance(*this));
}
