#include "stdafx.h"
#include "Scout.h"
#include "../../../../Utils/Logger.h"
#include "../Layer1FSM.h"

///////////////////////////////////////////////////
// GoToLocation State
///////////////////////////////////////////////////
ScoutFSM::GoToLocation::GoToLocation(ScoutFSM& behavior, Location* location)
: BehaviorState(behavior)
, m_moveAction(*this)
, m_location(location)
, m_locationReached(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
, m_transitionToScouting(*this,
                     UnitReachedLocation(behavior.getUnit(), m_location), 
                     //NoTransitionAction(), 
                     RaiseScoutArrived(&getBehavior()),
                     behavior.m_scoutingState)
, m_events(m_locationReached)
, m_transitions(m_transitionToScouting)
{}

void ScoutFSM::GoToLocation::onEnterImpl()
{
  //Set destination
  m_moveAction.setDestination(m_location->getCurrentCenter()); 
  //Give command
  Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_moveAction, Arbitration::ActionImportance(*this));
}




///////////////////////////////////////////////////
// Scouting state
///////////////////////////////////////////////////
ScoutFSM::Scouting::Scouting(ScoutFSM& behavior, Location* location)
: BehaviorState(behavior)
, m_moveAction(*this)
, m_location(location)
, m_newMove(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 10, 10)
, m_transitionToSelf(*this,
                     NoCondition(), 
                     NoTransitionAction(), 
                     *this)
, m_events(m_newMove)
, m_transitions(m_transitionToSelf)
{
  m_maxY = BWAPI::Broodwar->mapHeight() * TILE_SIZE;
  m_maxX = BWAPI::Broodwar->mapWidth() * TILE_SIZE;
  m_region = BWTA::getRegion(location->getCurrentCenter().x()/TILE_SIZE, location->getCurrentCenter().y()/TILE_SIZE);
}

//TODO: Fix, ca crash!!  (tester avec le scenario "scoutScenario")
void ScoutFSM::Scouting::onEnterImpl()
{
  //Get next destination
  BWAPI::Position destination(-m_location->getCurrentCenter().x(),m_location->getCurrentCenter().y()) ;
  //Get potential menace
  std::list<BWAPI::Unit*> danger;
  UnitGroup enemyGroup = Spar->getPerceptualState().getEnemiesNearUnit(m_context.m_unit,BWAPI::WeaponTypes::Arclite_Cannon.maxRange());
  for (UnitGroup::const_iterator it = enemyGroup.begin(); it != enemyGroup.end(); ++it)
  { 
    //if((*it)->getType().canAttack()){
    if(m_context.m_unit->getType().isFlyer()){
      if((*it)->getType().airWeapon() != BWAPI::WeaponTypes::None){
        if(m_context.m_unit->getDistance((*it)) < (*it)->getType().airWeapon().maxRange() + 64){
          danger.push_back((*it));
        }
      }
    }
    else{
      if((*it)->getType().groundWeapon() != BWAPI::WeaponTypes::None && !(*it)->getType().isWorker()){
        if(m_context.m_unit->getDistance((*it)) < (*it)->getType().groundWeapon().maxRange() + 64){
          danger.push_back((*it));
        }
      }	  
    }
  }
  //}
  //If the menace exist move away from it
  if(!danger.empty())
  {
    destination = Spar->getPerceptualState().getSafeRetreatPosition(m_context.m_unit);
    /*double fleeX = 0;
    double fleeY = 0;
    std::list<BWAPI::Unit*>::iterator it = danger.begin();
    double tmpX = (*it)->getPosition().x() - m_context.m_unit->getPosition().x();
    double tmpY = (*it)->getPosition().y() - m_context.m_unit->getPosition().y();
    double lenght = sqrt(tmpX*tmpX + tmpY*tmpY);
    fleeX = tmpX/lenght * 50;
    fleeY = tmpY/lenght * 50;
    it++;
    while(it != danger.end()){
      //it++;
      tmpX = (*it)->getPosition().x() - m_context.m_unit->getPosition().x();
      tmpY = (*it)->getPosition().y() - m_context.m_unit->getPosition().y();
      lenght = sqrt(tmpX*tmpX + tmpY*tmpY);
      fleeX += tmpX/lenght * 100;
      fleeY += tmpY/lenght * 100;
      it++;
    }
    lenght = sqrt(fleeX*fleeX + fleeY*fleeY);
    fleeX = fleeX/lenght * 100;
    fleeY = fleeY/lenght * 100;
    //SPAR_LOG(LogTypes::LAYER_1_DECISION, "Flee destination = (%f, %f)", m_behavior.m_unit->getPosition().x() -fleeX, m_behavior.m_unit->getPosition().y() -fleeY); 
    //destination = m_context.m_unit->getPosition() + BWAPI::Position(-fleeX,-fleeY);
    destination = getFleeDestination(-fleeX, -fleeY);
    if(!Spar->getPerceptualState().positionIsReachable(m_context.m_unit->getPosition(), destination)){
      destination = getFleeDestination(fleeY,-fleeX);
      if(!Spar->getPerceptualState().positionIsReachable(m_context.m_unit->getPosition(), destination)){
        destination = getFleeDestination(-fleeY,fleeX);
        if(!Spar->getPerceptualState().positionIsReachable(m_context.m_unit->getPosition(), destination)){
          BWAPI::Broodwar->printf("scout says: FFFFFFFFFFFFFFFUUUUUUUUUUUUUUUUU");
        }
      }  
    }*/
  }else
  {
   /* double tmpX = m_context.m_enemyCC.x() - m_context.m_unit->getPosition().x();
    double tmpY = m_context.m_enemyCC.y() - m_context.m_unit->getPosition().y();
    double lenght = sqrt(tmpX*tmpX + tmpY*tmpY);
    BWAPI::Position pos(-tmpY/lenght * 50,tmpX/lenght * 50);
    destination = m_context.m_unit->getPosition() + pos;
    if(!destination.isValid()){
      destination = m_context.m_unit->getPosition() + BWAPI::Position(tmpY/lenght * 50,-tmpX/lenght * 50);
    }*/
    destination = Spar->getPerceptualState().getScoutingLocation(m_context.m_unit, m_region);
  }

 // SPAR_LOG(LogTypes::LAYER_1_DECISION, "Scout destination = (%i, %i)", destination.x(), destination.y()); 

  //Set destination
  m_moveAction.setDestination(destination); 
  //Give command

  Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_moveAction, Arbitration::ActionImportance(*this));
}

//Add fleeX and fleeY to the scout's current position to get the destination.
//Note: checks to make sure the destination is not outside the map!
BWAPI::Position ScoutFSM::Scouting::getFleeDestination(double fleeX, double fleeY)
{
  double destX = m_context.m_unit->getPosition().x() + fleeX;
  if(destX < 0)
    destX = 0;
  else if(destX > m_maxX)
    destX = m_maxX;

  double destY = m_context.m_unit->getPosition().y() + fleeY;
  if(destY < 0)
    destY = 0;
  else if(destY > m_maxY)
    destY = m_maxY;

  return BWAPI::Position(static_cast<int>(destX), static_cast<int>(destY));
}