#pragma once
#include "../../../../Utils/FSM/Condition.h"

const double MAX_GROUP_DISTANCE_TO_DESTINATION = 128;

const double MAX_UNIT_DISTANCE_TO_DESTINATION = 32;

template<class UnitGroup>
class GroupReachedDestination : public Condition
{
public:
	GroupReachedDestination(const UnitGroup* unit,BWAPI::Position destination)
    : m_unit(unit),m_destination(destination)
  {
  }

  bool evaluate() const
  {
    for(UnitGroup::const_iterator it = m_unit->begin();
        it != m_unit->end();
        ++it)
    {
      if((*it)->getDistance(m_destination) > MAX_GROUP_DISTANCE_TO_DESTINATION)
      {
        return false;
	    }
    }
    return true;
  }

protected:
  const UnitGroup* const m_unit;
  const BWAPI::Position m_destination;
};

class UnitReachedDestination : public Condition
{
public:
  UnitReachedDestination(BWAPI::Unit* unit, BWAPI::Position destination)
    : m_unit(unit), m_destination(destination)
  {
    maxDist = MAX_UNIT_DISTANCE_TO_DESTINATION;
  }

  UnitReachedDestination(BWAPI::Unit* unit, BWAPI::Position destination, double md)
    : m_unit(unit), m_destination(destination)
  {
    maxDist = md;
  }

  bool evaluate() const
  {
	  return m_unit->getPosition().getApproxDistance(m_destination) <= maxDist;
  }

protected:
  const BWAPI::Unit* const  m_unit;
  const BWAPI::Position m_destination;
  double maxDist;
};

class UnitReachedRegion : public Condition
{
public:
  UnitReachedRegion(BWAPI::Unit* unit, BWTA::Region* destination)
    : m_unit(unit), m_destination(destination)
  {}

  bool evaluate(boost::tuple<>&) const
  {
    return evaluate();
  }

  bool evaluate() const
  {
    return m_destination->getPolygon().isInside(m_unit->getPosition());
  }

protected:
  const BWAPI::Unit* const  m_unit;
  const BWTA::Region* const m_destination;
};

class UnitReachedLocation : public Condition
{
public:
  UnitReachedLocation(BWAPI::Unit* unit, Location* destination)
    : m_unit(unit), m_destination(destination)
  {}

  bool evaluate(boost::tuple<>&) const
  {
    return evaluate();
  }

  bool evaluate() const
  {
    return m_destination->isInLocation(m_unit->getPosition());
  }

protected:
  const BWAPI::Unit* const  m_unit;
  const Location* const m_destination;
};

template<class UnitGroup>
class GroupReachedDestinationLocation : public Condition
{
public:
	GroupReachedDestinationLocation(const UnitGroup* unit,Location* destination)
    : m_unit(unit),m_destination(destination)
  {
  }

  bool evaluate() const
  {
    for(UnitGroup::const_iterator it = m_unit->begin();
        it != m_unit->end();
        ++it)
    {
      if((*it)->getDistance(m_destination->getCurrentCenter()) > MAX_GROUP_DISTANCE_TO_DESTINATION)
      {
        return false;
	    }
    }
    return true;
  }

protected:
  const UnitGroup* const m_unit;
  const Location* m_destination;
};