#pragma once
#include "Behavior.h"
#include "../../../Utils/GroupProcess.h"

/**
 * @brief A group behavior outputs commands to a group of units (possibly by being decomposed
 *        into multiple individual unit behaviors).
 */
class IGroupBehavior : public GroupProcess, public virtual IBehavior
{
public:
  IGroupBehavior(const Process* parent, unsigned int initialPriority, const PersistentUnitGroup* group)
    : GroupProcess(parent, group)
    , Process(parent)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
  {
  }

  virtual void executeImpl()
  {
    GroupProcess::executeImpl();
    IBehavior::executeImpl();
  }

  /////////
  // GUI //
  /////////
  virtual std::set<BWAPI::Unit*> getCommandedUnits() const
  {
    return std::set<BWAPI::Unit*>(getGroup()->begin(), getGroup()->end());
  }

protected:
  virtual void output(std::ostream& out) const
  {
    IBehavior::output(out);
  }

  virtual void terminateImpl()
  {
    GroupProcess::terminateImpl();
    IBehavior::terminateImpl();
  }
};

template <class DerivedBehavior>
class GroupBehavior : public virtual IGroupBehavior, public Behavior<DerivedBehavior>
{
public:
  GroupBehavior(const Process* parent, unsigned int initialPriority, const PersistentUnitGroup* group)
    : IGroupBehavior(parent, initialPriority, group)
    , Behavior(parent, initialPriority)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
  {
  }
};
