#pragma once
#include "../../../Utils/Component.h"
#include "ArbitratorUnitData.h"
#include "UnitBehavior.h"
#include <boost/iterator/transform_iterator.hpp>

/**
 * @brief This class holds controller-related data for each unit.
 */
class Layer1FSMUnitData : public ArbitratorUnitData, public Component
{
public:
  Layer1FSMUnitData(BWAPI::Unit* unit)
    : ArbitratorUnitData(unit)
    , UnitData(unit)
#ifdef USE_SPAR_GUI
    , m_behaviorTerminateEventHandler(*this)
#endif
  {
  }

#ifndef USE_SPAR_GUI

  void addBehavior(IUnitBehavior* behavior)
  {
    unused(behavior);
  }

#else

protected:
  struct BehaviorWrapper
  {
    BehaviorWrapper() {}
    BehaviorWrapper(IUnitBehavior* behavior)
      : m_behavior(behavior) {}

    IUnitBehavior* m_behavior;
    std::list<BehaviorWrapper>::iterator m_behaviorID;
    IUnitBehavior::OnTerminate::SubscriberID m_subscriberID;
  };
  struct extract_behavior : std::unary_function<BehaviorWrapper,const IUnitBehavior*>
  {
    const IUnitBehavior* operator()(const BehaviorWrapper& behaviorWrapper) const
    {
      return behaviorWrapper.m_behavior;
    }
  };

public:
  /**
   * @brief Attachs a new behavior to this unit.
   * @param behavior the new behavior.
   */
  void addBehavior(IUnitBehavior* behavior)
  {
    std::list<BehaviorWrapper>::iterator behaviorID = m_behaviors.insert(m_behaviors.end(), BehaviorWrapper(behavior));
    BehaviorWrapper& behaviorWrapper = *behaviorID;
    behaviorWrapper.m_behaviorID = behaviorID;
    behaviorWrapper.m_subscriberID = behavior->getTerminateEvent().subscribe(&m_behaviorTerminateEventHandler, &behaviorWrapper);
  }

  /**
   * @brief Notification that a behavior terminated.
   * @param data data that was sent along to the behavior when subscribing.
   */
  void onBehaviorTerminate(void* data)
  {
    BehaviorWrapper* behaviorWrapper = reinterpret_cast<BehaviorWrapper*>(data);
#ifdef _DEBUG_COMPONENT
    behaviorWrapper->m_behavior->getTerminateEvent().unsubscribe(behaviorWrapper->m_subscriberID);
#endif
    m_behaviors.erase(behaviorWrapper->m_behaviorID);
  }

  typedef boost::transform_iterator<extract_behavior, std::list<BehaviorWrapper>::const_iterator> behavior_iterator;
  behavior_iterator behaviorsBegin()
  {
    return behavior_iterator(m_behaviors.begin());
  }
  behavior_iterator behaviorsEnd()
  {
    return behavior_iterator(m_behaviors.end());
  }

protected:
  EVENT_HANDLER(Layer1FSMUnitData, onBehaviorTerminate) m_behaviorTerminateEventHandler;

  std::list<BehaviorWrapper> m_behaviors;
#endif
};