#include "stdafx.h"
#include "AvoidEnemies.h"
#include "../Layer1FSM.h"

///////////////////////////////////////////////////
// Idle state

AvoidEnemies::Idle::Idle(AvoidEnemies& behavior)
  : BehaviorState(behavior)
  , m_stopAction(*this)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToReadyState(
      *this,
      IsUnitIdle(behavior.m_unit), 
      NoTransitionAction(), 
      behavior.m_readyToAvoidState)
  , m_events(m_timerEvent)
  , m_transitions(m_transitionToReadyState)
{
}

void AvoidEnemies::Idle::onEnterImpl()
{
    Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.m_unit, m_stopAction, Arbitration::ActionImportance(*this));
}

///////////////////////////////////////////////////
// ReadyToAvoid state

AvoidEnemies::ReadyToAvoid::ReadyToAvoid(AvoidEnemies& behavior)
: BehaviorState(behavior)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToAvoidingState(
      *this,
      AtLeastOneUnit(), 
      NoTransitionAction(), 
      behavior.m_avoidingState)
  , m_events(m_timerEvent)
  , m_transitions(m_transitionToAvoidingState)
{
}

void AvoidEnemies::ReadyToAvoid::onEnterImpl()
{
}

///////////////////////////////////////////////////
// Avoiding state

AvoidEnemies::Avoiding::Avoiding(AvoidEnemies& behavior)
: BehaviorState(behavior)
  , m_moveAction(*this)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 5, 5)
  , m_transitionToSelf(
      *this,
      AtLeastOneUnit(),
      NoTransitionAction(), 
      *this)
  , m_transitionToIdle(
      *this,
      Not<AtLeastOneUnit>(AtLeastOneUnit()),
      NoTransitionAction(), 
      behavior.m_idleState)
  , m_events(m_timerEvent)
  , m_transitions(m_transitionToSelf, m_transitionToIdle)
{
}

void AvoidEnemies::Avoiding::onEnterImpl()
{  
  BWAPI::Position destination = Spar->getPerceptualState().
    getSafeRetreatPosition(m_context.m_unit); //, m_context.m_enemyUnits);

  m_moveAction.setDestination(destination);
  Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.m_unit, m_moveAction, Arbitration::ActionImportance(*this));
}

void AvoidEnemies::Avoiding::onTimer()
{
  m_context.m_enemyUnits = Spar->getPerceptualState().getEnemyUnits(CircularUnitDynamicLocation(m_context.m_unit));
  m_transitionToIdle.testTransition(boost::tuple<const UnitGroup&>(m_context.m_enemyUnits));
  m_transitionToSelf.testTransition(boost::tuple<const UnitGroup&>(m_context.m_enemyUnits));
}