#pragma once
#include "../UnitBehavior.h"
#include "../BehaviorState.h"
#include "../Arbitrator.h"
#include "../../../../PerceptualState/Locations/Dynamic/CircularUnitDynamicLocation.h"
#include "../Conditions/InDanger.h"
#include "../Conditions/HasCriticalHealth.h"
#include "../Conditions/ReachedDestination.h"
#include "../Conditions/AtLeastOneUnit.h"
#include "../Conditions/CompositeCondition.h"

class Dodge : public UnitBehavior<Dodge>
{
protected:
  class Safe;
  class Critical;
  friend class Safe;
  friend class Critical;

  class Safe : public BehaviorState<Dodge, Safe>
  {
  public:
    Safe(Dodge& behavior);
    void onEnterImpl();
    void onExitImpl();

    void onTimer()
    {
      m_transition.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Safe, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<Safe, HasCriticalHealth, NoTransitionAction, Critical>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions

    // Transitions
    TimerEvent m_timerEvent;
    Transition<Safe, HasCriticalHealth, NoTransitionAction, Critical> m_transition;
  };

  class Critical : public BehaviorState<Dodge, Critical>
  {
    typedef CompositeCondition<HasCriticalHealth,InDanger> ShouldBeDodging;

  public:
    Critical(Dodge& behavior);
    void onEnterImpl();
    void onExitImpl();

    void onTimer()
    {
      m_transitionToSelf.testTransition(boost::make_tuple());
      m_transitionToSelf.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Critical, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<Critical, InDanger, NoTransitionAction, Critical>&,
      Transition<Critical, Not<ShouldBeDodging>, NoTransitionAction, Safe>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions
    layer1::MoveAction m_moveAction;

    // Transitions
    TimerEvent m_timerEvent;
    Transition<Critical, InDanger, NoTransitionAction, Critical> m_transitionToSelf;
    Transition<Critical, Not<ShouldBeDodging>, NoTransitionAction, Safe> m_transitionToSafeState;
  };


public:
  Dodge(const Process* parent, unsigned int initialPriority, BWAPI::Unit* dodgingUnit)
    : UnitBehavior(parent, initialPriority, dodgingUnit)
    , IUnitBehavior(parent, initialPriority, dodgingUnit)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
    , m_safeState(*this)
    , m_criticalState(*this)
    , m_states(m_safeState, m_criticalState)
  {
  }

  virtual std::string getName() const
  {
    return "Dodge (" + getObjectDescription(m_unit) + ")";
  }

  typedef boost::tuple<
      Safe&,
      Critical&
    > States;
	const States m_states;

protected:
  virtual void executeImpl()
  {
    UnitBehavior::executeImpl();
    m_safeState.onEnter();
  }

  // Arguments

  // Variables

  // States
  Safe m_safeState;
  Critical m_criticalState;
};
