#include "stdafx.h"
#include "FocusFire.h"
#include "../../Layer1FSM.h"
#include <BWAPI.h>

///////////////////////////////////////////////////
// Engaging state
///////////////////////////////////////////////////

FocusFire::Engaging::Engaging(FocusFire& behavior)
  : BehaviorState(behavior)
  , m_unitDestroyedEventHandler(*this)
  , m_attackUnitAction(*this)
  , m_timerEvent(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 10, 10)
	, m_events(m_timerEvent)
  , m_transitions(/*m_transitionToSelf*/)
  /*, m_transitionToSelf(
      *this,
      m_enemyUnitsInLocationEvent,
      NoCondition(), 
      NoTransitionAction(),
      *this)*/
{
}

FocusFire::Engaging::~Engaging()
{
}

void FocusFire::Engaging::onEnterImpl()
{
  m_context.m_enemyUnits = Spar->getPerceptualState().getEnemyUnits(RectangularGroupDynamicLocation(m_context.getGroup()));
  for(PersistentUnitGroup::const_iterator it=m_context.getGroup()->begin(); it != m_context.getGroup()->end(); ++it)
  {
	  BWAPI::Unit* unit = *it;
	  Dodge* newBehavior = new Dodge(&m_context, m_context.getCurrentPriority() * 2 + 1,unit);
	  newBehavior->execute();
  }
}

void FocusFire::Engaging::onUnitDestroyed(void* data)
{
  //if (m_context.m_enemyUnits.size() != 0)
  //{
  //  findTargetAndFire();
  //}
  if (m_context.m_enemyUnits.size() == 0)
    m_context.terminate(Behavior::Success);
  else
    findTargetAndFire();
} 

void FocusFire::Engaging::onTimer()
{
  UnitGroup enemyUnits = Spar->getPerceptualState().getEnemyUnits(RectangularGroupDynamicLocation(m_context.getGroup()));

  assert(&enemyUnits == &m_context.m_enemyUnits);

  if (m_context.m_enemyUnits.size() == 0)
    m_context.terminate(Behavior::Success);
  else
    findTargetAndFire();
}

BWAPI::Unit* FocusFire::Engaging::getWeakest(const UnitGroup &group)
{
  double minHp = boost::numeric::bounds<double>::highest();
  BWAPI::Unit* target = NULL;
  for (UnitGroup::const_iterator it = group.begin(); it != group.end(); ++it)
  {
    BWAPI::Unit* unit = *it;
    const double hp = unit->getShields() + unit->getHitPoints();
    if (hp < minHp)
    {
      minHp = hp;
      target = unit;
    }
  }
  return target;
}

void FocusFire::Engaging::findTargetAndFire()
{
	static int iteration = 0;
	BWAPI::Position center = m_context.getGroup()->getCenter();

  BWAPI::Unit* nextTarget = getWeakest(m_context.m_enemyUnits);
	if(!nextTarget)
		nextTarget = m_context.m_enemyUnits.getNearest(center);

	Spar->getUnitManager().subscribeIndividualUnitDestroy(nextTarget, &m_unitDestroyedEventHandler,0);

	m_attackUnitAction.setUnitToAttack(nextTarget);

  const PersistentUnitGroup* group = m_context.getGroup();
  Layer1FSM::getInstance().getArbitrator().
      submitAction(m_context.getGroup(), m_attackUnitAction, Arbitration::ActionImportance(*this));
}

void FocusFire::Engaging::onExitImpl()
{
}
