#pragma once
#include "../../../Behaviors/EngageEnemies/EngageEnemies.h"
#include "../../../../../../PerceptualState/Locations/Dynamic/RectangularGroupDynamicLocation.h"
#include "../../../Conditions/AtLeastOneUnit.h"
#include "UnitCarefulAttack.h"
#include <functional>

// TODO: this behavior will cause units to run after a fleeing unit (and get hit in the process)

class GroupCarefulAttack : public EngageEnemies<GroupCarefulAttack>
{
protected:
  //class Engaging;
  //friend class Engaging;

  /**
  * @brief The Engaging state is active when the group is engaging the enemies.
  *        New sub-behaviors are spawned upon entering this state.
  */
  class Engaging : public BehaviorState<GroupCarefulAttack, Engaging>
  {
  public:
    Engaging(GroupCarefulAttack& behavior);
    void onEnterImpl();
    void onExitImpl() {}

    void onEnemyUnitDestroyed(void* data);

    typedef boost::tuple<
      //EnemyUnitsInLocationRequest&
    > Events;
		const Events m_events;

    typedef boost::tuple<
    > Transitions;
		const Transitions m_transitions;

  protected:
    void findTarget();
    void setTarget();
    class SetTarget : public std::unary_function<Process*,SafeListBase::FunctionReturn>
    {
    public:
      SetTarget(BWAPI::Unit* target)
        : m_target(target)
      {}
      SafeListBase::FunctionReturn operator()(Process* subBehavior) const
      {
        dynamic_cast<UnitCarefulAttack*>(subBehavior)->setTarget(m_target);
        return SafeListBase::FunctionReturn();
      }
    protected:
      BWAPI::Unit* const m_target;
    };

    // Variables
    EVENT_HANDLER(Engaging, onEnemyUnitDestroyed) m_enemyUnitDestroyedEventHandler;

    // Actions

    // Transitions
  };

public:
  GroupCarefulAttack(const Process* parent, unsigned int initialPriority, const PersistentUnitGroup* group)
    : EngageEnemies(parent, initialPriority, group)
    , IEngageEnemies(parent, initialPriority, group)
    , IGroupBehavior(parent, initialPriority, group)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
    , m_engagingState(*this)
    , m_target(NULL)
    , m_states(m_engagingState)
  {
  }

  virtual std::string getName() const
  {
    return "GroupCarefulAttack";
  }

  virtual void onUnitAddedToGroup(void* data, BWAPI::Unit* unit)
  {
    // TODO
  }

  virtual void onUnitRemovedFromGroup(void* data, BWAPI::Unit* unit)
  {
    // TODO
  }

  typedef boost::tuple<
      Engaging&
    > States;
	const States m_states;

protected:
  virtual void executeImpl()
  {
    EngageEnemies::executeImpl();
    m_engagingState.onEnter();
  }

  // Arguments

  // Variables
  UnitGroup m_enemyUnits;
  BWAPI::Unit *m_target;

  // States
  Engaging m_engagingState;
};
