#include "stdafx.h"
#include "SplitGroupCarefulAttack.h"
#include "GroupCarefulAttack.h"
#include "../MeleeEngage.h"
#include "../../../../../../Utils/Vector2.h"
#include <BWAPI.h>
#include <algorithm>
#include <vector>

static const int MELEE_RANGE = 25; //Zealot's range has been verified to be 15.

/**
* @brief Divides a UnitGroup into a std::vector of std::vector of units. All units of a type are in a
* same std::vector.
*/
std::vector<std::vector<BWAPI::Unit *>> splitInTypes(const UnitGroup &group) {
  std::vector<std::vector<BWAPI::Unit *>> groupedUnits;
	for (UnitGroup::const_iterator i = group.begin(); i != group.end(); i++) {
		bool found = false;
		for (std::vector<std::vector<BWAPI::Unit *>>::iterator j = groupedUnits.begin(); j != groupedUnits.end(); ++j) {
			if ((*i)->getType() == (*j->begin())->getType()) {
				found = true;
				j->push_back(*i);
				break;
			}
		}
		
		if (!found) {
			groupedUnits.push_back(std::vector<BWAPI::Unit *>());
			groupedUnits[groupedUnits.size()-1].push_back(*i);
		}
	}
  return groupedUnits;
}

struct Functor_SortByDistanceToPos {
	BWAPI::Position m_pos;
	Functor_SortByDistanceToPos(BWAPI::Position pos): m_pos(pos) {}
	static int squaredDistance(BWAPI::Position pos1, BWAPI::Position pos2) {
		return (int)(pow((float)pos1.x()-pos2.x(), 2) + pow((float)pos1.y()-pos2.y(), 2));
	}

	bool operator() (BWAPI::Unit *u1, BWAPI::Unit *u2) {
		return squaredDistance(m_pos, u1->getPosition()) < squaredDistance(m_pos, u2->getPosition());
	}
};

/**
@brief Divide a group of unit to face enemies. Makes groups of same type units
		that are in a perpendicular line to the position from the middle of the group.
@param myUnits units to divide into a group
@param enemyPos estimated enemy position
@param result result of the division in groups
*/
std::vector<UnitGroup> divideInGroups(const UnitGroup &myUnits, BWAPI::Position enemyPos) {
  std::vector<UnitGroup> result;
	std::vector<std::vector<BWAPI::Unit *>> myUnitsByType = splitInTypes(myUnits);
	
	BWAPI::Position myPos = myUnits.getCenter();
	
	Vector2i myPosToEnemies = Vector2i(enemyPos.x() - myPos.x(), enemyPos.y() - myPos.y());
		
	for (std::vector<std::vector<BWAPI::Unit *>>::iterator i = myUnitsByType.begin(); i != myUnitsByType.end(); ++i) {
		std::vector<BWAPI::Unit *> &units = *i;
		BWAPI::UnitType unitType = units[0]->getType();
		size_t nbPerGroup;
		if (unitType.size() == BWAPI::UnitSizeTypes::Large) {
			nbPerGroup = 3;
		} else if (unitType.size() == BWAPI::UnitSizeTypes::Medium) {
			nbPerGroup = 5;
		} else {
			nbPerGroup = 10;
		}

		if (units.size() > nbPerGroup) {
			sort(units.begin(), units.end(), Functor_SortByDistanceToPos(enemyPos));
			size_t curBeginIndex = 0;
			size_t curEndIndex = nbPerGroup;
			while (true) {
				if (curBeginIndex >= units.size()) {
					break;
				}
				if (curEndIndex > units.size()) {
					curEndIndex = units.size();
				}
				result.push_back(UnitGroup());
				result[result.size()-1].insert(units.begin()+curBeginIndex, units.begin()+curEndIndex);
				curBeginIndex += nbPerGroup;
				curEndIndex += nbPerGroup;
			}			
		} else {
			result.push_back(UnitGroup());
			result[result.size()-1].insert(units.begin(), units.end());
		}
	}
  return result;
}

/**
@brief Divide a group of unit to face enemies. Makes groups of same type units
		that are in a perpendicular line to the enemies.
@param myUnits units to divide into a group
@param enemyUnits enemy units against which the groups should be fighting.
@param result result of the division in groups
*/
std::vector<UnitGroup> divideInGroups(const UnitGroup &myUnits, const UnitGroup &enemyUnits) {
	//int sumx = 0;
	//int sumy = 0;

	// Use the middle position of the enemy group. Thats all we do with them for now.
	return divideInGroups(myUnits, enemyUnits.getCenter());
}


///////////////////////////////////////////////////
// Engaging state
///////////////////////////////////////////////////

SplitGroupCarefulAttack::Engaging::Engaging(SplitGroupCarefulAttack& behavior, const UnitGroup& enemyUnits)
  : BehaviorState(behavior)
  , m_enemyCenter(enemyUnits.getCenter())
  , m_groupBehavior(*this)
{
}

void SplitGroupCarefulAttack::Engaging::onEnterImpl()
{
  UnitGroup unitGroup = UnitGroup();
  unitGroup.insert(m_context.getGroup()->begin(), m_context.getGroup()->end());

  std::vector<UnitGroup> groups = divideInGroups(unitGroup, m_enemyCenter);

  for (std::vector<UnitGroup>::iterator curGroup = groups.begin(); curGroup != groups.end(); ++curGroup)
  {
	  BWAPI::UnitType unitType = (*curGroup->begin())->getType();

    PersistentUnitGroup* newGroup = new PersistentUnitGroup(std::string("Sub-group type '" + unitType.getName() + "'"), curGroup->begin(), curGroup->end());
		
	  if (unitType.groundWeapon().maxRange() <= MELEE_RANGE) {
      m_groupBehavior = new MeleeEngage(&m_context, m_context.getCurrentPriority(), newGroup);
	  } else {
      m_groupBehavior = new GroupCarefulAttack(&m_context, m_context.getCurrentPriority() * 2 + 1, newGroup);
	  }
    m_groupBehavior->execute();
  }

  //return std::vector; //new FocusFire(initialPriority, group);
}

void SplitGroupCarefulAttack::Engaging::onExitImpl()
{
}

