#include "stdafx.h"
#include "UnitCarefulAttack.h"
#include "../../../Layer1FSM.h"

///////////////////////////////////////////////////
// Unit in moving State

UnitCarefulAttack::Moving::Moving(UnitCarefulAttack& behavior)
  : BehaviorState(behavior)  
  , m_update(*this, Scheduler::FSM_SCHEDULING_PRIORITY,10,10)
  , m_transition(
	  *this,
	  NotNull(&behavior.m_target),
	  NoTransitionAction(),
	  behavior.m_safeState)
  , m_events(m_update)
  , m_transitions(m_transition)
{
}

void UnitCarefulAttack::Moving::onEnterImpl()
{
}

void UnitCarefulAttack::Moving::onExitImpl()
{
}

///////////////////////////////////////////////////
// Unit in safe State

UnitCarefulAttack::Safe::Safe(UnitCarefulAttack& behavior)
  : BehaviorState(behavior)  
  , m_update(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 10, 10)
  , m_attackUnitAction(*this)
  , m_transition(
	  *this,
	  ShouldBeDodging(HasCriticalHealth(behavior.m_unit), InDanger(behavior.m_unit)),
	  NoTransitionAction(),
	  behavior.m_criticalState)
  , m_events(m_update)
  , m_transitions(m_transition)
{
}

void UnitCarefulAttack::Safe::onEnterImpl()
{
	orderAttack();
}

void UnitCarefulAttack::Safe::orderAttack()
{
	m_attackUnitAction.setUnitToAttack(m_context.m_target);

  Layer1FSM::getInstance().getArbitrator().
    submitAction(m_context.m_unit, m_attackUnitAction, Arbitration::ActionImportance(*this));
}

void UnitCarefulAttack::Safe::onExitImpl()
{
}

///////////////////////////////////////////////////
// Unit in critical State

UnitCarefulAttack::Critical::Critical(UnitCarefulAttack& behavior)
  : BehaviorState(behavior)
  , m_moveAction(*this)
  //, m_location(m_context.m_unit)
  //, m_enemyUnitsInLocationRequest(m_location, behavior.m_group)
  //, m_enemyUnitsInLocationEvent(m_enemyUnitsInLocationRequest, Scheduler::FSM_SCHEDULING_PRIORITY, 12, 12)
  , m_update(*this, Scheduler::FSM_SCHEDULING_PRIORITY, 12, 12)
  , m_transitionToSafeState(
      *this,
      NoCondition(),
      NoTransitionAction(), 
      behavior.m_safeState)
  , m_events(m_update)
  , m_transitions(m_transitionToSafeState)
{
}

void UnitCarefulAttack::Critical::onEnterImpl()
{
  //log("dodging");
  if(m_context.m_group->size() > 1)
  {
	  BWAPI::Position destination = Spar->getPerceptualState().getSafeDodgePosition(m_context.m_unit);
	  m_moveAction.setDestination(destination);
	  //m_moveAction.subscribeOnActionEnded(, 0);
	  Layer1FSM::getInstance().getArbitrator().submitAction(m_context.m_unit, m_moveAction, Arbitration::ActionImportance(*this));
  }
}

void UnitCarefulAttack::Critical::onExitImpl()
{
}