#pragma once
#include "../../../UnitBehavior.h"
#include "../../../BehaviorState.h"
#include "../../../Arbitrator.h"
#include "../../../../../../PerceptualState/Locations/Dynamic/CircularUnitDynamicLocation.h"
#include "../../../Conditions/InDanger.h"
#include "../../../Conditions/HasCriticalHealth.h"
#include "../../../Conditions/ReachedDestination.h"
#include "../../../Conditions/AtLeastOneUnit.h"
#include "../../../Conditions/CompositeCondition.h"

class NotNull : public Condition
{
public:
  NotNull(BWAPI::Unit** unit)
    : m_unit(unit)
  {}
  bool evaluate()
  {
	  return (*m_unit) != NULL;
  }

protected:
  BWAPI::Unit** m_unit;

};

class UnitCarefulAttack : public UnitBehavior<UnitCarefulAttack>
{
protected:
  class Safe;
  class Critical;
  class Moving;
  friend class Safe;
  friend class Critical;
  friend class Moving;

  class Moving : public BehaviorState<UnitCarefulAttack, Moving>
  {
  public:
    Moving(UnitCarefulAttack& behavior);
    void onEnterImpl();
    void onExitImpl();

    void onTimer()
    {
      m_transition.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Moving, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<Moving, NotNull, NoTransitionAction, Safe>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions

    // Transitions
    TimerEvent m_update;
    Transition<Moving, NotNull, NoTransitionAction, Safe> m_transition;
  };

  class Safe : public BehaviorState<UnitCarefulAttack, Safe>
  {
    typedef CompositeCondition<HasCriticalHealth,InDanger> ShouldBeDodging;

  public:
    Safe(UnitCarefulAttack& behavior);
    void onEnterImpl();
    void onExitImpl();	

    void onTimer()
    {
      m_transition.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Safe, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<Safe, ShouldBeDodging, NoTransitionAction, Critical>&
    > Transitions;
		const Transitions m_transitions;

    void orderAttack();

  protected:
    // Variables

    // Actions
    layer1::AttackUnitAction m_attackUnitAction;

    // Transitions
    TimerEvent m_update;
    Transition<Safe, ShouldBeDodging, NoTransitionAction, Critical> m_transition;
  };

  class Critical : public BehaviorState<UnitCarefulAttack, Critical>
  {
  public:
    Critical(UnitCarefulAttack& behavior);
    void onEnterImpl();
    void onExitImpl();

    void onTimer()
    {
      m_transitionToSafeState.testTransition(boost::make_tuple());
    }

    typedef SCHEDULER_EVENT(Critical, onTimer) TimerEvent;

    typedef boost::tuple<
      TimerEvent&
    > Events;
		const Events m_events;

    typedef boost::tuple<
      Transition<Critical, NoCondition, NoTransitionAction, Safe>&
    > Transitions;
		const Transitions m_transitions;

  protected:
    // Variables

    // Actions
    layer1::MoveAction m_moveAction;

    // Transitions
    TimerEvent m_update;
    Transition<Critical, NoCondition, NoTransitionAction, Safe> m_transitionToSafeState;
  };


public:
  UnitCarefulAttack(const Process* parent, unsigned int initialPriority, BWAPI::Unit* dodgingUnit, const PersistentUnitGroup* group)
    : UnitBehavior(parent, initialPriority, dodgingUnit)
    , IUnitBehavior(parent, initialPriority, dodgingUnit)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
    , m_movingState(*this)
    , m_safeState(*this)
    , m_criticalState(*this)
    , m_group(group, *this)
    , m_target(NULL)
    , m_states(m_movingState, m_safeState, m_criticalState)
  {
  }

  virtual std::string getName() const
  {
    return "UnitCarefulAttack";
  }

  void setTarget(BWAPI::Unit* target)
  {
    m_target = target;
    if(m_currentState == &m_safeState)
    {
      m_safeState.orderAttack();
    }
  }

  typedef boost::tuple<
      Moving&,
      Safe&,
      Critical&
    > States;
	const States m_states;

protected:
  virtual void executeImpl()
  {
    UnitBehavior::executeImpl();
    m_movingState.onEnter();
  }

  // Actions

  // Arguments

  // Variables
  //UnitGroup m_enemyUnitsCache;
  AutoPtr<const PersistentUnitGroup> m_group;
  BWAPI::Unit* m_target;

  // States
  Moving m_movingState;
  Safe m_safeState;
  Critical m_criticalState;
};
