#pragma once
#include "PatrolBase.h"

/**
 * @brief Transition action returning the next checkpoint, possibly past-the-end of the checkpoints list if the
 *        patrol has finished.
 */
class NextCheckpoint : public TransitionAction
{
public:
  NextCheckpoint(std::list<BWAPI::Position>::const_iterator& currentCheckpoint, const std::list<BWAPI::Position>&)
    : m_currentCheckpoint(currentCheckpoint){}
  void execute()
  {
    ++m_currentCheckpoint;
  }
protected:
  std::list<BWAPI::Position>::const_iterator& m_currentCheckpoint;
};

/**
 * @brief The FinitePatrol behavior instructs a group to patrol along a sequence of checkpoints, ending 
 *        the behavior when the loop is completed.
 */
class FinitePatrol : public PatrolBase<NextCheckpoint>
{
public:
  FinitePatrol(const Process* parent, unsigned int initialPriority, const PersistentUnitGroup* patrollingGroup, std::list<BWAPI::Position>& checkPoints)
    : PatrolBase<NextCheckpoint>(parent, initialPriority, patrollingGroup, checkPoints)
    , IGroupBehavior(parent, initialPriority, patrollingGroup)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
  {
  }

  virtual std::string getName() const
  {
    return "FinitePatrol (" + getObjectDescription(m_checkPoints) + ")";
  }
};
