#pragma once
#include "PatrolBase.h"

/**
 * @brief Transition action returning the next checkpoint, returning to the first one when all checkpoints have
 *        been visited.
 */
class LoopNextCheckpoint : public TransitionAction
{
public:
  LoopNextCheckpoint(std::list<Position>::const_iterator& currentCheckpoint, const std::list<Position>& checkPoints)
    : m_currentCheckpoint(currentCheckpoint), m_checkPoints(checkPoints) {}
  void execute()
  {
    if (++m_currentCheckpoint == m_checkPoints.end())
      m_currentCheckpoint = m_checkPoints.begin();
  }
protected:
  std::list<Position>::const_iterator& m_currentCheckpoint;
  const std::list<Position>& m_checkPoints;
};

/**
 * @brief The Patrol behavior instructs a group to patrol along a sequence of checkpoints,
 *        returning to the first one when the loop is completed.
 */
class Patrol : public PatrolBase<LoopNextCheckpoint>
{
public:
  Patrol(const Process* parent, unsigned int initialPriority, const PersistentUnitGroup* patrollingGroup, std::list<Position>& checkPoints)
    : PatrolBase<NextCheckpoint>(initialPriority, patrollingGroup, checkPoints)
  {
  }

  virtual std::string getName() const
  {
    return "Patrol (" + getObjectDescription(m_checkPoints) + ")";
  }
};