#pragma once
#include "Behavior.h"

/**
 * @brief A unit behavior outputs commands to a single unit.
 */
class IUnitBehavior : public virtual IBehavior
{
public:
  IUnitBehavior(const Process* parent, unsigned int initialPriority, BWAPI::Unit* unit)
    : IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
    , m_unitDestroyEventHandler(*this)
    , m_unit(unit)
  {
  }

  BWAPI::Unit* getUnit()
  {
    return m_unit;
  }

  void onUnitDestroy(void*)
  {
    terminate(UnitDestroyed);
  }

  static const TerminationMode UnitDestroyed;

  /////////
  // GUI //
  /////////
  virtual std::set<BWAPI::Unit*> getCommandedUnits() const
  {
    return std::set<BWAPI::Unit*>(&m_unit, &m_unit + 1);
  }

protected:
  virtual void executeImpl();

  virtual void terminateImpl()
  {
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unit, m_unitDestroyId);
    IBehavior::terminateImpl();
  }

  EVENT_HANDLER(IUnitBehavior, onUnitDestroy) m_unitDestroyEventHandler;
  UnitManager::OnIndividualUnitDestroy::SubscriberID m_unitDestroyId;

  /**
   * @brief The unit associated with this behavior.
   */
  BWAPI::Unit* const m_unit;
};

template <class DerivedBehavior>
class UnitBehavior : public virtual IUnitBehavior, public Behavior<DerivedBehavior>
{
public:
  UnitBehavior(const Process* parent, unsigned int initialPriority, BWAPI::Unit* unit)
    : IUnitBehavior(parent, initialPriority, unit)
    , Behavior(parent, initialPriority)
    , IBehavior(parent, initialPriority)
    , IFSM(parent)
    , Process(parent)
  {
  }
};
