#pragma once
#include "GroupTacticalAction.h"
#include <vector>

namespace layer2
{
  // Could restrict the choice of builder probes among some probes

  class Build : public virtual GroupTacticalAction
  {
  public:
    Build(unsigned int priority, StaticLocation* location, BWAPI::UnitType buildingType)
      : GroupTacticalAction(priority)
      , TacticalAction(priority)
      , Process(Spar)
      , m_location(location)
      , m_buildingType(buildingType)
    {
    }

    virtual std::string getName() const
    {
      return "Build";
    }

    static const TerminationMode TerminatedNoSpace;
    static const TerminationMode TerminatedNoWorker;
    static const TerminationMode TerminatedBuildingDestroyed;

    DECLARE_EVENT3(OnBuildingConstructed, const Build*, BWAPI::Unit*, PrimitiveLocation*);
    //DECLARE_EVENT3(OnBuildingCancelled, const Build*, BWAPI::Unit*, PrimitiveLocation*);
    //DECLARE_EVENT3(OnBuildingCancelledNoSpace, const Build*, BWAPI::UnitType, PrimitiveLocation*);
    OnBuildingConstructed& getBuildingConstructedEvent() const { return m_buildingConstructedEvent; }
    //OnBuildingCancelled& getBuildingCancelledEvent() const { return m_buildingCancelledEvent; }
    //OnBuildingCancelledNoSpace& getBuildingCancelledNoSpaceEvent() const { return m_buildingCancelledNoSpaceEvent; }

  protected:
    virtual void output(std::ostream& out) const
    {
      GroupTacticalAction::output(out);

      out << "Build(" << *m_location << ", " << m_buildingType << ")";
    }

    StaticLocation* const m_location;
    const BWAPI::UnitType m_buildingType;

    // TODO: move to layer 2 implementation?
    enum BuildingProgress
    {
      reserved,
      building,
      finished
    };
    std::map<BWAPI::TilePosition,BuildingProgress> m_buildingProgress;

    EVENT(OnBuildingConstructed) m_buildingConstructedEvent;
  };
}
