#pragma once
#include "../../../../Scheduler/SchedulerTask.h"
#include "TacticalAction.h"
#include <vector>

namespace layer2
{
  class Recruit : public virtual TacticalAction
  {
  public:
    Recruit(unsigned int priority, StaticLocation* location, const std::map<BWAPI::UnitType, size_t>& unitTypes, bool isInfinite = false)
      : TacticalAction(priority)
      , Process(Spar)
      , m_location(location)
      , m_units(unitTypes)
      , m_isInfinite(isInfinite)
    {}

    Recruit(unsigned int priority, StaticLocation* location, BWAPI::UnitType unitType, bool isInfinite = false)
      : TacticalAction(priority)
      , Process(Spar)
      , m_location(location)
      , m_units(getUnitTypes(unitType, 1))
      , m_isInfinite(isInfinite)
    {}

    virtual std::string getName() const
    {
      return "Recruit";
    }

    const StaticLocation* getLocation() const { return m_location; }
    StaticLocation* getLocation() { return m_location; }

    const std::map<BWAPI::UnitType, size_t>& getUnitsToRecruit() const { return m_units; }

    DECLARE_EVENT3(OnUnitRecruited, const Recruit*, BWAPI::Unit*, PrimitiveLocation*);
    OnUnitRecruited& getUnitRecruitedEvent() const { return m_unitRecruitedEvent; }

  protected:
    virtual void output(std::ostream& out) const
    {
      TacticalAction::output(out);

      out << (m_isInfinite?"Infinite-":"") << "Recruit(" << *m_location << ", ";
      
      for (std::map<BWAPI::UnitType, size_t>::const_iterator it = m_units.begin();
           it != m_units.end();
           ++it)
      {
        const std::pair<BWAPI::UnitType, size_t>& pair = *it;
        out << "[" << pair.first << ", " << pair.second << "]";
      }
      out << ")";
    }

    static std::map<BWAPI::UnitType, size_t> getUnitTypes(BWAPI::UnitType unitType, size_t nb)
    {
      std::map<BWAPI::UnitType, size_t> unitTypes;
      unitTypes.insert(std::make_pair(unitType, nb));
      return unitTypes;
    }

    StaticLocation* const m_location;
    const std::map<BWAPI::UnitType, size_t> m_units;
    const bool m_isInfinite;

    EVENT(OnUnitRecruited) m_unitRecruitedEvent;
  };
}