#pragma once
#include "../../../../PerceptualState/PerceptualState.h"
#include "../../../../PerceptualState/Locations/Static/StaticLocation.h"
#include "../../../../PerceptualState/Locations/Static/PrimitiveLocation.h"
#include "../../../../Utils/Event.h"
#include "../../../../Utils/Process.h"
#include "../../../../SparAIContainer.h"
#include "../../../../UnitManager.h"
#include "../ActionResources.h"

//#include "../../../../Planning/Exp/Bindings.h"
//#include "../../../../Planning/Exp/ILogicalExp.h"
//#include "../../../../Planning/Exp/Logical/AndExp.h"
//#include "../../../../Planning/Exp/IEffect.h"
//#include "../../../../Planning/World/IClosedWorld.h"

//#include "../../Layer3Tactical/Script/World.h" // Let's happily break encapsulation

namespace layer2
{
  /**
   * @brief Base class for tactical actions.
   */
  class TacticalAction : public virtual Process
  {
  public:
    /**
     * @brief Constructs a new tactical action.
     */
    TacticalAction(unsigned int priority)
      : Process(Spar)
      , m_id(s_nextId++)
      , m_priority(priority)
      , m_isCancelling(false)
    {
    }

    virtual ~TacticalAction()
    {
    }

    virtual std::string getName() const = 0;

    unsigned int getPriority() const
    {
      return m_priority;
    }
    
    void cancel()
    {
      if (!m_isCancelling && !isTerminated())
      {
        SPAR_LOG(LogTypes::LAYER_2_DECISION, "Cancelling action: %s", getObjectDescription(this).c_str());

        m_isCancelling = true;

        cancelSpecificImpl();
      }
    }

    // Failures are action-dependent
      
    static const TerminationMode Cancelled;

    static const TerminationMode AllUnitsDestroyed;
    
    friend std::ostream& operator<<(std::ostream& out, const TacticalAction& action)
    {
      action.output(out);
      return out;
    }

    /////////
    // GUI //
    /////////
    virtual std::set<BWAPI::Unit*> getCommandedUnits() const
    {
      return std::set<BWAPI::Unit*>();
    }

  protected:
    virtual void executeImpl()
    {
      SPAR_LOG(LogTypes::LAYER_2_DECISION, "Executing action: %s", getObjectDescription(this).c_str());

      executeCommonImpl();
      executeSpecificImpl();
    }

    // THIS METHOD SHOULD ONLY BE CALLED FROM executeImpl()
    virtual void executeCommonImpl()
    {
      Process::executeImpl();
    }

    // THIS METHOD SHOULD ONLY BE CALLED FROM executeImpl()
    virtual void executeSpecificImpl() = 0;
    
    virtual void terminateImpl()
    {
      terminateSpecificImpl();
      terminateCommonImpl();
    }

    // THIS METHOD SHOULD ONLY BE CALLED FROM terminateImpl()
    virtual void terminateCommonImpl()
    {
      Process::terminateImpl();
    }

    // THIS METHOD SHOULD ONLY BE CALLED FROM terminateImpl()
    virtual void terminateSpecificImpl() = 0;

    virtual void cancelSpecificImpl() = 0;

    virtual void output(std::ostream& out) const
    {
      out << "#" << m_id << " (P" << m_priority << ")- ";
    }

    unsigned int m_priority; // higher value -> more priority
    const unsigned int m_id;
    bool m_isCancelling;

  private:
    static unsigned int s_nextId;
  };
}