#include "stdafx.h"
#include "L2Build.h"
#include "../Layer2Managers.h"

void L2Build::onUnitAddedToGroup(void*, BWAPI::Unit* unit)
{
  unused(unit);
  // Should only happen when reserveUnit() is called
}

void L2Build::onUnitRemovedFromGroup(void*, BWAPI::Unit* unit)
{
  unused(unit);
  assert(m_worker == NULL || m_worker == unit);

  if(m_building != NULL)
  {
    //the building is started. Dont care
  }
  else
  {
    //must start the building
    m_worker = findClosestWorker(m_buildingTP);
    if(m_worker == NULL)
    {
      terminate(TerminatedNoWorker);
    }
    else
    {
      bool reserved = addUnit(m_worker);
      assert(reserved);
      if(m_buildFSM!=NULL)
        m_buildFSM->externalTerminate();
      m_buildFSM = new BuildFSM(this,m_priority,m_worker,m_buildingTP,m_buildingType);
      m_buildFSM->execute();
    }
    //destroy other FSM
    //Build new FSM with new worker
  }
}

void L2Build::executeSpecificImpl()
{
  L2GroupTacticalAction::executeSpecificImpl();
  Layer2Managers::getInstance().requestResources(
    m_priority,
    Resources(m_buildingType),
    m_resourcesRequestID);
}


void L2Build::cancelSpecificImpl()
{
  L2GroupTacticalAction::cancelSpecificImpl();
  Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
  if(m_worker!=NULL)
  {
    if(m_building!=NULL)
    {
      if(!m_building->isCompleted())
      {
        m_building->cancelConstruction();
      }
    }
    else
    {
      m_worker->stop();
    }
  }
  if(m_buildFSM != NULL)
  {
    //m_buildFSM.externalTerminate();
  }
  terminate(Cancelled);
}

void L2Build::terminateSpecificImpl()
{
  if(m_reservingResources)
  {
    //Layer2Managers::getInstance().releaseResources(Resources(m_buildingType));
    Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
    m_reservingResources = false;
  }
  else if(m_buildFSM != NULL)
  {
    m_buildFSM->externalTerminate();
  }
  else
  {
    if(m_building!=NULL)
      Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_building, m_unitDestroyedSubscriberId);
    Spar->getPerceptualState().getUnitCompletedEvent().unsubscribe(m_unitCompletedSubscriberId);
    Spar->getPerceptualState().getMap()->getOwnUnitAddedEvent().unsubscribe(m_unitCreatedSubscriberId);
  }
  L2GroupTacticalAction::terminateSpecificImpl();
}

void L2Build::onResourcesAllocated(void* data, BWAPI::Unit* b)
{
  m_reservingResources = true;

  //Layer2Managers::getInstance().executeBuild(this);
  const PersistentUnitGroup* group = getGroup();
  BWAPI::Unit* worker = NULL;
  std::set<BWAPI::Unit*> unitSet = BWAPI::Broodwar->self()->getUnits();
  for(std::set<BWAPI::Unit*>::const_iterator it = unitSet.begin() ; (it!=unitSet.end()) ; ++it)
  {
    if((*it)->getType() == BWAPI::UnitTypes::Protoss_Probe)
    {
      if((*it)->getOrder() != BWAPI::Orders::PlaceBuilding && !m_buildingPlacer->isUsed(*it) && (*it)->isCompleted() )
      {
        worker = (*it);
      }        
    }
  }

  m_worker = worker;
  if(worker==NULL)
  {
    Layer2Managers::getInstance().releaseResources(Resources(m_buildingType));
    m_reservingResources = false;
    terminate(TerminatedNoWorker);
  } else {
    //bool reserved = reserveUnit(worker);
    //m_buildingPlacer->useWorker(m_worker);

    //get a location
    BWAPI::TilePosition tp = BWAPI::TilePositions::None;
    tp = m_buildingPlacer->getBuildingLocation(worker,m_buildingType,m_location);

    if(tp == BWAPI::TilePositions::None)
    {
      terminate(TerminatedNoSpace);
    }

    //send the probe to construct
    else
    {
      worker = findClosestWorker(tp);
      if(worker==NULL)
      {
        Layer2Managers::getInstance().releaseResources(Resources(m_buildingType));
        m_reservingResources = false;
        terminate(TerminatedNoWorker);
      } else {
        bool reserved = addUnit(worker);
        assert(reserved);
        m_worker = worker;
        //m_buildingPlacer->useWorker(m_worker);
        //Should put himself in a list of action waiting for resources
        m_buildingPlacer->reservePosition(tp,m_buildingType);
        m_buildingTP = tp;
        //bool testTopLeft = tp.isValid();
        //bool testTopRight = BWAPI::TilePosition(tp.x()+m_buildingType.tileWidth()-1,tp.y()).isValid();
        //bool testBottomLeft = BWAPI::TilePosition(tp.x(),tp.y()-m_buildingType.tileHeight()+1).isValid();
        //bool testBottomRight = BWAPI::TilePosition(tp.x()+m_buildingType.tileWidth()-1,tp.y()-m_buildingType.tileHeight()+1).isValid();
        //bool test2 = BWAPI::Broodwar->isExplored(tp);
        //bool test3 = worker->build(tp,m_buildingType);

        m_buildFSM = new BuildFSM(this, m_priority,worker,tp,m_buildingType);
        m_buildFSM->execute();

        //m_unitCreatedSubscriberId = Spar->getPerceptualState().getUnitCreateEvent().subscribe(&m_unitCreatedEventHandler, NULL);
        m_unitCreatedSubscriberId = Spar->getPerceptualState().getMap()->getOwnUnitAddedEvent().subscribe(&m_unitCreatedEventHandler, NULL);
      }
    }
  }
}

BWAPI::Unit* L2Build::findClosestWorker(BWAPI::TilePosition tp)
{
  BWAPI::Unit* worker = NULL;
  double minDist = 100000;
  std::set<BWAPI::Unit*> unitSet = BWAPI::Broodwar->self()->getUnits();
  for(std::set<BWAPI::Unit*>::const_iterator it = unitSet.begin() ; (it!=unitSet.end()) ; ++it)
  {
    if((*it)->getType() == BWAPI::UnitTypes::Protoss_Probe && (*it)->isCompleted() )
    {
      int dist = (*it)->getDistance(BWAPI::Position(tp.x()*TILE_SIZE,tp.y()*TILE_SIZE));
      const layer2::GroupTacticalAction* act = dynamic_cast<Layer2UnitData&>(Spar->getUnitManager().getUnitData(*it)).getCurrentAction();
      int prio;
      if(act == NULL)
      {
        prio = 0;
      }
      else
      {
        prio = act->getPriority();
      }
      if(prio < m_priority && dist < minDist)
      {
        worker = (*it);
        minDist = dist;
      }        
    }
  }
  return worker;
}

bool L2Build::asCommanded(BWAPI::Unit* unit)
{
  return (unit->getTilePosition() == m_buildingTP && unit->getType() == m_buildingType);
}

void L2Build::addReservation(BWAPI::TilePosition tp)
{
  //Set the building place in 
  m_buildingProgress.insert(std::pair<BWAPI::TilePosition,BuildingProgress>(tp,reserved));
}

void L2Build::onUnitCreated(void* data, OwnUnit* u)
{
  BWAPI::Unit* unit = u->getUnit();
  if(asCommanded(unit))
  {
    Layer2Managers::getInstance().releaseResources(Resources(m_buildingType));
    m_reservingResources = false;
    m_event.activate();
    m_building = unit;
    if(m_worker!=NULL)
    {
      removeUnit(m_worker);
      m_worker= NULL;
    }
    //Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
    Spar->getPerceptualState().getMap()->getOwnUnitAddedEvent().unsubscribe(m_unitCreatedSubscriberId);
    //m_buildFSM = NULL;
    m_buildingConstructedEvent.raise(this, m_building, Spar->getUnitManager().getUnitLocation(m_building));
    m_unitDestroyedSubscriberId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_building, &m_unitDestroyedEventHandler, NULL);
    //if nb pending>0
    //void buildingStarted(BWAPI::Unit* unit);    
  }
}

void L2Build::onTimeout()
{
  //Check if building was cancelled before the timeout

  if(m_building != NULL)
  {
    m_unitCompletedSubscriberId = Spar->getPerceptualState().getUnitCompletedEvent().subscribe(&m_unitCompletedEventHandler, NULL);
    m_event.deactivate();
  }
}

void L2Build::onUnitCompleted(void* data, BWAPI::Unit* unit)
{
  if(m_building != NULL && m_buildingType == unit->getType())
  {
    if(m_buildingType == BWAPI::UnitTypes::Protoss_Pylon)
    {
      m_buildingPlacer->setPylonFalse();
    }
    else if(m_buildingType == BWAPI::UnitTypes::Protoss_Nexus)
    {
      m_buildingPlacer->addNewNexus(unit);
    }
    terminate(Success);
  }
}

void L2Build::onUnitDestroyed(void* data)
{
  m_event.deactivate();
  terminate(Cancelled);
}