#pragma once
#include "../../Interface/Actions/Build.h"
#include "../../../Layer1Reactive/FSM/Behaviors/BuildFSM.h"
#include "../NewBuildingPlacer.h"
#include "L2GroupTacticalAction.h"

// TODO: cancel building if it is about to get destroyed
// TODO: build actions sometimes hang (and nothing ever happens)

class L2Build : public virtual layer2::Build, public L2GroupTacticalAction
{
public:
  /**
   * @brief Instantiates a Build action.
   * @param priority The priority of the action.
   * @param location The location at which the building should be constructed.
   * @param buildingType The type of building to construct.
   */
  L2Build(unsigned int priority, StaticLocation* location, BWAPI::UnitType buildingType)
    : Build(priority, location, buildingType)
    , L2GroupTacticalAction(priority)
    , GroupTacticalAction(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_event(*this,priority,0,buildingType.buildTime()-2)
    , m_resourceAllocatedEventHandler(*this)
    , m_unitCreatedEventHandler(*this)
    , m_unitCompletedEventHandler(*this)
    , m_buildFSM(*this)
    , m_resourcesRequestID(m_resourceAllocatedEventHandler)
    , m_reservingResources(false)
    , m_unitDestroyedEventHandler(*this)
  {
    m_buildingPlacer = NewBuildingPlacer::getInstance();
    m_building = NULL;
    m_worker = NULL;
  }

  virtual ActionResources getPendingResourcesConsumption() const
  {
    if(m_building == NULL)
    {
      return ActionResources(Resources(m_buildingType));
    }
    else
    {
      return ActionResources(0,0,0);
    }
  }

  void buildingStarted()
  {
    //--m_pendingNb;
  }

  void onResourcesAllocated(void* data, BWAPI::Unit* b);
  void onTimeout();
  void onUnitCreated(void* data, OwnUnit*);
  void onUnitDestroyed(void* data);

protected:
  virtual void executeSpecificImpl();
  virtual void cancelSpecificImpl();
  virtual void terminateSpecificImpl();

  virtual void onUnitAddedToGroup(void*, BWAPI::Unit* unit);
  virtual void onUnitRemovedFromGroup(void*, BWAPI::Unit* unit);

  bool asCommanded(BWAPI::Unit* unit);
  void onUnitCompleted(void* data, BWAPI::Unit* unit);
  void addReservation(BWAPI::TilePosition tp);
  BWAPI::Unit* findClosestWorker(BWAPI::TilePosition tp);

  BWAPI::Unit* m_building;
  BWAPI::Unit* m_worker;
  BWAPI::TilePosition m_buildingTP;
  NewBuildingPlacer* m_buildingPlacer;
  bool m_reservingResources;

  ProcessContainer<BuildFSM> m_buildFSM;
  SCHEDULER_EVENT(L2Build,onTimeout) m_event;
  Event<boost::tuples::tuple<OwnUnit*>>::SubscriberID m_unitCreatedSubscriberId;
  Event<boost::tuples::tuple<BWAPI::Unit*>>::SubscriberID m_unitCompletedSubscriberId;
  Layer2ManagersBase::ResourcesRequestID m_resourcesRequestID;
  Event<boost::tuples::tuple<>>::SubscriberID m_unitDestroyedSubscriberId;
    
  
  EVENT_HANDLER(L2Build, onUnitDestroyed) m_unitDestroyedEventHandler;
  EVENT_HANDLER1(L2Build, onResourcesAllocated, BWAPI::Unit*) m_resourceAllocatedEventHandler;
  EVENT_HANDLER1(L2Build, onUnitCreated, OwnUnit*) m_unitCreatedEventHandler;
  EVENT_HANDLER1(L2Build, onUnitCompleted, BWAPI::Unit*) m_unitCompletedEventHandler;
};
