#include "stdafx.h"
#include "L2Collect.h"
#include "../Layer2Managers.h"

void L2CollectMinerals::executeSpecificImpl()
{
  L2GroupTacticalAction::executeSpecificImpl();

  m_gatherBehaviorTerminatedSubscriberId = m_gatherBehavior.getTerminateEvent().subscribe(&m_gatherBehaviorTerminatedEventHandler, NULL);

  std::pair<BaseLocation::crystal_iterator, BaseLocation::crystal_iterator> crystals = m_baseLocation->getLastKnownCrystalsLeft();
  if(crystals.first == crystals.second)
  {
    //TODO: terminate(NoMoreMinerals);
    terminate(Success);
  } else {
    bool accessibleMineralFound = false;
    BaseLocation::crystal_iterator mineralIt = crystals.first;
    while(mineralIt != crystals.second && !accessibleMineralFound)
    {
      accessibleMineralFound = (*mineralIt)->isAccessible();
      ++mineralIt;
    }

    if(accessibleMineralFound)
    {
      const BWTA::Region* baseRegion = m_baseLocation->getParent()->getBWTARegion();
      for(PersistentUnitGroup::const_iterator it = getGroup()->begin(); it != getGroup()->end(); ++it)
      {
        assert((*it)->getType() == BWAPI::UnitTypes::Protoss_Probe);
        if(baseRegion != BWTA::getRegion((*it)->getTilePosition()))
        {
          m_movingGroup->insert(*it);
        } else {
          m_gatheringGroup->insert(*it);
        }        
      }
    } else {
      for(PersistentUnitGroup::const_iterator it = getGroup()->begin(); it != getGroup()->end(); ++it)
      {
        assert((*it)->getType() == BWAPI::UnitTypes::Protoss_Probe);
        m_movingGroup->insert(*it);
      }
    }
    
    if(m_movingGroup->size() > 0 && m_moveBehavior == NULL)
      startMoveBehavior();

    m_gatherBehavior.execute();
  }
}

void L2CollectMinerals::onUnitAddedToGroup(void*, BWAPI::Unit* unit)
{
  assert(unit->getType() == BWAPI::UnitTypes::Protoss_Probe);
  bool accessibleMineralFound = false;
  
  std::pair<BaseLocation::crystal_iterator, BaseLocation::crystal_iterator> crystals = m_baseLocation->getLastKnownCrystalsLeft();
  BaseLocation::crystal_iterator mineralIt = crystals.first;
  while(mineralIt != crystals.second && !accessibleMineralFound)
  {
    accessibleMineralFound = (*mineralIt)->isAccessible();
    ++mineralIt;
  }

  if(!accessibleMineralFound || m_baseLocation->getParent()->getBWTARegion() != BWTA::getRegion(unit->getTilePosition()))
  {
    m_movingGroup->insert(unit);
    if(m_movingGroup->size() > 0 && m_moveBehavior == NULL)
      startMoveBehavior();
  } else {
    m_gatheringGroup->insert(unit);
  }
}

void L2CollectMinerals::onUnitRemovedFromGroup(void*, BWAPI::Unit* unit)
{
  bool removed = m_movingGroup->remove(unit);
  if(!removed)
    m_gatheringGroup->remove(unit);
}

void L2CollectMinerals::startMoveBehavior()
{
  assert(m_moveBehavior == NULL);

  std::pair<BaseLocation::crystal_iterator, BaseLocation::crystal_iterator> crystals = m_baseLocation->getLastKnownCrystalsLeft();
  
  m_moveBehavior = new Move(this, getPriority(), m_movingGroup,
    (*crystals.first)->getTilePosition(), MINERAL_PROXIMITY, 5, 5);
  Process::zombifyOnTerminate(m_moveBehavior);
  m_unitAtDestinationSubscriberId = m_moveBehavior->getUnitAtDestinationEvent().subscribe(&m_unitAtDestinationEventHandler, NULL);
  m_moveBehaviorTerminatedSubscriberId = m_moveBehavior->getTerminateEvent().subscribe(&m_moveBehaviorTerminatedEventHandler, NULL);
  m_moveBehavior->execute();
}

void L2CollectMinerals::onUnitAtDestination(void* data, BWAPI::Unit* unit)
{
  m_movingGroup->remove(unit);
  m_gatheringGroup->insert(unit);
}

void L2CollectMinerals::onMoveBehaviorTerminated(void* data)
{
  m_moveBehavior->getUnitAtDestinationEvent().unsubscribe(m_unitAtDestinationSubscriberId);
  m_moveBehavior->getTerminateEvent().unsubscribe(m_moveBehaviorTerminatedSubscriberId);
  m_moveBehavior = NULL;
}

void L2CollectMinerals::onGatherBehaviorTerminated(void* data)
{
  if(isExecuting())
  {
    m_gatherBehavior.getTerminateEvent().unsubscribe(m_gatherBehaviorTerminatedSubscriberId);
    if(m_moveBehavior != NULL)
    {
      m_moveBehavior->getUnitAtDestinationEvent().unsubscribe(m_unitAtDestinationSubscriberId);
      m_moveBehavior->getTerminateEvent().unsubscribe(m_moveBehaviorTerminatedSubscriberId);
      m_moveBehavior = NULL;
    }
  }

  terminate(Success);
}

void L2CollectMinerals::cancelSpecificImpl()
{
  L2Collect::cancelSpecificImpl();

  if(isExecuting())
    m_gatherBehavior.getTerminateEvent().unsubscribe(m_gatherBehaviorTerminatedSubscriberId);
  m_gatherBehavior.externalTerminate();
  if(m_moveBehavior != NULL)
  {
    m_moveBehavior->externalTerminate();
    m_moveBehavior->getUnitAtDestinationEvent().unsubscribe(m_unitAtDestinationSubscriberId);
    m_moveBehavior->getTerminateEvent().unsubscribe(m_moveBehaviorTerminatedSubscriberId);
    m_moveBehavior = NULL;
  }

  terminate(Cancelled);
  // TODO
}

void L2CollectMinerals::terminateSpecificImpl()
{
  L2Collect::terminateSpecificImpl();

  if(m_moveBehavior != NULL)
  {
    m_moveBehavior->getUnitAtDestinationEvent().unsubscribe(m_unitAtDestinationSubscriberId);
    m_moveBehavior->getTerminateEvent().unsubscribe(m_moveBehaviorTerminatedSubscriberId);
    m_moveBehavior = NULL;
  }
}


/////////////////////////////////////
// CollectGas
/////////////////////////////////////

//TODO: Manage multiple geyser in one location

void L2CollectGas::executeSpecificImpl()
{
  L2GroupTacticalAction::executeSpecificImpl();
  std::pair<StaticLocation::own_unit_type_const_iterator, StaticLocation::own_unit_type_const_iterator> assimilator = this->m_baseLocation->getOwnUnits(BWAPI::UnitTypes::Protoss_Assimilator);
  if(assimilator.first != assimilator.first)
  {
    for(PersistentUnitGroup::const_iterator it = getGroup()->begin(); it!=getGroup()->end() ; ++it)
    {
      (*it)->gather((*assimilator.first)->getUnit());
    }   
  }
}


void L2CollectGas::cancelSpecificImpl()
{
  L2Collect::cancelSpecificImpl();
  terminate(Cancelled);
}

void L2CollectGas::terminateSpecificImpl()
{
  L2Collect::terminateSpecificImpl();
}

void L2CollectGas::onUnitAddedToGroup(void*, BWAPI::Unit* unit)
{
  std::pair<StaticLocation::own_unit_type_const_iterator, StaticLocation::own_unit_type_const_iterator> assimilator = this->m_baseLocation->getOwnUnits(BWAPI::UnitTypes::Protoss_Assimilator);
  if(assimilator.first != assimilator.second)
  {
    unit->gather((*assimilator.first)->getUnit());
  }
}

void L2CollectGas::onUnitRemovedFromGroup(void*, BWAPI::Unit* unit)
{
}
