#pragma once
#include "../../../Layer1Reactive/FSM/Behaviors/GatherMinerals.h"
#include "../../../Layer1Reactive/FSM/Behaviors/Move.h"
#include "../../Interface/Actions/Collect.h"
#include "L2GroupTacticalAction.h"

class L2Collect : public virtual layer2::Collect, public L2GroupTacticalAction
{
public:
  L2Collect(unsigned int priority, const PersistentUnitGroup* initialGroup, BaseLocation* baseLocation)
    : Collect(priority, initialGroup, baseLocation)
    , L2GroupTacticalAction(priority, initialGroup)
    , GroupTacticalAction(priority, initialGroup)
    , TacticalAction(priority)
    , Process(Spar)
  {
  }

};

class L2CollectMinerals : public virtual layer2::CollectMinerals, public L2Collect
{
public:
  L2CollectMinerals(unsigned int priority, const PersistentUnitGroup* initialGroup, BaseLocation* baseLocation)
    : CollectMinerals(priority, initialGroup, baseLocation)
    , L2Collect(priority, initialGroup, baseLocation)
    , Collect(priority, initialGroup, baseLocation)
    , GroupTacticalAction(priority, initialGroup)
    , TacticalAction(priority)
    , Process(Spar)
    , m_baseLocation(baseLocation)
    , m_gatherBehavior(this, priority, m_gatheringGroup, baseLocation)
    , m_gatherBehaviorTerminatedEventHandler(*this)
    , m_moveBehavior(NULL)
    , m_unitAtDestinationEventHandler(*this)
    , m_moveBehaviorTerminatedEventHandler(*this)
    , m_movingGroup(new PersistentUnitGroup("Collect Mineral, Moving group"), *this)
    , m_gatheringGroup(new PersistentUnitGroup("Collect Mineral, Gathering group"), *this)
  {
  }

private:
  // Groups must be declared before their behaviors !
  AutoPtr<PersistentUnitGroup> m_movingGroup;
  AutoPtr<PersistentUnitGroup> m_gatheringGroup;

  GatherMinerals m_gatherBehavior;
  EVENT_HANDLER(L2CollectMinerals, onGatherBehaviorTerminated) m_gatherBehaviorTerminatedEventHandler;
  Event<boost::tuples::tuple<>>::SubscriberID m_gatherBehaviorTerminatedSubscriberId;

  Move* m_moveBehavior;
  EVENT_HANDLER1(L2CollectMinerals, onUnitAtDestination, BWAPI::Unit*) m_unitAtDestinationEventHandler;
  Event<boost::tuples::tuple<BWAPI::Unit*>>::SubscriberID m_unitAtDestinationSubscriberId;
  EVENT_HANDLER(L2CollectMinerals, onMoveBehaviorTerminated) m_moveBehaviorTerminatedEventHandler;
  Event<boost::tuples::tuple<>>::SubscriberID m_moveBehaviorTerminatedSubscriberId;

protected:
  // must be smaller than the probe sight range (256)
  static const size_t MINERAL_PROXIMITY = 128;

  virtual void executeSpecificImpl();
  virtual void cancelSpecificImpl();
  virtual void terminateSpecificImpl();

  virtual void onUnitAddedToGroup(void*, BWAPI::Unit* unit);
  virtual void onUnitRemovedFromGroup(void*, BWAPI::Unit* unit);
  
  void startMoveBehavior();

  void onUnitAtDestination(void* data, BWAPI::Unit* unit);
  void onMoveBehaviorTerminated(void* data);
  void onGatherBehaviorTerminated(void* data);

  BaseLocation* m_baseLocation;
};

class L2CollectGas : public virtual layer2::CollectGas, public L2Collect
{
public:
  L2CollectGas(unsigned int priority, const PersistentUnitGroup* initialGroup, BaseLocation* baseLocation)
    : CollectGas(priority, initialGroup, baseLocation)
    , L2Collect(priority, initialGroup, baseLocation)
    , Collect(priority, initialGroup, baseLocation)
    , GroupTacticalAction(priority, initialGroup)
    , TacticalAction(priority)
    , Process(Spar)
  {
  }

protected:
  virtual void executeSpecificImpl();
  virtual void cancelSpecificImpl();
  virtual void terminateSpecificImpl();

  virtual void onUnitAddedToGroup(void*, BWAPI::Unit* unit);
  virtual void onUnitRemovedFromGroup(void*, BWAPI::Unit* unit);
};
