#include "stdafx.h"
#include "L2Defend.h"
#include "../Layer2Managers.h"
#include "../../../../PerceptualState/Locations/Dynamic/UnitDynamicLocation.h"

void L2Defend::executeSpecificImpl()
{
  L2GroupTacticalAction::executeSpecificImpl();
  subscribeBegin();
  m_currentLocation = regionAttackedByEnemy();
  m_onBuildingAddedSubscriberId = Spar->getPerceptualState().getMap()->getOwnUnitAddedEvent().subscribe(&m_onBuildingAddedEventHandler, NULL);
  if(m_currentLocation != NULL)
  {
    waiting = false;
    if(m_currentPosition!=NULL)
    {
      delete m_currentPosition;
    }
    m_currentPosition = new BWAPI::Position((*m_currentLocation->getEnemyUnits().first)->getPosition());
    m_currentTarget = (*m_currentLocation->getEnemyUnits().first);
    m_groupAttackFSM = new GroupAttack(this, 100, getGroup(), new UnitDynamicLocation((*m_currentLocation->getEnemyUnits().first)->getUnit()));
    Process::zombifyOnTerminate(m_groupAttackFSM);
    m_onEnemyRemovedSubscriberId = m_currentLocation->getEnemyUnitRemovedEvent().subscribe(&m_onEnemyRemovedEventHandler , NULL);
    m_groupAttackFSM->execute();
  }
  else
  {
    waiting = true;
    if(m_currentPosition!=NULL)
    {
      delete m_currentPosition;
    }
    m_currentPosition = new BWAPI::Position((*Spar->getPerceptualState().getMap()->getOwnUnits(BWAPI::UnitTypes::Protoss_Nexus).first)->getPosition());
    getGroup()->move(*m_currentPosition);
  }  
}


void L2Defend::cancelSpecificImpl()
{
  L2GroupTacticalAction::cancelSpecificImpl();
  terminate(Cancelled);
}

void L2Defend::terminateSpecificImpl()
{
  unsubscribeAllEnemyUnitAdded();
  Spar->getPerceptualState().getMap()->getOwnUnitAddedEvent().unsubscribe(m_onBuildingAddedSubscriberId);
  if(!waiting)
  {
    m_currentLocation->getEnemyUnitRemovedEvent().unsubscribe(m_onEnemyRemovedSubscriberId);
  }
  if(m_groupAttackFSM != NULL)
  {
    m_groupAttackFSM->externalTerminate();
  }
  L2GroupTacticalAction::terminateSpecificImpl();
}


void L2Defend::onUnitAddedToGroup(void*, BWAPI::Unit* unit)
{
  if(m_groupAttackFSM == NULL && !waiting)
  {
    m_currentLocation = regionAttackedByEnemy();
    if(m_currentLocation != NULL)
    {
      StaticLocation::enemy_unit_iterator it = m_currentLocation->getEnemyUnits().first;
      m_currentTarget = NULL;
      while(it!=m_currentLocation->getEnemyUnits().second && m_currentTarget == NULL)
      {
        if((*it)->isAccessible())
        {
          m_currentTarget = (*it);
        }
        it++;
      }
      if(m_groupAttackFSM!=NULL)
      {
        m_groupAttackFSM->externalTerminate();
      }
      if(m_currentTarget!=NULL)
      {
        if(m_currentPosition!=NULL)
        {
          delete m_currentPosition;
        }
        m_currentPosition = new BWAPI::Position(m_currentTarget->getPosition());
        m_groupAttackFSM = new GroupAttack(this, 100, getGroup(), new UnitDynamicLocation(m_currentTarget->getUnit()));
        Process::zombifyOnTerminate(m_groupAttackFSM);
        m_groupAttackFSM->execute();
      }
      else
      {
        if(m_currentPosition!=NULL)
        {
          delete m_currentPosition;
        }
        m_currentPosition = new BWAPI::Position((*m_currentLocation->getEnemyUnits().first)->getPosition());
        m_currentTarget = (*m_currentLocation->getEnemyUnits().first);
        m_groupAttackFSM = new GroupAttack(this, 100, getGroup(), new CoordinatesLocation(m_currentTarget->getPosition()));
        Process::zombifyOnTerminate(m_groupAttackFSM);
        m_groupAttackFSM->execute();
      }
    }
  }
  if(waiting && m_currentPosition)
  {
    unit->move(*m_currentPosition);
  }
  else if(!waiting && m_currentPosition)
  {
    unit->attack(*m_currentPosition);
  }
}

void L2Defend::onUnitRemovedFromGroup(void*, BWAPI::Unit* unit)
{
  // TODO
}

void L2Defend::onEnemyAdded(void*, EnemyUnit* unit)
{
  if(waiting){
    waiting = false;
    m_currentTarget = unit;
    if(m_currentPosition!=NULL)
    {
      delete m_currentPosition;
    }
    m_currentPosition = new BWAPI::Position(unit->getPosition());
    m_currentLocation = unit->getLocation();
    m_groupAttackFSM = new GroupAttack(this, 100, getGroup(), new UnitDynamicLocation(m_currentTarget->getUnit()));
    Process::zombifyOnTerminate(m_groupAttackFSM);
    m_onEnemyRemovedSubscriberId = m_currentLocation->getEnemyUnitRemovedEvent().subscribe(&m_onEnemyRemovedEventHandler , NULL);
    m_groupAttackFSM->execute();
  }
}

void L2Defend::onEnemyRemoved(void*, EnemyUnit* unit)
{
  int i = m_currentLocation->getNbEnemyUnits();
  if(m_currentLocation->getNbEnemyUnits() == 0){
    if(m_groupAttackFSM!=NULL)
    {
      m_groupAttackFSM->externalTerminate();
    }
    m_currentLocation->getEnemyUnitRemovedEvent().unsubscribe(m_onEnemyRemovedSubscriberId);
    m_currentLocation->getEnemyUnitAddedEvent().unsubscribe(m_onEnemyRemovedSubscriberId);
    m_currentLocation = regionAttackedByEnemy();

    if(m_currentLocation != NULL)
    {
      assert(m_currentLocation->getEnemyUnits().first != m_currentLocation->getEnemyUnits().second);
      m_onEnemyRemovedSubscriberId = m_currentLocation->getEnemyUnitRemovedEvent().subscribe(&m_onEnemyRemovedEventHandler , NULL);
      if(m_currentPosition!=NULL)
      {
        delete m_currentPosition;
      }
      m_currentPosition = new BWAPI::Position((*(m_currentLocation->getEnemyUnits().first))->getPosition());
      m_currentTarget = *m_currentLocation->getEnemyUnits().first;
      m_groupAttackFSM = new GroupAttack(this, 100, getGroup(), new UnitDynamicLocation(m_currentTarget->getUnit()));
      Process::zombifyOnTerminate(m_groupAttackFSM);
      m_groupAttackFSM->execute();
    }
    else
    {
      waiting = true;
      if(m_currentPosition!=NULL)
      {
        delete m_currentPosition;
      }
      m_currentPosition = new BWAPI::Position((*Spar->getPerceptualState().getMap()->getOwnUnits(BWAPI::UnitTypes::Protoss_Nexus).first)->getPosition());
      getGroup()->move(*m_currentPosition);
    }
  }
  else if(unit==m_currentTarget)
  {
    assert(m_currentLocation->getEnemyUnits().first != m_currentLocation->getEnemyUnits().second);
    StaticLocation::enemy_unit_iterator it = m_currentLocation->getEnemyUnits().first;
    m_currentTarget = NULL;
    while(it!=m_currentLocation->getEnemyUnits().second && m_currentTarget == NULL)
    {
      if((*it)->isAccessible())
      {
        m_currentTarget = (*it);
      }
      it++;
    }
    if(m_groupAttackFSM!=NULL)
    {
      m_groupAttackFSM->externalTerminate();
    }
    if(m_currentTarget!=NULL)
    {
      if(m_currentPosition!=NULL)
      {
        delete m_currentPosition;
      }
      m_currentPosition = new BWAPI::Position(m_currentTarget->getPosition());
      m_groupAttackFSM = new GroupAttack(this, 100, getGroup(), new UnitDynamicLocation(m_currentTarget->getUnit()));
      Process::zombifyOnTerminate(m_groupAttackFSM);
      m_groupAttackFSM->execute();
    }
    else
    {
      if(m_currentPosition!=NULL)
      {
        delete m_currentPosition;
      }
      m_currentPosition = new BWAPI::Position((*m_currentLocation->getEnemyUnits().first)->getPosition());
      m_currentTarget = (*m_currentLocation->getEnemyUnits().first);
      m_groupAttackFSM = new GroupAttack(this, 100, getGroup(), new CoordinatesLocation(m_currentTarget->getPosition()));
      Process::zombifyOnTerminate(m_groupAttackFSM);
      m_groupAttackFSM->execute();
    }
  }
}


StaticLocation* L2Defend::regionAttackedByEnemy()
{
  typedef std::list<PrimitiveLocation*>::iterator primitive_sub_location_iterator;
  std::pair<primitive_sub_location_iterator, primitive_sub_location_iterator> pi;
  pi = Spar->getPerceptualState().getMap()->getPrimitiveSubLocations();
  for(primitive_sub_location_iterator i = pi.first ; i!=pi.second ; ++i)
  {
    if((*i)->getNbOwnBuildings()>0)
    {
      if(dynamic_cast<BaseLocation*>(*i))
      {
        Region* region = dynamic_cast<Region*>((*i)->getParent());
        if(region->getRegionLocation()->getNbEnemyUnits() > 0)
        {
          return region->getRegionLocation();
        }
        else if((*i)->getNbEnemyUnits() > 0)
        {
          return (*i);
        }
      }
      else if((*i)->getNbEnemyUnits()>0)
      {
        return (*i);
      }
    } 
  }
  return NULL;
}

void L2Defend::onBuildingAdded(void*, OwnUnit* unit)
{
  if(unit->getType().isBuilding())
  {
    PrimitiveLocation* tempLocation = unit->getLocation();
    //CompositeLocation* tempParent = tempLocation->getParent();
    if(tempLocation->getNbOwnBuildings()==1)
    {
      if(unit->getType() == BWAPI::UnitTypes::Protoss_Nexus)
      {
        Region* lolTest = dynamic_cast<Region*>(tempLocation->getParent());
        subscribeEnemyUnitAdded(lolTest->getRegionLocation());
      }
      else
      {
        subscribeEnemyUnitAdded(tempLocation);
      }
    }
  }
}

void L2Defend::unsubscribeAllEnemyUnitAdded()
{
  typedef std::map<StaticLocation*,Event<boost::tuples::tuple<EnemyUnit*>>::SubscriberID>::iterator mapIterator;
  for(mapIterator it = mapSubscription.begin() ; it != mapSubscription.end() ; ++it)
  {
    (it->first)->getEnemyUnitAddedEvent().unsubscribe(it->second);
  }
}

void L2Defend::subscribeEnemyUnitAdded(StaticLocation* cl)
{
  typedef Event<boost::tuples::tuple<EnemyUnit*>>::SubscriberID a;
  if(mapSubscription.find(cl) == mapSubscription.end())
  {
    mapSubscription.insert(std::pair<StaticLocation*,a>(cl,cl->getEnemyUnitAddedEvent().subscribe(&m_onEnemyAddedEventHandler, NULL)));
  }
}

void L2Defend::subscribeBegin()
{
  typedef std::list<PrimitiveLocation*>::iterator primitive_sub_location_iterator;
  std::pair<primitive_sub_location_iterator, primitive_sub_location_iterator> pi;
  pi = Spar->getPerceptualState().getMap()->getPrimitiveSubLocations();
  for(primitive_sub_location_iterator i = pi.first ; i!=pi.second ; ++i)
  {
    if((*i)->getNbOwnBuildings()>0)
    {
      subscribeEnemyUnitAdded((*i));
    } 
  }
}
