#pragma once
#include "../../Interface/Actions/Defend.h"
#include "../../../Layer1Reactive/FSM/Behaviors/EngageEnemies/GroupAttack.h"
#include "L2GroupTacticalAction.h"

// TODO: possibly more than two locations so units can patrol?

class L2Defend : public virtual layer2::Defend, public L2GroupTacticalAction
{
public:
  L2Defend(unsigned int priority, const PersistentUnitGroup* initialGroup, Location* location)
    : Defend(priority, initialGroup, location)
    , L2GroupTacticalAction(priority, initialGroup)
    , GroupTacticalAction(priority, initialGroup)
    , TacticalAction(priority)
    , Process(Spar)
    , m_groupAttackFSM(*this)
    , m_onEnemyAddedEventHandler(*this)
    , m_onBuildingAddedEventHandler(*this)
    , m_onEnemyRemovedEventHandler(*this)
  {
    m_currentLocation = NULL;
    m_currentPosition = NULL;
  }

  ~L2Defend()
  {
    delete m_currentPosition;
  }

  //void onTerminate(void* data);
  void onEnemyAdded(void* data, EnemyUnit*);
  StaticLocation* regionAttackedByEnemy();
  void onBuildingAdded(void* data, OwnUnit*);
  void onEnemyRemoved(void* data, EnemyUnit*);

protected:
  virtual void executeSpecificImpl();
  virtual void cancelSpecificImpl();
  virtual void terminateSpecificImpl();
  virtual void onUnitAddedToGroup(void*, BWAPI::Unit* unit);
  virtual void onUnitRemovedFromGroup(void*, BWAPI::Unit* unit);
  void unsubscribeAllEnemyUnitAdded();
  void subscribeEnemyUnitAdded(StaticLocation* cl);
  void subscribeBegin();

  Event<boost::tuples::tuple<OwnUnit*>>::SubscriberID m_onBuildingAddedSubscriberId;
  Event<boost::tuples::tuple<EnemyUnit*>>::SubscriberID m_onEnemyAddedSubscriberId;
  Event<boost::tuples::tuple<EnemyUnit*>>::SubscriberID m_onEnemyRemovedSubscriberId;
  //Event<boost::tuples::tuple<>>::SubscriberID m_onTerminateSubscriberId;

  EVENT_HANDLER1(L2Defend, onBuildingAdded, OwnUnit*) m_onBuildingAddedEventHandler;
  EVENT_HANDLER1(L2Defend, onEnemyAdded, EnemyUnit*) m_onEnemyAddedEventHandler;
  EVENT_HANDLER1(L2Defend, onEnemyRemoved, EnemyUnit*) m_onEnemyRemovedEventHandler;
  //EVENT_HANDLER(L2Defend, onTerminate) m_onTerminateEventHandle;

private:
  ProcessContainer<GroupAttack> m_groupAttackFSM;
  bool waiting;
  StaticLocation* m_currentLocation;
  BWAPI::Position* m_currentPosition;
  EnemyUnit* m_currentTarget;
  std::map<StaticLocation*,Event<boost::tuples::tuple<EnemyUnit*>>::SubscriberID> mapSubscription;
};
