#include "stdafx.h"
#include "L2Recruit.h"
#include "../Layer2Managers.h"

L2Recruit::~L2Recruit()
{
  for (RecruitListType::iterator it = m_individualRecruits.begin(); it != m_individualRecruits.end(); ++it)
    delete *it;
}

void L2Recruit::executeSpecificImpl()
{
  L2TacticalAction::executeSpecificImpl();
  for (std::map<BWAPI::UnitType, size_t>::const_iterator it = m_units.begin();
       it != m_units.end();
       ++it)
  {
    for(size_t i = 0; i < it->second; ++i)
    {
      RecruitListID id 
        = m_individualRecruits.insert(m_individualRecruits.end(), new L2IndividualRecruit(this, it->first));
      (*id)->execute(id);
    }
  }
}

void L2Recruit::cancelSpecificImpl()
{
  L2TacticalAction::cancelSpecificImpl();
  m_canceled = true;
  RecruitListID it;
  RecruitListID nextIt = m_individualRecruits.begin();
  while(nextIt != m_individualRecruits.end())
  {
    it = nextIt;
    ++nextIt;
    (*it)->cancel(false);
  }
}

void L2Recruit::terminateSpecificImpl()
{
  L2TacticalAction::terminateSpecificImpl();
  m_canceled = true;
  RecruitListID it;
  RecruitListID nextIt = m_individualRecruits.begin();
  while(nextIt != m_individualRecruits.end())
  {
    it = nextIt;
    ++nextIt;
    (*it)->terminate();
  }
}

void L2Recruit::individualRecruitStarted(RecruitListID id)
{
  if(!m_canceled && m_isInfinite)
  {
    RecruitListID newId = m_individualRecruits.insert(m_individualRecruits.end(), new L2IndividualRecruit(this, (*id)->getUnitToBuild()));
    (*newId)->execute(newId);
  }
}

void L2Recruit::individualRecruitTerminated(RecruitListID id)
{
  Spar->addZombieComponent(*id);
  m_individualRecruits.erase(id);
  if(m_individualRecruits.empty())
    terminate(Success);
}




// Class L2IndividualRecruit



void L2IndividualRecruit::execute(L2Recruit::RecruitListID id)
{  
  assert(m_unitToBuild.whatBuilds().first.isBuilding());
  m_id = id;

  if(Techs::getUnitTypeTech(m_unitToBuild)->isAcquired())
  {
    requestResources();
  } else {
    m_unitCreatedSubscriberId = Spar->getPerceptualState().getUnitCreateEvent().subscribe(&m_unitCreatedEventHandler, NULL);
    m_currentState = WAITING_FOR_TECH;
  }
}

void L2IndividualRecruit::requestResources()
{
  Layer2Managers::getInstance().requestResources(
    m_l2Recruit->getPriority(), 
    ActionResources(Resources(m_unitToBuild),m_l2Recruit->getLocation(), m_unitToBuild.whatBuilds().first), 
    m_resourcesRequestID);
  m_currentState = WAITING_FOR_RESOURCES;
}

void L2IndividualRecruit::onResourcesAllocated(void* data, BWAPI::Unit* building)
{  
  m_buildingToUse = building;

  if(Techs::getUnitTypeTech(m_unitToBuild)->isAcquired())
  {
    assert(m_unitToBuild.whatBuilds().first == building->getType());
    assert(dynamic_cast<Layer2ManagersUnitData&>(Spar->getUnitManager().getUnitData(m_buildingToUse)).isInUse());  
    assert(m_buildingToUse != NULL);
    assert(m_buildingToUse->isCompleted());
    assert(m_buildingToUse->getTrainingQueue().empty());
    assert(!m_buildingToUse->isUnpowered());
    // TODO: Check if the training queue is empty (or at least have some space)
    
    Resources oldResources = Resources(BWAPI::Broodwar->self());
    assert(m_buildingToUse->canIssueCommand(BWAPI::UnitCommand::train(m_buildingToUse, m_unitToBuild)));
    bool actionSent = m_buildingToUse->train(m_unitToBuild); 
    
    if(actionSent)
    {
      assert(oldResources == Resources(BWAPI::Broodwar->self()) + Resources(m_unitToBuild));
      Layer2Managers::getInstance().releaseResources(Resources(m_unitToBuild));
      m_l2Recruit->individualRecruitStarted(m_id);

      m_unitCreatedSubscriberId = Spar->getPerceptualState().getUnitCreateEvent().subscribe(&m_unitCreatedEventHandler, NULL);
      
      m_buildingDestroyedSubscriberId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_buildingToUse, &m_buildingDestroyedEventHandler,NULL);

      //TODO: getLatencyFrames and getRemainingLatencyFrames function do not return the same values, which one is the right one ?
      m_timeoutEvent.setWaitingTimeBeforeExecution(BWAPI::Broodwar->getLatencyFrames() + L2Recruit::RECRUIT_TIMEOUT);
      m_timeoutEvent.activate();

      m_currentState = ACTION_SENT;
    } else {
      Layer2Managers::getInstance().releaseResources(Resources(m_unitToBuild));
      recruitFailed();
    }
  } else {
    Layer2Managers::getInstance().releaseResources(Resources(m_unitToBuild));
    recruitFailed();
  }
}

void L2IndividualRecruit::onUnitCreated(void* data, BWAPI::Unit* unit)
{
  if(m_currentState != WAITING_FOR_TECH)
  {
    if(!m_buildingToUse->exists()) // TODO: CRASH when m_buildingToUse is NULL
    {
      m_timeoutEvent.deactivate();
      Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
      Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_buildingToUse, m_buildingDestroyedSubscriberId);
      recruitFailed();
    }

    if(unit->getType() == m_unitToBuild && unit->getPosition() == m_buildingToUse->getPosition())
    {
      m_unitInConstruction = unit;
      Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
      m_timeoutEvent.deactivate();
      Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_buildingToUse, m_buildingDestroyedSubscriberId);

      // We don't register to onUnitCompleted now for better performance. -2 to make sure we don't miss the event.
      m_timeoutEvent.setWaitingTimeBeforeExecution(m_buildingToUse->getRemainingTrainTime()-2);
      m_timeoutEvent.activate();
      m_unitDestroyedSubscriberId = Spar->getUnitManager().subscribeIndividualUnitDestroy(m_unitInConstruction, &m_unitDestroyedEventHandler, NULL);
      
      m_currentState = UNIT_CREATED;

      if(m_canceled)
      {
        m_buildingToUse->cancelTrain(0);
        recruitDone();
      }
    }
  } else {
    if(Techs::getUnitTypeTech(m_unitToBuild)->isAcquired())
    {
      Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
      requestResources();
    }
  }
}

void L2IndividualRecruit::onUnitCompleted(void* data, BWAPI::Unit* unit)
{
  if(m_unitInConstruction == unit)
  {
    Spar->getPerceptualState().getUnitCompletedEvent().unsubscribe(m_unitCompletedSubscriberId);
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
    m_l2Recruit->getUnitRecruitedEvent().raise(m_l2Recruit, m_unitInConstruction, Spar->getUnitManager().getUnitLocation(m_buildingToUse));
    recruitDone();
  }
}

void L2IndividualRecruit::onBuildingDestroyed(void* data)
{;
  assert(m_currentState == ACTION_SENT);
  Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
  Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_buildingToUse, m_buildingDestroyedSubscriberId);
  m_timeoutEvent.deactivate();
  recruitFailed();
}

void L2IndividualRecruit::onUnitDestroyed(void* data)
{
  assert(m_currentState == UNIT_CREATED || m_currentState == WAITING_TO_FINISH);
  if(m_currentState == UNIT_CREATED)
  {
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
    m_timeoutEvent.deactivate();
  } else if(m_currentState == WAITING_TO_FINISH) {
    Spar->getPerceptualState().getUnitCompletedEvent().unsubscribe(m_unitCompletedSubscriberId);
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
  }
  recruitFailed();
}

void L2IndividualRecruit::onTimeout()
{
  assert(m_currentState == ACTION_SENT || m_currentState == UNIT_CREATED);
  if(m_currentState == UNIT_CREATED)
  {
    m_unitCompletedSubscriberId = Spar->getPerceptualState().getUnitCompletedEvent().subscribe(&m_unitCompletedEventHandler, NULL);
    m_timeoutEvent.deactivate();
    m_currentState = WAITING_TO_FINISH;
  } else if(m_currentState == ACTION_SENT)
  {
    // This should never happen
    assert(false);
    m_timeoutEvent.deactivate();
    Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_buildingToUse, m_buildingDestroyedSubscriberId);
    recruitFailed();
  }
}

void L2IndividualRecruit::recruitDone()
{
  if(m_buildingToUse != NULL && m_buildingToUse->exists())
    dynamic_cast<Layer2ManagersUnitData&>(Spar->getUnitManager().getUnitData(m_buildingToUse)).setInUseFlag(false);
  m_currentState = TERMINATED;
  m_l2Recruit->individualRecruitTerminated(m_id);
}

void L2IndividualRecruit::recruitFailed()
{
  SPAR_LOG(LogTypes::LAYER_2_DECISION, "Individual Recruit Failed (%d) %s %d"
    , BWAPI::Broodwar->getFrameCount()
    , m_unitToBuild.getName().c_str()
    , (m_unitInConstruction==NULL ? -1 : m_unitInConstruction->getID()));
  if(m_buildingToUse != NULL && m_buildingToUse->exists())
    dynamic_cast<Layer2ManagersUnitData&>(Spar->getUnitManager().getUnitData(m_buildingToUse)).setInUseFlag(false);
  // Try again until it works
  m_buildingToUse = NULL;

  if(!m_canceled)
    execute(m_id);
}

void L2IndividualRecruit::cancel(bool cancelCurrentUnit)
{
  if(cancelCurrentUnit)
  {
    m_canceled = true;
    switch(m_currentState)
    {
    case UNINITIALIZED:
      recruitDone();
      break;
    case WAITING_FOR_TECH:
      Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
      recruitDone();
      break;
    case WAITING_FOR_RESOURCES:
      Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
      recruitDone();
      break;
    case ACTION_SENT:
      // Wait until the unit is created (or the building is destroyed)
      break;
    case UNIT_CREATED:
      Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
      if(m_buildingToUse->exists() && !m_buildingToUse->getTrainingQueue().empty())
      {
        //TODO: Find the right unit in the queue to remove
        assert(m_buildingToUse->getTrainingQueue().size()==1);
        m_buildingToUse->cancelTrain(0);
      }
      m_timeoutEvent.deactivate();
      recruitDone();
      break;
    case WAITING_TO_FINISH:
      Spar->getPerceptualState().getUnitCompletedEvent().unsubscribe(m_unitCompletedSubscriberId);
      Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
      if(m_buildingToUse->exists() && !m_buildingToUse->getTrainingQueue().empty())
      {
        //TODO: Find the right unit in the queue to remove
        assert(m_buildingToUse->getTrainingQueue().size()==1);
        m_buildingToUse->cancelTrain(0);
      }
      recruitDone();
    case TERMINATED:
      break;
    }
  } else {
    switch(m_currentState)
    {
    case UNINITIALIZED:
      recruitDone();
      break;
    case WAITING_FOR_TECH:
      Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
      recruitDone();
      break;
    case WAITING_FOR_RESOURCES:
      Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
      recruitDone();
      break;
    case ACTION_SENT:
    case UNIT_CREATED:
    case WAITING_TO_FINISH:
    case TERMINATED:
      break;
    }
  }
}

void L2IndividualRecruit::terminate()
{
  m_canceled = true;
  switch(m_currentState)
  {
  case UNINITIALIZED:
    recruitDone();
    break;
  case WAITING_FOR_TECH:
    Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
    recruitDone();
    break;
  case WAITING_FOR_RESOURCES:
    Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
    recruitDone();
    break;
  case ACTION_SENT:
    // At this point the unit will be lost, we cannot cancel it anymore.
    Spar->getPerceptualState().getUnitCreateEvent().unsubscribe(m_unitCreatedSubscriberId);
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_buildingToUse, m_buildingDestroyedSubscriberId);
    m_timeoutEvent.deactivate();
    recruitDone();
    break;
  case UNIT_CREATED:
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
    if(m_buildingToUse->exists() && !m_buildingToUse->getTrainingQueue().empty())
    {
      //TODO: Find the right unit in the queue to remove
      assert(m_buildingToUse->getTrainingQueue().size()==1);
      m_buildingToUse->cancelTrain(0);
    }
    m_timeoutEvent.deactivate();
    recruitDone();
    break;
  case WAITING_TO_FINISH:
    Spar->getPerceptualState().getUnitCompletedEvent().unsubscribe(m_unitCompletedSubscriberId);
    Spar->getUnitManager().unsubscribeIndividualUnitDestroy(m_unitInConstruction, m_unitDestroyedSubscriberId);
    if(m_buildingToUse->exists() && !m_buildingToUse->getTrainingQueue().empty())
    {
      //TODO: Find the right unit in the queue to remove
      assert(m_buildingToUse->getTrainingQueue().size()==1);
      m_buildingToUse->cancelTrain(0);
    }
    recruitDone();
    break;
  case TERMINATED:
    break;
  }
  Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
}