#pragma once
#include "../../../../Scheduler/SchedulerTask.h"
#include "../../Interface/Actions/Recruit.h"
#include "L2TacticalAction.h"
#include "../Layer2ManagersBase.h"

class L2IndividualRecruit;

class L2Recruit : public virtual layer2::Recruit, public L2TacticalAction
{
  friend L2IndividualRecruit;
public:
  static const size_t RECRUIT_TIMEOUT = 20;

  typedef std::list<L2IndividualRecruit*> RecruitListType;
  typedef RecruitListType::iterator RecruitListID;

   L2Recruit(unsigned int priority, StaticLocation* location, const std::map<BWAPI::UnitType, size_t>& unitTypes, bool isInfinite = false)
    : Recruit(priority, location, unitTypes, isInfinite)
    , L2TacticalAction(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_canceled(false)
    , m_isInfinite(isInfinite)
  {}

  L2Recruit(unsigned int priority, StaticLocation* location, BWAPI::UnitType unitType, bool isInfinite = false)
    : Recruit(priority, location, unitType, isInfinite)
    , L2TacticalAction(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_canceled(false)
    , m_isInfinite(isInfinite)
  {}

  ~L2Recruit();

protected:
  virtual void executeSpecificImpl();
  virtual void cancelSpecificImpl();
  virtual void terminateSpecificImpl();

  void individualRecruitTerminated(RecruitListID id);
  void individualRecruitStarted(RecruitListID id);

  bool m_isInfinite;
  bool m_canceled;
  
  RecruitListType m_individualRecruits;
};

class L2IndividualRecruit : public Component
{
public:
  L2IndividualRecruit(L2Recruit* l2Recruit, BWAPI::UnitType unitType)
    : Component()
    , m_timeoutEvent(*this, Scheduler::SCHEDULING_MEDIUM_PRIORITY,1,0)
    , m_l2Recruit(l2Recruit)
    , m_unitToBuild(unitType)
    , m_resourceAllocatedEventHandler(*this)
    , m_unitCreatedEventHandler(*this)
    , m_unitCompletedEventHandler(*this)
    , m_unitDestroyedEventHandler(*this)
    , m_buildingDestroyedEventHandler(*this)
    , m_buildingToUse(NULL)
    , m_unitInConstruction(NULL)
    , m_currentState(UNINITIALIZED)
    , m_canceled(false)
    , m_resourcesRequestID(m_resourceAllocatedEventHandler)
  {}

  void execute(L2Recruit::RecruitListID it);
  void cancel(bool cancelCurrentUnit);
  void terminate();
  BWAPI::UnitType getUnitToBuild() { return m_unitToBuild; }

  void onResourcesAllocated(void* data, BWAPI::Unit* building);
  void onTimeout();
  void onUnitCreated(void* data, BWAPI::Unit* unit);
  void onUnitCompleted(void* data, BWAPI::Unit* unit);
  void onUnitDestroyed(void* data);
  void onBuildingDestroyed(void* data);

protected:
  void requestResources();
  void recruitDone();
  void recruitFailed();

  enum RecruitState
  {
    UNINITIALIZED,
    WAITING_FOR_TECH,
    WAITING_FOR_RESOURCES,
    ACTION_SENT,
    UNIT_CREATED,
    WAITING_TO_FINISH,
    TERMINATED
  };

  RecruitState m_currentState;
  bool m_canceled;
  BWAPI::Unit* m_buildingToUse;
  BWAPI::UnitType m_unitToBuild;
  BWAPI::Unit* m_unitInConstruction;
  L2Recruit* m_l2Recruit;
  L2Recruit::RecruitListID m_id;

  SCHEDULER_EVENT(L2IndividualRecruit,onTimeout) m_timeoutEvent;
  Event<boost::tuples::tuple<BWAPI::Unit*>>::SubscriberID m_unitCreatedSubscriberId;
  Event<boost::tuples::tuple<BWAPI::Unit*>>::SubscriberID m_unitCompletedSubscriberId;
  Event<boost::tuples::tuple<>>::SubscriberID m_unitDestroyedSubscriberId;
  Event<boost::tuples::tuple<>>::SubscriberID m_buildingDestroyedSubscriberId;
  Layer2ManagersBase::ResourcesRequestID m_resourcesRequestID;

  EVENT_HANDLER1(L2IndividualRecruit, onResourcesAllocated, BWAPI::Unit*) m_resourceAllocatedEventHandler;
  EVENT_HANDLER1(L2IndividualRecruit, onUnitCreated, BWAPI::Unit*) m_unitCreatedEventHandler;
  EVENT_HANDLER1(L2IndividualRecruit, onUnitCompleted, BWAPI::Unit*) m_unitCompletedEventHandler;
  EVENT_HANDLER(L2IndividualRecruit, onUnitDestroyed) m_unitDestroyedEventHandler;
  EVENT_HANDLER(L2IndividualRecruit, onBuildingDestroyed) m_buildingDestroyedEventHandler;
};

