#include "stdafx.h"
#include "../Layer2Managers.h"
#include "L2Research.h"

void L2Research::executeSpecificImpl()
{
  L2TacticalAction::executeSpecificImpl();
}

void L2Research::cancelSpecificImpl()
{
  L2TacticalAction::cancelSpecificImpl();

  if(m_currentState == WAITING_FOR_RESOURCES)
  {
    Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
  }
  else if(m_currentState == WAITING_TO_FINISH)
  {
    assert(m_researchProcess != NULL);
    m_researchProcess->getTerminateEvent().unsubscribe(m_researchTerminateSubscriberID);
    m_researchProcess->cancel();
  }
  terminate(Cancelled);
}

void L2Research::terminateSpecificImpl()
{
  L2TacticalAction::terminateSpecificImpl();
  Layer2Managers::getInstance().cancelResourcesRequest(m_resourcesRequestID);
  if (m_researchProcess != NULL)
    m_researchProcess->getTerminateEvent().unsubscribe(m_researchTerminateSubscriberID);
  if(m_building != NULL)
    dynamic_cast<Layer2ManagersUnitData&>(Spar->getUnitManager().getUnitData(m_building)).setInUseFlag(false);
  m_building = NULL;
}

void L2Research::onResearchProcessTerminate(void*)
{
  m_researchProcess->getTerminateEvent().unsubscribe(m_researchTerminateSubscriberID);
  if(m_building != NULL)
    dynamic_cast<Layer2ManagersUnitData&>(Spar->getUnitManager().getUnitData(m_building)).setInUseFlag(false);
  m_building = NULL;
  PerceptualState::ResearchInProgress* research = m_researchProcess;
  m_researchProcess = NULL;
  terminate(research->getTerminationMode());
}

void L2UpgradeResearch::executeSpecificImpl()
{
  L2Research::executeSpecificImpl(); 
  Layer2Managers::getInstance().requestResources(
    m_priority, 
    ActionResources(Resources(m_upgrade->getUpgrade(),m_upgrade->getLevel()),Spar->getPerceptualState().getMap(), m_upgrade->getUpgrade().whatUpgrades()), 
    m_resourcesRequestID);
  m_currentState = WAITING_FOR_RESOURCES;
}

void L2UpgradeResearch::onResourcesAllocated(void* data, BWAPI::Unit* b)
{
  if(b==NULL || b->getType()!= m_upgrade->getUpgrade().whatUpgrades())
  {
    //SHUT
    //DOWN
    //EVERYTHING!
    cancel();
  }
  else
  {
    m_building = b;
    if(m_building->getRemainingResearchTime() == 0 &&
      m_building->getRemainingUpgradeTime() == 0 &&
      m_building->getRemainingTrainTime() == 0)
    {
      Resources oldResources = Resources(BWAPI::Broodwar->self());
      m_building->upgrade(m_upgrade->getUpgrade());
      assert(oldResources == Resources(BWAPI::Broodwar->self()) + Resources(m_upgrade->getUpgrade(),m_upgrade->getLevel()));

      //TELL Perceptual that an upgrade as been started
      m_researchProcess = Spar->getPerceptualState().onUpgradeStarted(this,m_upgrade,m_building);
      Layer2Managers::getInstance().releaseResources(Resources(m_upgrade->getUpgrade(),m_upgrade->getLevel()));      
      m_researchTerminateSubscriberID = m_researchProcess->getTerminateEvent().subscribe(&m_researchTerminateEventHandler, NULL);
      m_currentState = WAITING_TO_FINISH;
    }
    else
    {
      //SHUT
      //DOWN
      //EVERYTHING!
      cancel();
    }
  }
}

void L2TechResearch::executeSpecificImpl()
{
  L2Research::executeSpecificImpl(); 
  Layer2Managers::getInstance().requestResources(
    m_priority, 
    ActionResources(Resources(m_tech->getTech()),Spar->getPerceptualState().getMap(), m_tech->getTech().whatResearches()), 
    m_resourcesRequestID);
  m_currentState = WAITING_FOR_RESOURCES;
}

void L2TechResearch::onResourcesAllocated(void* data, BWAPI::Unit* b)
{
  if(b==NULL || b->getType()!= m_tech->getTech().whatResearches())
  {
    //SHUT
    //DOWN
    //EVERYTHING!
    cancel();
  }
  else
  {
    m_building = b;
    if(m_building->getRemainingResearchTime() == 0 &&
      m_building->getRemainingUpgradeTime() == 0 &&
      m_building->getRemainingTrainTime() == 0)
    {
      Resources oldResources = Resources(BWAPI::Broodwar->self());
      m_building->research(m_tech->getTech());
      assert(oldResources == Resources(BWAPI::Broodwar->self()) + Resources(m_tech->getTech()));
    
      //TELL Perceptual that a tech as been started
      m_researchProcess = Spar->getPerceptualState().onResearchStarted(this,m_tech,m_building);
      Layer2Managers::getInstance().releaseResources(Resources(m_tech->getTech()));
      m_researchTerminateSubscriberID = m_researchProcess->getTerminateEvent().subscribe(&m_researchTerminateEventHandler, NULL);
      m_currentState = WAITING_TO_FINISH;
    }
    else
    {
      //SHUT
      //DOWN
      //EVERYTHING!
      cancel();
    }
  }
}
