#pragma once
#include "../../Interface/Actions/Research.h"
#include "L2TacticalAction.h"

class L2Research : public virtual layer2::Research, public L2TacticalAction
{
public:
  L2Research(unsigned int priority)
    : Research(priority)
    , L2TacticalAction(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_researchTerminateEventHandler(*this)
    , m_resourceAllocatedEventHandler(*this)
    , m_resourcesRequestID(m_resourceAllocatedEventHandler)
    , m_researchProcess(NULL)
    , m_currentState(UNINITIALIZED)
    , m_building(NULL)
  {
  }

protected:
  virtual void executeSpecificImpl();
  virtual void cancelSpecificImpl();
  virtual void terminateSpecificImpl();
  virtual void onResourcesAllocated(void* data, BWAPI::Unit* b) = 0;

  void onResearchProcessTerminate(void* data);  
  
  enum ResearchState
  {
    UNINITIALIZED,
    WAITING_FOR_RESOURCES,
    ACTION_SENT,
    WAITING_TO_FINISH,
    TERMINATED
  };

  ResearchState m_currentState;
  BWAPI::Unit* m_building;
  Layer2ManagersBase::ResourcesRequestID m_resourcesRequestID;
  PerceptualState::ResearchInProgress* m_researchProcess;

  EVENT_HANDLER1(L2Research, onResourcesAllocated, BWAPI::Unit*) m_resourceAllocatedEventHandler;

  PerceptualState::ResearchInProgress::OnTerminate::SubscriberID m_researchTerminateSubscriberID;
  EVENT_HANDLER(L2Research, onResearchProcessTerminate) m_researchTerminateEventHandler;
};

class L2UpgradeResearch : public virtual layer2::UpgradeResearch, public L2Research
{
public:
  L2UpgradeResearch(unsigned int priority, UpgradeTypeTech* upgrade)
    : UpgradeResearch(priority, upgrade)
    , L2Research(priority)
    , Research(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_upgrade(upgrade)
  {    
  }

protected:
  UpgradeTypeTech* m_upgrade;

  virtual void executeSpecificImpl();
  virtual void onResourcesAllocated(void* data, BWAPI::Unit* b);
};

class L2TechResearch : public virtual layer2::TechResearch, public L2Research
{
public:
  L2TechResearch(unsigned int priority, TechTypeTech* tech)
    : TechResearch(priority, tech)
    , L2Research(priority)
    , Research(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_tech(tech)
  {
  }

protected:
  TechTypeTech* m_tech;

  virtual void executeSpecificImpl();
  virtual void onResourcesAllocated(void* data, BWAPI::Unit* b);
};
