#include "stdafx.h"
#include "../Layer2Managers.h"
#include "L2Scout.h"
#include <math.h>

void L2Scout::executeSpecificImpl()
{
  if(m_unit==NULL){
    L2GroupTacticalAction::executeSpecificImpl();

    const PersistentUnitGroup* group = getGroup();
    BWAPI::Unit* scout = NULL;

    int smallestDist = boost::numeric::bounds<int>::highest();
    BWAPI::Unit* bestUnit;
    std::set<BWAPI::Unit*> unitSet = BWAPI::Broodwar->self()->getUnits();
    for(std::set<BWAPI::Unit*>::const_iterator it = unitSet.begin(); it!=unitSet.end(); ++it)
    {
      //TODO: is there anything else to test ?
      if((*it)->exists() && !(*it)->getType().isBuilding() && (*it)->isCompleted() && 
        (dynamic_cast<Layer2UnitData&>(Spar->getUnitManager().getUnitData(*it)).getCurrentAction() == NULL ||
        dynamic_cast<Layer2UnitData&>(Spar->getUnitManager().getUnitData(*it)).getCurrentAction()->getPriority() < m_priority) &&
        ((*it)->getType().isFlyer() || (*it)->getPosition().hasPath(m_location->getCurrentCenter())))
      {
        int dist;
        if((*it)->getType().isFlyer())
        {
          BWAPI::Position diff = ((*it)->getPosition() - m_location->getCurrentCenter());
          dist = static_cast<int>(sqrt((double)(diff.x()*diff.x() + diff.y()*diff.y())));
        }
        else
        {
          dist = (*it)->getPosition().getApproxDistance(m_location->getCurrentCenter());
        }

        if(dist<smallestDist)
        {
          bestUnit = (*it);
          smallestDist = dist;
        }
      }
    }

    assert(bestUnit!=NULL);
    bool reserved = addUnit(bestUnit);
    // TODO: pick another one
    assert(reserved);

    for(PersistentUnitGroup::const_iterator it = getGroup()->begin();it!=getGroup()->end();it++){
      BWAPI::Unit* worker = (*it);
      m_scoutFSM = new ScoutFSM(this, 100, m_location, worker);
      m_scoutArrivedSubscriberId = m_scoutFSM->getScoutArrivedEvent().subscribe(&m_scoutArrivedEventHandler, NULL);
      m_scoutFSM->execute();
    }
  }
  else if(m_unit->getType() == BWAPI::UnitTypes::Protoss_Observer)
  {
    m_obsScout = new ObsScout(this, 100, m_location, m_unit);
    m_scoutArrivedSubscriberId = m_obsScout->getScoutArrivedEvent().subscribe(&m_scoutArrivedEventHandler, NULL);
    m_obsScout->execute();
  }
}

void L2Scout::onUnitAddedToGroup(void*, BWAPI::Unit* unit)
{
  // Should only happen when reserveUnit() is called
}

void L2Scout::onUnitRemovedFromGroup(void*, BWAPI::Unit* unit)
{
  terminate(AllUnitsDestroyed);
}

void L2Scout::cancelSpecificImpl()
{
  L2GroupTacticalAction::cancelSpecificImpl();
  terminate(Cancelled);
}

void L2Scout::terminateSpecificImpl()
{
  if(m_unit==NULL)
  {
    m_scoutFSM->getScoutArrivedEvent().unsubscribe(m_scoutArrivedSubscriberId);
  }
  else if(m_obsScout!=NULL)
  {
    m_obsScout->getScoutArrivedEvent().unsubscribe(m_scoutArrivedSubscriberId);
  }
  L2GroupTacticalAction::terminateSpecificImpl();
  // NULL case is handled by addZombieComponent()
}

void L2Scout::onScoutArrived(void* data)
{
  if(m_unit==NULL){
    m_scoutArrivedEvent.raise(this, m_location,NULL);
  }
  else
  {
    m_scoutArrivedEvent.raise(this, m_location,m_unit);
    terminate(Success);
  }
  SPAR_LOG(LogTypes::LAYER_1_DECISION, "Scout has arrived!");
}
