#pragma once
#include "../../Interface/Actions/Scout.h"
#include "L2GroupTacticalAction.h"
#include "../../../Layer1Reactive/FSM/Behaviors/Scout.h"
#include "../../../Layer1Reactive/FSM/Behaviors/ObsScout.h"

class L2Scout : public virtual layer2::Scout, public L2GroupTacticalAction
{
public:
  L2Scout(unsigned int priority, StaticLocation* location)
    : Scout(priority, location)
    , L2GroupTacticalAction(priority)
    , GroupTacticalAction(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_scoutFSM(*this)
    , m_obsScout(*this)
    , m_scoutArrivedEventHandler(*this)
  {
    m_unit = NULL;
  }

  L2Scout(unsigned int priority, StaticLocation* location, BWAPI::Unit* scout)
    : Scout(priority, location)
    , L2GroupTacticalAction(priority)
    , GroupTacticalAction(priority)
    , TacticalAction(priority)
    , Process(Spar)
    , m_scoutFSM(*this)
    , m_obsScout(*this)
    , m_scoutArrivedEventHandler(*this)
    , m_unit(scout)
  {
    assert(scout->getType()==BWAPI::UnitTypes::Protoss_Observer);
  }

  void onScoutArrived(void* data);

protected:
  virtual void executeSpecificImpl();
  virtual void cancelSpecificImpl();
  virtual void terminateSpecificImpl();

  virtual void onUnitAddedToGroup(void*, BWAPI::Unit* unit);
  virtual void onUnitRemovedFromGroup(void*, BWAPI::Unit* unit);

  Event<boost::tuples::tuple<>>::SubscriberID m_scoutArrivedSubscriberId;
  EVENT_HANDLER(L2Scout, onScoutArrived) m_scoutArrivedEventHandler;

private:
  ProcessContainer<ScoutFSM> m_scoutFSM;
  ProcessContainer<ObsScout> m_obsScout;
  BWAPI::Unit* m_unit;
};
