#include "stdafx.h"
#include "Layer2Managers.h"
#include "../Interface/Common.h"
#include "ActionsImplementation/Common.h"

using namespace layer2;

Layer2Managers* Layer2Managers::s_instance;

bool Layer2Managers::includedIn(Resources r1, Resources r2)
{
  return r1.getGas() <= r2.getGas() && r1.getMinerals() <= r2.getMinerals()
      && r1.getSupply() <= r2.getSupply();
}

Layer2Managers::Layer2Managers(Layer2Cueing<Layer1GameState>& layer2Cueing, Layer1FSM& layer1FSM)
: Layer2Actions(layer2Cueing, layer1FSM)
, m_requestCounter(0)
//, m_actionEndedEventHandler(*this)
{
  s_instance = this;
}

Layer2Managers::~Layer2Managers()
{}

void Layer2Managers::onStart()
{
  Layer2Actions::onStart();
  NewBuildingPlacer::onStart();
}

void Layer2Managers::onFrame()
{
  Layer2Actions::onFrame();

  //TODO: This should not be called every frame !
  
  Resources remainingResource = Resources(BWAPI::Broodwar->self()) - m_reservedResources;
  m_resourcesRequests.do_while(UseResources(m_resourcesRequests, remainingResource, this));

  //check for actions to be updated
  // Do Nothing ? managed by the Scheduler ?
  //for(std::list<L2TacticalAction*>::iterator it = m_actionsWaitingForFeedback.begin() ; it!=m_actionsWaitingForFeedback.end() ; it++)
  //{
  //  //update action
  //}

  //check for new action and start them
  //for(std::list<L2TacticalAction*>::iterator it = m_newActions.begin() ; it!=m_newActions.end() ; it++)
  //{
  //  (*it)->execute();
  //}
  //m_newActions.clear();

  //for(std::list<L2TacticalAction*>::iterator it = m_actionsEnded.begin() ; it!=m_actionsEnded.end() ; it++)
  //{
  //  //Clear those actions
  //}

#ifdef DRAW_ON_SCREEN
  if (Spar->getConfiguration().displayOwnUnitsActions())
  {
    const std::set<BWAPI::Unit*>& units=BWAPI::Broodwar->self()->getUnits();
    for(std::set<BWAPI::Unit*>::const_iterator i = units.begin(); i != units.end(); ++i) {
      UnitData* data = reinterpret_cast<UnitData*>((*i)->getClientInfo());
      if (data != NULL)
      {
        const GroupTacticalAction* action = dynamic_cast<Layer2UnitData&>(Spar->getUnitManager().getUnitData(*i)).getCurrentAction();
        if (action != NULL)
        {
          const int x = (*i)->getPosition().x();
          const int y = (*i)->getPosition().y();
          const char activeColor = '\x17';
          std::string actionName = action->getName();
          BWAPI::Broodwar->drawTextMap(x, y, "%c%s: %d", activeColor, actionName.c_str(), action->getPriority());
        }
      }
    }
  }
#endif
}

void Layer2Managers::onEnd()
{
  Layer2Actions::onEnd();

  m_resourcesRequests.do_while(ClearResourcesRequestsFunctor());

  NewBuildingPlacer::onEnd();
}

void Layer2Managers::onSendText(std::string text)
{
  //...
}

//void Layer2Managers::onActionExecuted(L2TacticalAction* action)
//{
//  // No duplicated actions
//  assert(std::find(m_currentActions.begin(), m_currentActions.end(), action) == m_currentActions.end());
//
//  action->getTerminateEvent().subscribe(&m_actionEndedEventHandler, action);
//  action->setActionID(m_currentActions.insert(m_currentActions.end(), action));
//}

//void Layer2Managers::onActionEnded(void* data)
//{
//  const L2TacticalAction* l2Action = reinterpret_cast<const L2TacticalAction*>(data);
//  assert(std::find(m_currentActions.begin(), m_currentActions.end(), l2Action) != m_currentActions.end());
//  assert(l2Action->isActionIDValid());
//  m_currentActions.erase(l2Action->getActionID());
//}

void Layer2Managers::requestResources(unsigned int priority, const ActionResources& resources, Layer2Managers::ResourcesRequestID& id)
{
  // TODO: automatic mechanism to erase request resource when its owner is deleted (or terminates)

  ResourcesRequestsType::iterator pos = m_resourcesRequests.insert(ResourcesRequest(++m_requestCounter, priority, resources, id));
  id.posInContainer = pos;
  id.isValid = true;
}

void Layer2Managers::cancelResourcesRequest(ResourcesRequestID& id)
{
  if (id.isValid)
  {
    id.isValid = false;
    m_resourcesRequests.erase(id.posInContainer);
  }
}

void Layer2Managers::changeResourcesRequestPriority(ResourcesRequestID& id, unsigned int newPriority)
{
  assert(id.isValid);

  ResourcesRequest rr = *id.posInContainer;
  m_resourcesRequests.erase(id.posInContainer);
  rr.priority = newPriority;
  m_resourcesRequests.insert(rr);
}

void Layer2Managers::reserveResource(const Resources& resources)
{
  m_reservedResources += resources;
}

void Layer2Managers::releaseResources(const Resources& resources)
{
  m_reservedResources -= resources;
  assert(m_reservedResources.getMinerals() >= 0);
  assert(m_reservedResources.getGas() >= 0);
  assert(m_reservedResources.getSupply() >= 0);
}

Resources Layer2Managers::getLowestPriorityAvailableResources() const
{
  return getAvailableResources(boost::numeric::bounds<int>::lowest());
}

Resources Layer2Managers::getHighestPriorityAvailableResources() const
{
  return Resources(BWAPI::Broodwar->self()) - m_reservedResources;
}

Resources Layer2Managers::getAvailableResources(unsigned int priority) const
{
  //
  // Returns the total amount of resources which are currently available for a process which would like
  // to execute an action of this priority.
  // This corresponds to the amount of resources currently registered (Broodwar->self()), 
  // - minus the resources not yet deducted due to latency problems (minerals count being updated 2-3 frames later...)
  // - minus the resources reserved for higher-priority actions
  //
  // Note: if Build actions cannot be cancelled once the probe is sent (even if higher priority actions need resources), 
  //       the amount of resources needed for all these pending Build actions must still be deduced, regardless of their priority.
  //
  
  return m_resourcesRequests.do_while(AvailableResourcesFunctor(Resources(BWAPI::Broodwar->self()) - m_reservedResources, priority)).m_resources;
}

size_t Layer2Managers::getNbAvailableBuildings(const BWAPI::UnitType buildingType, const StaticLocation& location) const
{
  size_t nbAvailable = 0;
  std::pair<StaticLocation::own_unit_type_const_iterator, StaticLocation::own_unit_type_const_iterator> buildingList =
    location.getOwnUnits(buildingType);
  for (StaticLocation::own_unit_const_iterator buildingIt = buildingList.first; buildingIt != buildingList.second; ++buildingIt)
  {
    BWAPI::Unit* unit = (*buildingIt)->getUnit();
    if(unit->isCompleted())
    {
      Layer2ManagersUnitData& unitData = dynamic_cast<Layer2ManagersUnitData&>(Spar->getUnitManager().getUnitData(unit));
      if(!unitData.isInUse())
        ++nbAvailable;
    }
  }
  return nbAvailable;
}
