#pragma once
#include "../../../Utils/SafeMultiset.h"

// This class is only used to avoid circular references

class L2TacticalAction;

class Layer2ManagersBase
{
public:
  struct ResourcesRequest;
  struct SortByPriorityThenID
  {
    bool operator()(const ResourcesRequest& rr1, const ResourcesRequest& rr2) const;
  };
  typedef SafeMultiset<ResourcesRequest, SortByPriorityThenID> ResourcesRequestsType;

  struct ResourcesRequestID
  {
    ResourcesRequestID(EventHandler1<BWAPI::Unit*>& p_resourcesAllocatedEventHandler)
      : eventHandler(p_resourcesAllocatedEventHandler)
      , isValid(false)
    {}
    EventHandler1<BWAPI::Unit*>& eventHandler;
    bool isValid;
    ResourcesRequestsType::iterator posInContainer;
  };

  struct ResourcesRequest
  {
    ResourcesRequest(unsigned int p_id, unsigned int p_priority, const ActionResources& p_resources, ResourcesRequestID& p_resourcesId)
      : id(p_id)
      , priority(p_priority)
      , actionResources(p_resources)
      , resourcesId(p_resourcesId)
    {}
    unsigned int id;
    unsigned int priority;
    const ActionResources actionResources;
    ResourcesRequestID& resourcesId;
  };

  //typedef ResourcesRequestsType::iterator ResourcesRequestID;
  //typedef std::list<L2TacticalAction*>::iterator ActionID;
};

inline bool Layer2ManagersBase::SortByPriorityThenID::operator()(const Layer2ManagersBase::ResourcesRequest& rr1, 
                                                           const Layer2ManagersBase::ResourcesRequest& rr2) const
{
  return rr1.priority > rr2.priority || (rr1.priority == rr2.priority && rr1.id < rr2.id);
}
