#include "stdafx.h"
#include "NewBuildingPlacer.h"

int PYLON_BUILD_RADIUS = 10;

NewBuildingPlacer* NewBuildingPlacer::s_newBuildingPlacer;

NewBuildingPlacer::NewBuildingPlacer()
{
  m_distance = 1;
  needPylon = true;
  reservedEmplacements = new bool*[(BWAPI::Broodwar->mapWidth()+1)];
  for(int i = 0;i<=BWAPI::Broodwar->mapWidth();i++)
  {
    reservedEmplacements[i]= new bool[(BWAPI::Broodwar->mapHeight()+1)];
    for(int j = 0; j<=BWAPI::Broodwar->mapHeight();j++)
    {
      reservedEmplacements[i][j] = false;
    }
  }
}

NewBuildingPlacer::~NewBuildingPlacer()
{
  for(int i=0; i<=BWAPI::Broodwar->mapWidth();i++)
  {
    delete [] reservedEmplacements[i];
  }
  delete[] reservedEmplacements;
}



void NewBuildingPlacer::setBuildingDistance(int dist)
{
  m_distance = dist;
}

bool NewBuildingPlacer::needsAdditionalPylons()
{
  // TODO: keep a flag if a build actions terminates because there's no room left
  return needPylon;
}

void NewBuildingPlacer::setPylonFalse()
{
  needPylon = false;
}

BWAPI::TilePosition NewBuildingPlacer::getBuildingLocation(BWAPI::Unit* builder, BWAPI::UnitType toBuild,StaticLocation* location)
{
  int dist = m_distance;
  int maxDist = PYLON_BUILD_RADIUS;

  //Find a place for a pylon
  if(toBuild==BWAPI::UnitTypes::Protoss_Pylon)
  {
    dist = 2;
    maxDist = 15;

    while(nexusWithoutPylon.begin() != nexusWithoutPylon.end())
    {
      if( (*nexusWithoutPylon.begin()) != NULL && (*nexusWithoutPylon.begin())->exists())
      {
        BWAPI::TilePosition tempTP = getBuildLocationNear(builder, (*nexusWithoutPylon.begin())->getTilePosition(),toBuild,dist,maxDist);
        if(tempTP!=BWAPI::TilePositions::None)
        {
          return tempTP;
        }
      }
      nexusWithoutPylon.erase(nexusWithoutPylon.begin());
    }

    //We need to place the first pylon
    while(dist>=1)
    {
      std::pair<StaticLocation::own_building_const_iterator, StaticLocation::own_building_const_iterator> building = location->getOwnBuildings();

      if(building.first == building.second)
      {
        //the base is empty how do we know where to build our pylon location
        BWAPI::TilePosition tempTP = getBuildLocationNear(builder, BWAPI::TilePosition(location->getCurrentCenter().x()/TILE_SIZE,location->getCurrentCenter().y()/TILE_SIZE),toBuild,dist,maxDist);
      }

      //The base has no pylon but has building. Build Pylon close
      for( ; building.first != building.second ; building.first = ++(building.first) )
      {
        BWAPI::Unit* unit = (*building.first)->getUnit();
        assert(unit->getPlayer() == BWAPI::Broodwar->self());
        assert(unit->getType() != BWAPI::UnitTypes::Protoss_Probe);
        assert(unit->getType() != BWAPI::UnitTypes::Resource_Mineral_Field);
        assert(unit->getType() != BWAPI::UnitTypes::Resource_Vespene_Geyser);
        BWAPI::TilePosition tempTP = getBuildLocationNear(builder, unit->getTilePosition(),toBuild,dist,maxDist);
        if(tempTP != BWAPI::TilePositions::None)
        {
          return tempTP;
        }
      }
      dist--;
    }
  }
  
  //Find a place for a nexus
  else if(toBuild==BWAPI::UnitTypes::Protoss_Nexus)
  {
    if(dynamic_cast<BaseLocation*> (location))
    {
      BaseLocation* bl = dynamic_cast<BaseLocation*> (location);
      const BWTA::BaseLocation* base = bl->getBWTABaseLocation();
      return base->getTilePosition();
    }else
    {
      //Why a nexus here
      return BWAPI::TilePositions::None;
    }
  }
  
  //Find a place for an assimilator
  else if(toBuild==BWAPI::UnitTypes::Protoss_Assimilator)
  {    
    if(dynamic_cast<BaseLocation*> (location))
    {
      BaseLocation* bl = dynamic_cast<BaseLocation*> (location);
      const BWTA::BaseLocation* base = bl->getBWTABaseLocation();
      const std::set<BWAPI::TilePosition>& geysers = base->getGeysersPosition();
      for(std::set<BWAPI::TilePosition>::const_iterator it = geysers.begin(); it != geysers.end() ; ++it)
      {
        BWAPI::TilePosition geyser = (*it);
        // TODO: getUnitsOnTile returns empty if the geyser is not visible (probe must go near the geyser first)
        const std::set<BWAPI::Unit*>& units = BWAPI::Broodwar->getUnitsOnTile(geyser.x(),geyser.y());
        bool canBuild = true;
        for(std::set<BWAPI::Unit*>::const_iterator i = units.begin(); (i!=units.end() && canBuild) ; ++i)
        {
          if((*i)->getType() != BWAPI::UnitTypes::Resource_Vespene_Geyser)
          {
            canBuild = false;
          }
        }
        if(canBuild)
        {
          return geyser;
        }
      }
      //If here there is no empty geyser
      //Send ERROR
    }else
    {
      //Why an assimilator here
      return BWAPI::TilePositions::None;
    }
    //Must get a list of all empty gas in location
    return BWAPI::TilePositions::None;
  }
  
  //This is neither a pylon nor a nexus nor an assimilator. We need power.
  else
  {
    while(dist>=1)
    {
      std::pair<StaticLocation::own_unit_type_const_iterator, StaticLocation::own_unit_type_const_iterator> building = location->getOwnUnits(BWAPI::UnitTypes::Protoss_Pylon);

      if(building.first == building.second)
      {
        needPylon = true;
        return BWAPI::TilePositions::None;

      }

      for( ; building.first != building.second ; building.first = ++(building.first) )
      {
        BWAPI::Unit* unit = (*building.first)->getUnit();
        if(unit->isCompleted())
        {
          BWAPI::TilePosition tempTP = getBuildLocationNear(builder, unit->getTilePosition(),toBuild,dist,maxDist);
          if(tempTP != BWAPI::TilePositions::None)
          {
            //Maybe should reserve the position here
            return tempTP;
          }
        }
      }
      dist--;
    }
  }
  needPylon = true;
  return BWAPI::TilePositions::None;
}

bool NewBuildingPlacer::reservePosition(BWAPI::TilePosition tp, BWAPI::UnitType toBuild)
{
  //TODO
  //Check if position is resevable
  if(!canBuildHere(NULL,tp,toBuild))
    return false;
  //Add it to the reserve list
  for(int x = tp.x() ; x < (tp.x() + toBuild.tileWidth() - 1) ; x++)
  {
    for(int y = tp.y() ; y < (tp.y() + toBuild.tileHeight() - 1) ; y++)
    {
      reservedEmplacements[x][y] = true;
    }
  }
  return true;
}

bool NewBuildingPlacer::unreservePosition(BWAPI::TilePosition tp, BWAPI::UnitType toBuild)
{
  for(int x = tp.x() ; x < (tp.x() + toBuild.tileWidth() - 1) ; x++)
  {
    for(int y = tp.y() ; y < (tp.y() + toBuild.tileHeight() - 1) ; y++)
    {
      reservedEmplacements[x][y] = false;
    }
  }
  return true;
}

bool NewBuildingPlacer::overReservedPosition(BWAPI::TilePosition tp,BWAPI::UnitType toBuild)
{
  if(!tp.isValid())
    return true;
  for(int x = tp.x() ; x < (tp.x() + toBuild.tileWidth() - 1) ; x++)
  {
    for(int y = tp.y() ; y < (tp.y() + toBuild.tileHeight() - 1) ; y++)
    {
      if(x<0 || x>=BWAPI::Broodwar->mapWidth() || y<0 || y>=BWAPI::Broodwar->mapHeight())
        return true;
      if(reservedEmplacements[x][y])
        return true;
    }
  }
  return false;
}

bool NewBuildingPlacer::canBuildHere(BWAPI::Unit* builder, BWAPI::TilePosition position, BWAPI::UnitType type)
{
  bool isFree = (!overReservedPosition(position,type));
  bool canBuild = (BWAPI::Broodwar->canBuildHere(builder,position,type,true));
  return (isFree && canBuild);
}


//TODO: CHECK THIS 
bool NewBuildingPlacer::canBuildHereWithSpace(BWAPI::Unit* builder, BWAPI::TilePosition position, BWAPI::UnitType type, int buildDist)
{
  if(!canBuildHere(builder,position,type))
    return false;
  if(buildDist<1)
    return true;
  else if(!canWalkAround(position,type))
    return false;

  //test if space around building is OK
  int width = type.tileWidth();
  int height = type.tileHeight();
  int buildingTopX = position.x();
  int buildingTopY = position.y();

  //Test the top part
  int topX = buildingTopX + width - 1 + buildDist;
  int topY = buildingTopY - 1;
  int botX = buildingTopX - buildDist;
  int botY = buildingTopY - buildDist;

  for(int i = botX ; i <= topX ; i++)
  {
    for(int j = botY ; j <= topY ; j++)
    {
      BWAPI::TilePosition tp = BWAPI::TilePosition(i,j);
      if(tileHasBuilding(tp))
        return false;
    }
  }

  //Test the bottom part
  topX = buildingTopX + width - 1 + buildDist;
  topY = buildingTopY + height - 1 + buildDist;
  botX = buildingTopX - buildDist;
  botY = buildingTopY + height;

  for(int i = botX ; i <= topX ; i++)
  {
    for(int j = botY ; j <= topY ; j++)
    {
      BWAPI::TilePosition tp = BWAPI::TilePosition(i,j);
      if(tileHasBuilding(tp))
        return false;
    }
  }

  //Test the left part
  topX = buildingTopX - 1;
  topY = buildingTopY + height - 1;
  botX = buildingTopX - buildDist;
  botY = buildingTopY;

  for(int i = botX ; i <= topX ; i++)
  {
    for(int j = botY ; j <= topY ; j++)
    {
      BWAPI::TilePosition tp = BWAPI::TilePosition(i,j);
      if(tileHasBuilding(tp))
        return false;
    }
  }

  //Test the right part
  topX = buildingTopX + width - 1 + buildDist;
  topY = buildingTopY + height - 1 ;
  botX = buildingTopX + width;
  botY = buildingTopY;

  for(int i = botX ; i <= topX ; i++)
  {
    for(int j = botY ; j <= topY ; j++)
    {
      BWAPI::TilePosition tp = BWAPI::TilePosition(i,j);
      if(tileHasBuilding(tp))
        return false;
    }
  }

  return true;
}

bool NewBuildingPlacer::canWalkAround(BWAPI::TilePosition position, BWAPI::UnitType type)
{
  int x = position.x()-1;
  int xmax = position.x() + type.tileWidth() + 1;
  int y = position.y()-1;
  int ymax = position.y() + type.tileHeight() + 1;
  for(int i = x ; i<=xmax ; i++)
  {
    for(int j = y ; j<=ymax ; j++)
    {
      if(!tileIsWalkable(BWAPI::TilePosition(i,j)))
      {
        return false;
      }
      else//Check for gas or mineral
      {
        std::set<BWAPI::Unit*> units = BWAPI::Broodwar->getUnitsOnTile(i,j);
        for(std::set<BWAPI::Unit*>::iterator i = units.begin() ; i!=units.end() ; ++i)
        {
          if((*i)->getType() == BWAPI::UnitTypes::Protoss_Assimilator ||
            (*i)->getType() == BWAPI::UnitTypes::Resource_Mineral_Field ||
            (*i)->getType() == BWAPI::UnitTypes::Resource_Mineral_Field_Type_2 ||
            (*i)->getType() == BWAPI::UnitTypes::Resource_Mineral_Field_Type_3 ||
            (*i)->getType() == BWAPI::UnitTypes::Resource_Vespene_Geyser)
          {
            return false;
          }
        }
      }
    }
  }
  return true;
}

bool NewBuildingPlacer::tileIsWalkable(BWAPI::TilePosition position)
{
  //Check if the tile is walkable
  int x1=(position.x()*4);
  int x2=((position.x()*4)+3);
  int y1=(position.y()*4);
  int y2=((position.y()*4)+3);

  return !((!BWAPI::Broodwar->isWalkable(x1, y1)) || 
    (!BWAPI::Broodwar->isWalkable(x2, y1)) ||
    (!BWAPI::Broodwar->isWalkable(x2, y2)) || 
    (!BWAPI::Broodwar->isWalkable(x1, y2)));
}

bool NewBuildingPlacer::tileHasBuilding(BWAPI::TilePosition position)
{
  if(!position.isValid())
    //One of the tile around the building is not valid, The build distance is not respected.
    return false;
  const std::set<BWAPI::Unit*>& unitOnTile = BWAPI::Broodwar->getUnitsOnTile(position.x(),position.y());
  if(!unitOnTile.empty())
  {
    for(std::set<BWAPI::Unit*>::const_iterator it = unitOnTile.begin() ; it != unitOnTile.end() ; ++it)
    {
      //There is a unit on the tile that is not the builder
      if(!(*it)->getType().canMove() && (*it)->getPlayer()==BWAPI::Broodwar->self() )
      {
        //the unit cant move. Must be a building, so the distance between two building is not respected.
        return true;
      }
    }
  }
  if(reservedEmplacements[position.x()][position.y()])
  {
    //The tile is reserved for another building
    return true;
  }
  return false;
}

BWAPI::TilePosition NewBuildingPlacer::getBuildLocationNear(BWAPI::Unit* builder, BWAPI::TilePosition position, BWAPI::UnitType type, int buildDist, int maxDistance)
{
  int x      = position.x();
  int xMax   = x + maxDistance;
  int y      = position.y();
  int yMax   = y + maxDistance;
  int length = 1;
  int j      = 0;
  bool first = true;
  int dx     = 0;
  int dy     = 1;

  while ( x<=xMax || y<=yMax ) //We'll ride the spiral to the end
  {
    //if we can build here, return this tile position
    if (x >= 0 && x < BWAPI::Broodwar->mapWidth() && y >= 0 && y < BWAPI::Broodwar->mapHeight())
      if (this->canBuildHereWithSpace(builder, BWAPI::TilePosition(x, y), type, buildDist))
        return BWAPI::TilePosition(x, y);

    //otherwise, move to another position
    x = x + dx;
    y = y + dy;
    //count how many steps we take in this direction
    j++;
    if (j == length) //if we've reached the end, its time to turn
    {
      //reset step counter
      j = 0;

      //Spiral out. Keep going.
      if (!first)
        length++; //increment step counter if needed

      //first=true for every other turn so we spiral out at the right rate
      first =! first;

      //turn counter clockwise 90 degrees:
      if (dx == 0)
      {
        dx = dy;
        dy = 0;
      }
      else
      {
        dy = -dx;
        dx = 0;
      }
    }
    //Spiral out. Keep going.
  }

  //if (buildDist>1)
  //Not sure about that
  //return getBuildLocationNear(builder, position, type, buildDist-1,maxDistance);
  return BWAPI::TilePositions::None;
}

//Should not be here
bool NewBuildingPlacer::isUsed(BWAPI::Unit* unit)
{
  return !(std::find(usedWorker.begin(),usedWorker.end(),unit) == usedWorker.end());
}

void NewBuildingPlacer::useWorker(BWAPI::Unit* unit)
{
  usedWorker.push_back(unit);
}

void NewBuildingPlacer::releaseWorker(BWAPI::Unit* unit)
{
  usedWorker.remove(unit);
}

void NewBuildingPlacer::addNewNexus(BWAPI::Unit* nexus)
{
  nexusWithoutPylon.insert(nexusWithoutPylon.end(),nexus);
}