#pragma once
#include "../../PerceptualState/Techs/Techs.h"
#include "../Layer3Tactical.h"
#include "Layer3ScriptTypes.h"
#include "Locations/Common.h"
#include "Techs/L3Tech.h"
#include "Tasks/BuildCannonsTask.h"
#include "Tasks/BuildExpansionsTask.h"
#include "Tasks/BuildGatewaysTask.h"
#include "Tasks/BuildPylonsTask.h"
#include "Tasks/ManageArmyTask.h"
#include "Tasks/ManageResourcesTask.h"
#include "Tasks/ProgressTechTask.h"
#include "Tasks/RecruitArmyTask.h"
#include "Tasks/ScoutTask.h"

class Layer3Script : public Layer3Tactical<Layer3TacticalRecognition<Layer2Cueing<Layer1GameState>>, Layer2Managers>,
                     public Layer3ScriptTypes
{
public:

#pragma region Types_Matching

  template <class LocationType>
  struct LocationTypes {};
  template <>
  struct LocationTypes<AirLocation> {
    typedef L3AirLocation Type;
  };
  template <>
  struct LocationTypes<BaseLocation> {
    typedef L3BaseLocation Type;
  };
  template <>
  struct LocationTypes<Chokepoint> {
    typedef L3Chokepoint Type;
  };
  template <>
  struct LocationTypes<Region> {
    typedef L3Region Type;
  };
  template <>
  struct LocationTypes<RegionLocation> {
    typedef L3RegionLocation Type;
  };
  template <>
  struct LocationTypes<Map> {
    typedef L3Map Type;
  };

  class L3UnitData : public virtual UnitData
  {
  public:
    L3UnitData(BWAPI::Unit* unit)
      : UnitData(unit)
    {
    }
  };

  template <class TechType>
  struct TechTypes {};
  template <>
  struct TechTypes<UnitTypeTech> {
    typedef L3UnitTypeTech Type;
  };
  template <>
  struct TechTypes<TechTypeTech> {
    typedef L3TechTypeTech Type;
  };
  template <>
  struct TechTypes<UpgradeTypeTech> {
    typedef L3UpgradeTypeTech Type;
  };

#pragma endregion Types_Matching

  static Layer3Script& getInstance()
  {
    return *s_instance;
  }

  Layer3Script(Layer3TacticalRecognition<Layer2Cueing<Layer1GameState>>& layer3TacticalRecognition, Layer2Managers& layer2Managers);
  ~Layer3Script();
  void onStart();
  void onFrame();
  void onEnd();
  void onSendText(std::string text) {}

  //
  // Action-dependent events
  //
  void onAttackMoveEnded(void* data);

  //
  // Action-independent events
  //
  void onUnitReleased(void* data, BWAPI::Unit* unit, const layer2::GroupTacticalAction* action);

  void onUnitRecruited(BWAPI::Unit* unit, StaticLocation* location);
  void onDependentTechsUpdate();

  Layer2Managers& getLayer2() const { return m_layer2Actions; }
  L3BaseLocation* getMainBase() const { return m_mainBase; }
  int countSupplyForOneRoundTroops() const;
  ResourcesSaturation countProbesSaturationRatio() const;
  L3BaseLocation* getNearestOwnMiningBaseLocation(BWAPI::Position position) const;

  /////////////////////////
  // LEVEL 4 - INTERFACE //
  /////////////////////////
  void setTargetTechs(const TechSet& targetTechs)
  {
    m_progressTech.setTargetTechs(targetTechs);
    // Techs needed will vary according to game flow (when something is finished/when something gets destroyed...)

    updateResourcesRatio();
  }

  void setTargetArmyComposition(const ArmyComposition& targetArmyComposition)
  {
    m_recruitArmy.setTargetArmyComposition(targetArmyComposition);

    updateResourcesRatio();
  }

  /////////
  // GUI //
  /////////
  const ArmyComposition& getTargetArmyComposition() const
  {
    return m_recruitArmy.getTargetArmyComposition();
  }
  const TechSet& getTargetTechs() const
  {
    return m_progressTech.getTargetTechs();
  }
  const ProgressTechTask* getProgressTechTask() const
  {
    return &m_progressTech;
  }

protected:
  static Layer3Script* s_instance;

  ResourcesAcquisitionRatio computeResourcesAcquisitionRatio() const;
  void updateResourcesRatio()
  {
    m_manageResources.updateResourcesRatio(computeResourcesAcquisitionRatio());
  }

  // Action-independent events
  EVENT_HANDLER2(Layer3Script, onUnitReleased, BWAPI::Unit*, const layer2::GroupTacticalAction*) m_unitReleasedEventHandler;

  // Periodic tasks
  ManageResourcesTask m_manageResources;
  BuildPylonsTask m_buildPylons;
  BuildGatewaysTask m_buildGateways;
  BuildExpansionsTask m_buildExpansions;
  ProgressTechTask m_progressTech;
  RecruitArmyTask m_recruitArmy;
  ManageArmyTask m_manageArmy;
  ScoutTask m_scout;
  BuildCannonsTask m_buildCannons;

  L3BaseLocation* m_mainBase;
};
