#pragma once
#include "../../../Common.h"
#include "../../../PerceptualState/Common.h"
#include "../../../SituationAnalysis/Layer1GameState/Layer1GameState.h"
#include "../../../SituationAnalysis/Layer2Cueing/Layer2Cueing.h"
#include "../../../SituationAnalysis/Layer3TacticalRecognition/Layer3TacticalRecognition.h"
#include "../../Layer2Actions/Interface/Common.h"
#include "../../Layer2Actions/Managers/Layer2Managers.h"
#include "../Layer3Tactical.h"

class Layer3ScriptTypes
{
//protected:
public:
  // Resources
  static const unsigned int PRIORITY_BUILD_CANNONS = 190;
  static const unsigned int PRIORITY_RECRUIT_OBSERVER = 320;
  static const unsigned int PRIORITY_BUILD_PYLONS = 300;
  static const unsigned int PRIORITY_BUILD_ASSIMILATOR = 275;
  static const unsigned int PRIORITY_RECRUIT_PROBES = 250;
  static const unsigned int PRIORITY_PROGRESS_TECH = 240;
  static const unsigned int PRIORITY_RECRUIT_ARMY = 230;
  static const unsigned int PRIORITY_BUILD_GATEWAYS = 200;
  static const unsigned int PRIORITY_EXPAND = 200;
  // No resources
  static const unsigned int PRIORITY_INITIAL_SCOUT = 500;
  static const unsigned int PRIORITY_COLLECT_MINERALS = 100;
  static const unsigned int PRIORITY_COLLECT_GAS = 100;
  static const unsigned int PRIORITY_ATTACK_MOVE = 50;

public:
  class ProbesResourcesDistribution
  {
  public:
    ProbesResourcesDistribution(size_t nbProbesMinerals, size_t nbProbesGas)
      : m_nbProbesMinerals(nbProbesMinerals)
      , m_nbProbesGas(nbProbesGas)
    {}
    size_t getNbProbesOnMinerals() const { return m_nbProbesMinerals; }
    size_t getNbProbesOnGas() const { return m_nbProbesGas; }
    size_t getTotalNbProbes() const { return m_nbProbesMinerals + m_nbProbesGas; }
  protected:
    size_t m_nbProbesMinerals;
    size_t m_nbProbesGas;
  };

  class ResourcesSaturation : public ProbesResourcesDistribution
  {
  public:
    ResourcesSaturation(size_t nbProbesMinerals, size_t nbCrystals, double mineralsSaturationRatio,
               size_t nbProbesGas, size_t nbGeysers)
      : ProbesResourcesDistribution(nbProbesMinerals, nbProbesGas)
      , m_nbCrystals(nbCrystals)
      , m_mineralsSaturationRatio(mineralsSaturationRatio)
      , m_nbGeysers(nbGeysers)
    {}
    size_t getNbCrystals() const { return m_nbCrystals; }
    double getMineralsSaturationRatio() const { return m_mineralsSaturationRatio; }
    size_t getNbGeysers() const { return m_nbGeysers; } // Total number of geysers, including depleted ones
    double getGasSaturationRatio() const { return m_nbProbesGas > 0 ? double(m_nbGeysers * 3) / double(m_nbProbesGas) : 0.; }
  protected:
    size_t m_nbCrystals;
    double m_mineralsSaturationRatio;
    size_t m_nbGeysers;
  };

  class ResourcesAcquisitionRatio
  {
  public:
    ResourcesAcquisitionRatio(double minerals, double gas)
      : m_pctMinerals(minerals / (minerals + gas))
      , m_pctGas(gas / (minerals + gas))
    {}
    double getPctMinerals() const { return m_pctMinerals; }
    double getPctGas() const { return m_pctGas; }
    bool operator==(const ResourcesAcquisitionRatio& other) const
    {
      return m_pctMinerals == other.m_pctMinerals &&
             m_pctGas == other.m_pctGas;
    }
    bool operator!=(const ResourcesAcquisitionRatio& other) const
    {
      return m_pctMinerals != other.m_pctMinerals ||
             m_pctGas != other.m_pctGas;
    }
  private:
    // Resources quantities are defined in relative terms w.r.t. each other
    double m_pctMinerals;
    double m_pctGas;
  };

  // Tactical action types
  typedef Layer2Managers::TacticalActionTypes<layer2::TacticalAction>::Type TacticalAction;
  typedef Layer2Managers::TacticalActionTypes<layer2::GroupTacticalAction>::Type GroupTacticalAction;
  typedef Layer2Managers::TacticalActionTypes<layer2::Attack>::Type Attack;
  typedef Layer2Managers::TacticalActionTypes<layer2::AttackMove>::Type AttackMove;
  typedef Layer2Managers::TacticalActionTypes<layer2::Build>::Type Build;
  typedef Layer2Managers::TacticalActionTypes<layer2::CollectMinerals>::Type CollectMinerals;
  typedef Layer2Managers::TacticalActionTypes<layer2::CollectGas>::Type CollectGas;
  typedef Layer2Managers::TacticalActionTypes<layer2::Defend>::Type Defend;
  typedef Layer2Managers::TacticalActionTypes<layer2::Expand>::Type Expand;
  typedef Layer2Managers::TacticalActionTypes<layer2::Harass>::Type Harass;
  typedef Layer2Managers::TacticalActionTypes<layer2::Move>::Type Move;
  //typedef Layer2Managers::TacticalActionTypes<layer2::OffensiveBuild>::Type OffensiveBuild;
  typedef Layer2Managers::TacticalActionTypes<layer2::Recruit>::Type Recruit;
  typedef Layer2Managers::TacticalActionTypes<layer2::UpgradeResearch>::Type UpgradeResearch;
  typedef Layer2Managers::TacticalActionTypes<layer2::TechResearch>::Type TechResearch;
  typedef Layer2Managers::TacticalActionTypes<layer2::Scout>::Type Scout;
};
