#include "stdafx.h"
#include "L3BaseLocation.h"
#include "../Layer3Script.h"

using namespace layer2;

void L3BaseLocation::buildAssimilator()
{
  if (getNbGeysers() > getNbOwnUnits(BWAPI::UnitTypes::Protoss_Assimilator) + getLastKnownNbEnemyRefineries() + m_buildAssimilators.size())
  {
    Build *buildAssimilator = new Build(PRIORITY_BUILD_ASSIMILATOR, this, BWAPI::UnitTypes::Protoss_Assimilator);
    m_buildAssimilators.add(buildAssimilator);
    buildAssimilator->execute();
  }
}

void L3BaseLocation::executeCollectResources()
{
  if (m_collectMinerals->isWaitingToExecute() && getLastKnownNbCrystalsLeft() > 0)
  {
    m_collectMinerals->execute();
  }
  if (m_collectGas->isWaitingToExecute() && getNbGeysers() > 0 && getNbCompletedOwnUnits(BWAPI::UnitTypes::Protoss_Assimilator) > 0)
  {
    m_collectGas->execute();
  }
}

void L3BaseLocation::onCollectMineralsEnded(void* data)
{
  m_collectMinerals = new CollectMinerals(PRIORITY_COLLECT_MINERALS, reinterpret_cast<const CollectMinerals*>(data)->getGroup(), this);
  m_collectMinerals->getTerminateEvent().subscribe(&m_collectMineralsEndedEventHandler, m_collectMinerals);
}

void L3BaseLocation::onCollectGasEnded(void* data)
{
  m_collectGas = new CollectGas(PRIORITY_COLLECT_GAS, reinterpret_cast<const CollectGas*>(data)->getGroup(), this);
  m_collectGas->getTerminateEvent().subscribe(&m_collectGasEndedEventHandler, m_collectGas);
}

void L3BaseLocation::executeRecruitProbes()
{
  if (m_recruitProbes == NULL)
  {
    m_recruitProbes = new Recruit(PRIORITY_RECRUIT_PROBES, this, BWAPI::UnitTypes::Protoss_Probe, true);
    m_recruitProbes->getUnitRecruitedEvent().subscribe(&m_probeRecruitedEventHandler, NULL);
    m_recruitProbes->execute();
  }
  else
  {
    assert(m_recruitProbes->isExecuting());
  }
}

void L3BaseLocation::onProbeRecruited(void* data, const layer2::Recruit* recruit, BWAPI::Unit* probe, StaticLocation* location)
{
  assert(probe->getType() == BWAPI::UnitTypes::Protoss_Probe);
  assert(dynamic_cast<StaticLocation*>(this) == location);

  m_collectMinerals->addUnit(probe);
}

void L3BaseLocation::cancelRecruitProbes()
{
  if (m_recruitProbes != NULL)
  {
    assert(m_recruitProbes->isExecuting());

    m_recruitProbes->cancel();
  }
}

bool L3BaseLocation::hasEnemyArmyUnits() const
{
  UnitGroup enemyUnits = Spar->getPerceptualState().getEnemyUnits(CircularLocation(getCurrentCenter(), BASE_LOCATION_RADIUS));
  for (UnitGroup::const_iterator it = enemyUnits.begin(); it != enemyUnits.end(); ++it)
  {
    BWAPI::Unit* unit = *it;
    if (!unit->getType().isWorker() && (unit->getType().groundWeapon() != BWAPI::WeaponTypes::None || unit->getType() == BWAPI::UnitTypes::Protoss_High_Templar))
      return true;
  }
  return false;
}