#pragma once
#include "../Layer3ScriptTypes.h"
#include "L3PrimitiveLocation.h"

class L3BaseLocation : public virtual BaseLocation, public Layer3ScriptTypes, public L3PrimitiveLocation
{
public:
  template <class LocationTypes>
  L3BaseLocation(LocationTypes* dummy, const BWTA::BaseLocation* location, Region* parent)
    : BaseLocation(dummy, location, parent)
    , PrimitiveLocation(parent)
    , L3PrimitiveLocation(parent)
    , StaticLocation(parent)
    , m_collectMineralsEndedEventHandler(*this)
    , m_collectGasEndedEventHandler(*this)
    , m_probeRecruitedEventHandler(*this)
    , m_collectMinerals(*this, new CollectMinerals(PRIORITY_COLLECT_MINERALS, &PersistentUnitGroup::EmptyGroup, this))
    , m_collectGas(*this, new CollectGas(PRIORITY_COLLECT_GAS, &PersistentUnitGroup::EmptyGroup, this))
    , m_recruitProbes(*this)
    , m_buildAssimilators(*this)
  {
    m_collectMinerals->getTerminateEvent().subscribe(&m_collectMineralsEndedEventHandler, m_collectMinerals);
    m_collectGas->getTerminateEvent().subscribe(&m_collectGasEndedEventHandler, m_collectGas);
  }

  ~L3BaseLocation()
  {
  }

  size_t getNbProbesOnMinerals() const
  {
    return m_collectMinerals->getGroup()->size();
  }
  CollectMinerals* getCollectMinerals()
  {
    return m_collectMinerals;
  }

  size_t getNbProbesOnGas() const
  {
    return m_collectGas->getGroup()->size();
  }
  CollectGas* getCollectGas()
  {
    return m_collectGas;
  }

  bool hasEnemyArmyUnits() const;

  void buildAssimilator();

  void executeCollectResources();

  void executeRecruitProbes();
  void cancelRecruitProbes();

  //
  // Events
  //
  void onCollectMineralsEnded(void* data);
  void onCollectGasEnded(void* data);
  void onProbeRecruited(void* data, const layer2::Recruit* recruit, BWAPI::Unit* probe, StaticLocation* location);

protected:
  std::string probesName(const std::string suffix)
  {
    std::stringstream ss;
    ss << "Location#" << getId() << "-Probes" << suffix;
    return ss.str();
  }

  EVENT_HANDLER(L3BaseLocation, onCollectMineralsEnded) m_collectMineralsEndedEventHandler;
  EVENT_HANDLER(L3BaseLocation, onCollectGasEnded) m_collectGasEndedEventHandler;
  EVENT_HANDLER3(L3BaseLocation, onProbeRecruited, const layer2::Recruit*, BWAPI::Unit*, PrimitiveLocation*) m_probeRecruitedEventHandler;

  ProcessContainer<CollectMinerals> m_collectMinerals;
  ProcessContainer<CollectGas> m_collectGas;
  ProcessContainer<Recruit> m_recruitProbes;
  ProcessContainerList<Build> m_buildAssimilators;
};