#include "stdafx.h"
#include "BuildExpansionsTask.h"
#include "../Layer3Script.h"

const static double SATURATION_RATIO_TO_TRIGGER_EXPAND = 0.85;

void BuildExpansionsTask::run()
{
  // Expansions
  if (m_expand == NULL && 
      m_layer3Script.countProbesSaturationRatio().getMineralsSaturationRatio() >= SATURATION_RATIO_TO_TRIGGER_EXPAND)
  {
    BaseLocation* expandLocation = getNextExpandLocation();
    if (expandLocation != NULL)
    {
      m_expand = new Expand(
        PRIORITY_EXPAND,
        expandLocation);
      m_expand->execute();
    }
  }
}

BaseLocation* BuildExpansionsTask::getNextExpandLocation() const
{
  BaseLocation* bestLocation = NULL;
  double minDistance = boost::numeric::bounds<double>::highest();

  const Map* map = Spar->getPerceptualState().getMap();
  for (Map::base_location_const_iterator it = map->baseLocationsBegin(); it != map->baseLocationsEnd(); ++it)
  {
    L3BaseLocation* candidate = dynamic_cast<L3BaseLocation*>(*it);
    if (candidate->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Nexus) == 0 &&
        !candidate->hasEnemyResourceDepot() &&
        !candidate->hasEnemyArmyUnits())
    {
      BWAPI::TilePosition mainBasePos(m_layer3Script.getMainBase()->getCurrentCenter());
      BWAPI::TilePosition candiatePos(candidate->getCurrentCenter());
      if(BWTA::isConnected(mainBasePos, candiatePos))
      {
        double distance = BWTA::getGroundDistance(mainBasePos, candiatePos);
          //= m_layer3Script.getMainBase()->getCurrentCenter().getDistance(candidate->getCurrentCenter());
        if (distance < minDistance && distance > 0)
        {
          minDistance = distance;
          bestLocation = candidate;
        }
      }
    }
  }
  return bestLocation;
}