#pragma once
#include "Layer3Task.h"

//
// Input: gas/mineral ratio
//

class ManageArmyTask : public Layer3Task
{
private:
  struct UnitPower
  {
    UnitPower()
      : AttackPower(0)
      , HitPointPower(0)
    {}UnitPower(double attackPower, double hitPointPower)
      : AttackPower(attackPower)
      , HitPointPower(hitPointPower)
    {}
    double AttackPower;
    double HitPointPower;
  };

  struct RegionWrapper
  {
    RegionWrapper(Region* region)
      : region(region)
      , lastVisitedFrame(0)
      , fromRegion(NULL)
    {}
    Region* region;
    RegionWrapper* fromRegion;
    int lastVisitedFrame;
  };

public:
  ManageArmyTask(Layer3Script& layer3Script)
  : Layer3Task(layer3Script)
  , m_attackMoveEndedEventHandler(*this)
  , m_unitMorphCompletedEventHandler(*this)
  , m_scoutArrivedEventHandler(*this)
  , m_army(new PersistentUnitGroup("army"), *this)
  , m_defendGroup(new PersistentUnitGroup("Defenders"), *this)
  , m_attackGroup(new PersistentUnitGroup("Attackers"), *this)
  , m_highTemplars(new PersistentUnitGroup("High Templars"), *this)
  , m_attackingObservers(new PersistentUnitGroup("Attacking Observers"), *this)
  , m_defendingObservers(new PersistentUnitGroup("Defending Observers"), *this)
  , m_scoutingObservers(new PersistentUnitGroup("Scouting Observers"), *this)
  , m_attackAction(NULL)
  , m_defendAction(*this)
  , m_scout(*this) 
  , m_mainRegion(NULL)
  , m_goalRegionAttack(NULL)
  , m_currentRegionAttack(NULL)
{ 
}

  ~ManageArmyTask()
  {
    for(stdext::hash_map<Region*, RegionWrapper*>::iterator it = m_regionWrappers.begin(); it != m_regionWrappers.end(); ++it)
      delete (it->second);
  }


  void run();

  virtual void runV()
  {
    run();
  }

  void addUnit(BWAPI::Unit* unit);

  Scout* getNewScout(BWAPI::Unit* scout);
  void onAttackMoveEnded(void* data);
  void OnUnitMorphCompleted(void* data, BWAPI::Unit* unit);
  void onScoutArrived(void* data, const layer2::Scout* action, StaticLocation* arrivedAtLocation, BWAPI::Unit* unit);

private:
  RegionWrapper* ManageArmyTask::getAttackLocation(RegionWrapper* fromRegion, UnitPower armyPower, bool canBeDetected, double groundRatio, double minSuperiorityRatio);
  ManageArmyTask::UnitPower ManageArmyTask::getUnitPower(BWAPI::UnitType unitType, bool ownUnit, bool canBeDetected, double groundOnlyRatio) const;

  EVENT_HANDLER(ManageArmyTask, onAttackMoveEnded) m_attackMoveEndedEventHandler;
  Event<boost::tuples::tuple<>>::SubscriberID m_attackMoveEndedSubscriberId;
  EVENT_HANDLER1(ManageArmyTask, OnUnitMorphCompleted, BWAPI::Unit*) m_unitMorphCompletedEventHandler;
  Event<boost::tuples::tuple<BWAPI::Unit*>>::SubscriberID m_unitMorphCompletedSubscriberId;
  EVENT_HANDLER3(ManageArmyTask, onScoutArrived, const layer2::Scout*, StaticLocation*,BWAPI::Unit*) m_scoutArrivedEventHandler; 

  stdext::hash_map<Region*, RegionWrapper*> m_regionWrappers;
  RegionWrapper* m_mainRegion;
  RegionWrapper* m_goalRegionAttack;
  RegionWrapper* m_currentRegionAttack;

  // Actions
  AttackMove* m_attackAction;
  ProcessContainer<Defend> m_defendAction;
  ProcessContainer<Scout> m_scout;

  // Groups
  AutoPtr<PersistentUnitGroup> m_army;
  AutoPtr<PersistentUnitGroup> m_attackGroup;
  AutoPtr<PersistentUnitGroup> m_defendGroup;
  AutoPtr<PersistentUnitGroup> m_highTemplars;

  AutoPtr<PersistentUnitGroup> m_attackingObservers;
  AutoPtr<PersistentUnitGroup> m_defendingObservers;
  AutoPtr<PersistentUnitGroup> m_scoutingObservers; 

  void executeScout(Scout* scout);
};
