#include "stdafx.h"
#include "ProgressTechTask.h"
#include "../Layer3Script.h"

// TODO: should be dependent on resources available
static const size_t MAX_SIMULTANEOUS_TECH_ACTIONS = 1;

size_t ProgressTechTask::getRootTechsToAcquire(Tech* target, std::set<Tech*>& result)
{
  size_t nbTechs = 0;
  if (target != NULL && !target->isAcquired())
  {
    const std::set<Tech*>& dependencies = target->getDependencies();
    for (std::set<Tech*>::const_iterator it = dependencies.begin(); it != dependencies.end(); ++it)
    {
      Tech* dependency = *it;
      nbTechs += getRootTechsToAcquire(dependency, result);
    }
    if (nbTechs == 0)
    {
      result.insert(target);
      ++nbTechs;
    }
  }
  return nbTechs;
}

void ProgressTechTask::onTechActionTerminate(void* data)
{
  TechActionWrapper* wrapper = reinterpret_cast<TechActionWrapper*>(data);
  m_currentTechActions.erase(wrapper->m_posInContainer);
}

void ProgressTechTask::run()
{
  std::map<Tech*, std::map<Tech*, TechActionWrapper>::iterator> currentTechs;
  for (std::map<Tech*, TechActionWrapper>::iterator it = m_currentTechActions.begin(); it != m_currentTechActions.end(); ++it)
  {
    currentTechs.insert(std::make_pair(it->first, it));
  }
  std::set<Tech*> rootTechs;
  for (TechSet::const_iterator it = m_currentTechTargets.begin(); it != m_currentTechTargets.end(); ++it)
  {
    Tech* techTarget = *it;
    getRootTechsToAcquire(techTarget, rootTechs);
  }
  for (std::set<Tech*>::const_iterator it = rootTechs.begin(); it != rootTechs.end(); ++it)
  {
    Tech* rootTech = *it;
    std::map<Tech*, std::map<Tech*, TechActionWrapper>::iterator>::iterator existing = currentTechs.find(rootTech);
    if (existing != currentTechs.end())
    {
      currentTechs.erase(existing);
    }
    else
    {
      if (shouldResearchTech(rootTech))
      {
        TacticalAction* action = dynamic_cast<L3Tech*>(rootTech)->instantiateAction(PRIORITY_PROGRESS_TECH);
        Process::zombifyOnTerminate(action);
        std::pair<std::map<Tech*, TechActionWrapper>::iterator, bool> result = m_currentTechActions.insert(std::make_pair(rootTech, TechActionWrapper(action)));
        assert(result.second);
        TechActionWrapper& wrapper = result.first->second;
        wrapper.m_posInContainer = result.first;
        action->getTerminateEvent().subscribe(&m_techActionTerminateEventHandler, &wrapper);

        action->execute();
      }
    }
  }

  //for (std::map<Tech*, std::map<Tech*, TechActionWrapper>::iterator>::iterator it = currentTechs.begin(); it != currentTechs.end(); ++it)
  //{
  //  std::map<Tech*, TechActionWrapper>::iterator toCancel = it->second;
  //  toCancel->second.m_techAction->cancel();
  //}
}

bool ProgressTechTask::shouldResearchTech(Tech* tech) const
{
  // Allow one building tech + one other tech if there's no building tech at the moment

  size_t nbBuildingTechs = 0;
  size_t nbOtherTechs = 0;
  for (std::map<Tech*, TechActionWrapper>::const_iterator it = m_currentTechActions.begin(); it != m_currentTechActions.end(); ++it)
  {
    UnitTypeTech* unitTypeTech = dynamic_cast<UnitTypeTech*>(it->first);
    if (unitTypeTech != NULL)
    {
      assert(unitTypeTech->getType().isBuilding());
      ++nbBuildingTechs;
    }
    else
    {
      ++nbOtherTechs;
    }
  }

  return nbBuildingTechs == 0 &&
         (dynamic_cast<UnitTypeTech*>(tech) != NULL || nbOtherTechs == 0);
}