#pragma once
#include "Layer3Task.h"
#include "../../Interface/TechSet.h"
#include "../../../../PerceptualState/Techs/Techs.h"

//
// Input: techs to acquire
//

class ProgressTechTask : public Layer3Task
{
private:
  struct TechActionWrapper
  {
    TechActionWrapper(TacticalAction* techAction)
      : m_techAction(techAction)
    {}
    TacticalAction* const m_techAction;
    std::map<Tech*, TechActionWrapper>::const_iterator m_posInContainer;
  };
  struct transform_to_pair : std::unary_function<std::pair<Tech*, TechActionWrapper>, std::pair<const Tech*, const TacticalAction*>>
  {
    result_type operator()(const argument_type& wrapper) const
    {
      return std::make_pair(wrapper.first, wrapper.second.m_techAction);
    }
  };

public:
  ProgressTechTask(Layer3Script& layer3Script)
    : Layer3Task(layer3Script)
    , m_techActionTerminateEventHandler(*this)
  {
  }

  ~ProgressTechTask()
  {
    for (std::map<Tech*, TechActionWrapper>::const_iterator it = m_currentTechActions.begin(); it != m_currentTechActions.end(); ++it)
    {
      delete it->second.m_techAction;
    }
  }

  void setTargetTechs(const TechSet& targetTechs)
  {
    m_currentTechTargets = targetTechs;
  }

  void run();

  virtual void runV()
  {
    run();
  }

  void onTechActionTerminate(void* data);

  /////////
  // GUI //
  /////////
  const TechSet& getTargetTechs() const
  {
    return m_currentTechTargets;
  }
  typedef boost::transform_iterator<transform_to_pair, std::map<Tech*, TechActionWrapper>::const_iterator> TechActionIterator;
  std::pair<TechActionIterator, TechActionIterator> getCurrentTechActions() const
  {
    return std::make_pair(boost::make_transform_iterator(m_currentTechActions.begin(), transform_to_pair()),
                          boost::make_transform_iterator(m_currentTechActions.end(), transform_to_pair()));
  }

private:
  static size_t getRootTechsToAcquire(Tech* target, std::set<Tech*>& result);
  bool shouldResearchTech(Tech* tech) const;

  EVENT_HANDLER(ProgressTechTask, onTechActionTerminate) m_techActionTerminateEventHandler;

  TechSet m_currentTechTargets;
  std::map<Tech*, TechActionWrapper> m_currentTechActions;
};
