#include "stdafx.h"
#include "UpdateTargetArmyAndTechsTask.h"
#include "../Layer4Heuristic.h"
#include "../../../Layer3Tactical/Interface/ArmyComposition.h"
#include "../../../../PerceptualState/Techs/Techs.h"

static const int PSI_TO_TRIGGER_FORGE = 50;

static const int PSI_TO_TRIGGER_OBSERVERS = 75;

//
// Warning: ugly ad-hoc code follows
//

// TODO: build army before teching straight away!

// TODO: template when stabilized
ArmyComposition UpdateTargetArmyAndTechsTask::estimateBestCounter(const Map* map) const
{
  ArmyComposition counter;

  // Input: own army, current resources, enemy army

  BWAPI::Race enemyRace = BWAPI::Broodwar->enemy()->getRace();

  if (enemyRace == BWAPI::Races::Protoss)
  {
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Zealot, 1);
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Dragoon, 3);
    //counter.addUnlimited(BWAPI::UnitTypes::Protoss_High_Templar, 1);
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Dark_Templar, 1);
  }
  else if (enemyRace == BWAPI::Races::Terran)
  {
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Zealot, 2);
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Dragoon, 3);
    //counter.addUnlimited(BWAPI::UnitTypes::Protoss_High_Templar, 1);
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Dark_Templar, 1);
  }
  else if (enemyRace == BWAPI::Races::Zerg)
  {
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Zealot, 3);
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Dragoon, 1);
    //counter.addUnlimited(BWAPI::UnitTypes::Protoss_High_Templar, 1);
  }
  else
  {
    // 1a2a3a for the win!
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Zealot, 1);
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_Dragoon, 1);
  }

  if(BWAPI::Broodwar->self()->gas() > 500)
  {
    counter.addUnlimited(BWAPI::UnitTypes::Protoss_High_Templar, 1);
  }

  if (Techs::ObserverTech->isAcquired())
  {
    // Let's say 3 observers: scout, defense and attack
    const size_t nbObservers = Spar->getPerceptualState().getMap()->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Observer);
    if (nbObservers < 5)
    {
      counter.addLimited(BWAPI::UnitTypes::Protoss_Observer, 5 - nbObservers);
    }
  }

  // TODO: add limited number of corsairs here, and possibly shuttles/reavers/carriers/arbiter if we have some behaviors for them

  return counter;
}

TechSet UpdateTargetArmyAndTechsTask::getTargetTechs(const ArmyComposition& armyComposition) const
{
  const Map* map = Spar->getPerceptualState().getMap();

  TechSet techs;

  std::map<BWAPI::UnitType, size_t> infiniteComposition = armyComposition.getUnlimitedComposition();
  for (std::map<BWAPI::UnitType, size_t>::const_iterator it = infiniteComposition.begin(); it != infiniteComposition.end(); ++it)
  {
    std::pair<BWAPI::UnitType, size_t> units = *it;
    techs.add(Techs::getUnitTypeTech(units.first), 0);
  }

  if (map->getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Dark_Templar) > 0 || map->getNbEnemyUnits(BWAPI::UnitTypes::Protoss_High_Templar) > 0 || 
        map->getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Archon) > 0 || map->getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Dark_Archon) > 0 || map->getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Templar_Archives) > 0 ||
      map->getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Arbiter) > 0  || map->getNbEnemyUnits(BWAPI::UnitTypes::Protoss_Arbiter_Tribunal) > 0 ||
      map->getNbEnemyUnits(BWAPI::UnitTypes::Terran_Ghost) > 0  || map->getNbEnemyUnits(BWAPI::UnitTypes::Terran_Covert_Ops) > 0 ||
      map->getNbEnemyUnits(BWAPI::UnitTypes::Terran_Wraith) > 0  || (map->getNbEnemyUnits(BWAPI::UnitTypes::Terran_Starport) > 0 && map->getNbEnemyUnits(BWAPI::UnitTypes::Terran_Control_Tower) > 0) ||
      map->getNbEnemyUnits(BWAPI::UnitTypes::Zerg_Lurker) > 0  || (map->getNbEnemyUnits(BWAPI::UnitTypes::Zerg_Lair) > 0 && map->getNbEnemyUnits(BWAPI::UnitTypes::Zerg_Hydralisk_Den) > 0))
  {
    techs.add(Techs::ForgeTech, 0);
    techs.add(Techs::ObserverTech, 0);
  }

  if (!Techs::ForgeTech->isAcquired() &&
      BWAPI::Broodwar->self()->supplyUsed() >= PSI_TO_TRIGGER_FORGE * 2)
  {
    techs.add(Techs::ForgeTech, 0);
  }

  if (!Techs::ObserverTech->isAcquired() &&
      BWAPI::Broodwar->self()->supplyUsed() >= PSI_TO_TRIGGER_OBSERVERS * 2)
  {
    techs.add(Techs::ObserverTech, 0);
  }

  addImprovementTechs(armyComposition, techs);

  return techs;
}

void UpdateTargetArmyAndTechsTask::addImprovementTechs(const ArmyComposition& armyComposition, TechSet& techs) const
{
  const Map* map = Spar->getPerceptualState().getMap();

  if (!Techs::PsionicStormTech->isAcquired() &&
      Techs::PsionicStormTech->allDependenciesAcquired() &&
      map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_High_Templar) >= 1)
  {
    techs.add(Techs::PsionicStormTech, 0);
  }
  if (!Techs::SingularityCharge->isAcquired() &&
      Techs::SingularityCharge->allDependenciesAcquired() &&
      map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Dragoon) >= 4)
  {
    techs.add(Techs::SingularityCharge, 0);
  }
  if (!Techs::LegEnhancements->isAcquired() &&
      Techs::LegEnhancements->allDependenciesAcquired() &&
      map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Zealot) >= 4)
  {
    techs.add(Techs::LegEnhancements, 0);
  }
  if (!Techs::CarrierCapacity->isAcquired() &&
      Techs::CarrierCapacity->allDependenciesAcquired() &&
      map->getNbOwnUnits(BWAPI::UnitTypes::Protoss_Carrier) >= 1)
  {
    techs.add(Techs::CarrierCapacity, 0);
  }

  if (map->getNbCompletedOwnUnits(BWAPI::UnitTypes::Protoss_Forge) > 0)
  {
    for (size_t i = 0; i < array_size(Techs::GroundWeapons); ++i) {
      if (!Techs::GroundWeapons[i]->isAcquired() &&
          Techs::GroundWeapons[i]->allDependenciesAcquired()) {
        techs.add(Techs::GroundWeapons[i], 0);
        break;
      }
    }
    for (size_t i = 0; i < array_size(Techs::GroundArmor); ++i) {
      if (!Techs::GroundArmor[i]->isAcquired() &&
          Techs::GroundArmor[i]->allDependenciesAcquired()) {
        techs.add(Techs::GroundArmor[i], 0);
        break;
      }
    }
    for (size_t i = 0; i < array_size(Techs::PlasmaShields); ++i) {
      if (!Techs::PlasmaShields[i]->isAcquired() &&
          Techs::PlasmaShields[i]->allDependenciesAcquired()) {
        techs.add(Techs::PlasmaShields[i], 0);
        break;
      }
    }
  }
}

void UpdateTargetArmyAndTechsTask::run()
{
  ArmyComposition counter = estimateBestCounter(Spar->getPerceptualState().getMap());

  TechSet techs = getTargetTechs(counter);

  m_layer4Heuristic.getLayer3().setTargetArmyComposition(counter);

  m_layer4Heuristic.getLayer3().setTargetTechs(techs);
}
