#include "stdafx.h"
#include "Threat.h"
#include "../../SparAIContainer.h"

// TODO: bunkers

// TODO also consider minRange() (i.e., Terran Siege Tank)

// TODO: consider unit speed
// TODO: consider unit upgrades (if not done already)

static const int FRAME_COUNT_TO_REMOVE_INFLUENCE = 50;

void Threat::update(InfluenceMap<Threat>& map)
{
  map.reset();

  if (map.getPlayer() == BWAPI::Broodwar->self())
  {
    const std::set<BWAPI::Unit*>& units = map.getPlayer()->getUnits();
    update(map, units.begin(), units.end());
  }
  else if (map.getPlayer() == BWAPI::Broodwar->enemy())
  {
    std::pair<Map::enemy_unit_const_iterator, Map::enemy_unit_const_iterator> enemyUnits = Spar->getPerceptualState().getMap()->getEnemyUnits();
    update(map, enemyUnits.first, enemyUnits.second);
  }
}

void Threat::update(InfluenceMap<Threat>& map, BWAPI::Unit* unit)
{
  BWAPI::UnitType unitType = unit->getType();
  BWAPI::WeaponType groundWeapon = unitType.groundWeapon();
  BWAPI::WeaponType airWeapon = unitType.airWeapon();
  
  int groundDamage = groundWeapon.damageAmount() + groundWeapon.damageBonus();
  int airDamage = airWeapon.damageAmount() + airWeapon.damageBonus();

  int groundRange = static_cast<int>((groundWeapon.maxRange() + 1.5*TILE_SIZE) / TILE_SIZE);
  int airRange =    static_cast<int>((airWeapon.maxRange()    + 1.5*TILE_SIZE) / TILE_SIZE);
  
  updateInfluence(map, unit->getPosition(), groundRange, groundDamage, airRange, airDamage, unit->getType().topSpeed());
}

void Threat::update(InfluenceMap<Threat>& map, const EnemyUnit* unit)
{
  if (unit->isAccessible())
  {
    update(map, unit->getUnit());
  }
  else
  {
    BWAPI::UnitType unitType = unit->getType();
    BWAPI::WeaponType groundWeapon = unitType.groundWeapon();
    BWAPI::WeaponType airWeapon = unitType.airWeapon();
    
    int groundDamage = groundWeapon.damageAmount() + groundWeapon.damageBonus();
    int airDamage = airWeapon.damageAmount() + airWeapon.damageBonus();

    int groundRange = static_cast<int>((groundWeapon.maxRange() + 1.5*TILE_SIZE) / TILE_SIZE);
    int airRange =    static_cast<int>((airWeapon.maxRange()    + 1.5*TILE_SIZE) / TILE_SIZE);
    
    if (unitType.isBuilding())
    {
      updateInfluence(map, unit->getPosition(), groundRange, groundDamage, airRange, airDamage, unit->getType().topSpeed());
    }
    else
    {
      // Could decrease the influence according to the number of frames, until 0
      if (BWAPI::Broodwar->getFrameCount() - unit->getLastEvadedFrame() < FRAME_COUNT_TO_REMOVE_INFLUENCE)
        updateInfluence(map, unit->getPosition(), groundRange, groundDamage, airRange, airDamage, unit->getType().topSpeed());
    }
  }
}

Threat::Functor::result_type Threat::Functor::operator()(Threat::Functor::first_argument_type x, 
                                                         Threat::Functor::second_argument_type y)
{
  x += m_x;
  y += m_y;

  if (m_map.isValid(x, y))
  {
    Threat influence = m_map.getInfluence(x, y);
    influence.ground += m_ground;
    influence.air += m_air;
    m_map.setInfluence(x, y, influence);
  }
}
