#pragma once
#include "InfluenceMap.h"
#include <BWAPI.h>
#include <cassert>

struct Visibility
{
  bool visible;
  unsigned long lastVisited;

  Visibility() { reset(); }

  void reset()
  {
    visible = false;
    lastVisited = 0L;
  }

  void display(int x, int y, int w, int h, BWAPI::Player* player) const
  {
    // do nothing
    unused(x);
    unused(y);
    unused(w);
    unused(h);
    unused(player);
  }

  bool isNull() const { return !visible; }

  static void update(InfluenceMap<Visibility>& map)
  {
    assert(map.getPlayer() == BWAPI::Broodwar->self());

    assert(map.getWidth()  == BWAPI::Broodwar->mapWidth()
        && map.getHeight() == BWAPI::Broodwar->mapHeight());

    for (size_t y=0; y<map.getHeight(); ++y)
    {
      for (size_t x=0; x<map.getWidth(); ++x)
      {
        Visibility influence = map.getInfluence(x, y);
        influence.update(BWAPI::Broodwar->isVisible(x, y));
        map.setInfluence(x, y, influence);
      }
    }
  }

private:
  void update(bool v)
  {
    lastVisited = (!v && !visible) ? lastVisited+1 : 0L;
    visible = v;
  }
};