#pragma once
#include "PrimitiveLocation.h"

class Map;
class Region;

class Chokepoint : public virtual PrimitiveLocation
{
public:
  template <class LocationTypes>
  Chokepoint(LocationTypes*, BWTA::Chokepoint* chokepoint, Region* firstRegion, Region* secondRegion, Map* parent)
    : PrimitiveLocation(parent)
    , StaticLocation(parent)
    , m_chokepoint(chokepoint)
    , m_firstRegion(firstRegion)
    , m_secondRegion(secondRegion)
  {
  }

  Map* getParent();
  const Map* getParent() const;

  virtual BWAPI::Position getCurrentCenter() const
  {
    return m_chokepoint->getCenter();
  }

  virtual bool isInLocation(BWAPI::Position position) const
  {
    // TODO: improve this: should be only the tiles crossed by the chokepoint line segment

    const std::pair<BWAPI::Position, BWAPI::Position>& sides = m_chokepoint->getSides();

    return (sides.first.x() <= position.x() && position.x() <= sides.second.x()) &&
           (sides.first.y() <= position.y() && position.y() <= sides.second.y());
  }

  std::pair<Region*, Region*> getRegions() 
  {
    return std::make_pair(m_firstRegion, m_secondRegion);
  }
  std::pair<const Region*, const Region*> getRegions() const
  {
    return std::make_pair(m_firstRegion, m_secondRegion);
  }

  const BWTA::Chokepoint* getChokepoint() const
  {
    return m_chokepoint;
  }

protected:
  friend class Map;

  virtual void output(std::ostream& out) const
  {
    out << "Chokepoint " << m_chokepoint->getCenter();
  }

  BWTA::Chokepoint* const m_chokepoint;
  Region* const m_firstRegion;
  Region* const m_secondRegion;
};