#pragma once
#include "StaticLocationBase.h"
#include "PrimitiveLocationBase.h"

class CompositeLocation;
class StaticLocation;

class CompositeLocationBase : public virtual StaticLocationBase
{
protected:
  friend class StaticLocation;

  template <class UnitType>
  class UnitWrapper
  {
  public:
    UnitWrapper(UnitType* unit)
      : m_unit(unit)
    {}
    UnitType* getUnit() { return m_unit; }
    const UnitType* getUnit() const { return m_unit; }
  protected:
    friend class CompositeLocation;
    UnitType* const m_unit;
    typename stdext::hash_multimap<BWAPI::UnitType, UnitWrapper>::const_iterator m_posInContainer;
  };
  template <class UnitType>
  struct extract_unit : std::unary_function<UnitWrapper<UnitType>, UnitType*>
  {
    UnitType* operator()(UnitWrapper<UnitType>& unitWrapper) const
    {
      return unitWrapper.getUnit();
    }
  };
  template <class UnitType>
  struct extract_unit_const : std::unary_function<UnitWrapper<UnitType>, const UnitType*>
  {
    extract_unit_const() {}
    extract_unit_const(const extract_unit<UnitType>&) {}

    const UnitType* operator()(const UnitWrapper<UnitType>& unitWrapper) const
    {
      return unitWrapper.getUnit();
    }
  };

  typedef boost::transform_iterator<composite_functor<extract_unit_const<OwnUnit>, extract_pair_second_const<stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<OwnUnit>>::const_iterator::value_type>>, 
                                    stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<OwnUnit>>::const_iterator> 
          own_unit_const_iterator;
  typedef own_unit_const_iterator
          own_unit_type_const_iterator;
  typedef boost::filter_iterator<IsBuilding<OwnUnit>, own_unit_const_iterator>
          own_building_const_iterator;

  typedef boost::transform_iterator<composite_functor<extract_unit<EnemyUnit>, extract_pair_second<stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<EnemyUnit>>::iterator::value_type>>, 
                                    stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<EnemyUnit>>::iterator> 
          enemy_unit_iterator;
  typedef enemy_unit_iterator 
          enemy_unit_type_iterator;
  typedef boost::filter_iterator<IsBuilding<EnemyUnit>, enemy_unit_iterator>
          enemy_building_iterator;
  typedef boost::transform_iterator<composite_functor<extract_unit_const<EnemyUnit>, extract_pair_second_const<stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<EnemyUnit>>::const_iterator::value_type>>, 
                                    stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<EnemyUnit>>::const_iterator> 
          enemy_unit_const_iterator;
  typedef enemy_unit_const_iterator 
    enemy_unit_type_const_iterator;
  typedef boost::filter_iterator<IsBuilding<EnemyUnit>, enemy_unit_const_iterator>
          enemy_building_const_iterator;

  typedef boost::transform_iterator<composite_functor<extract_unit<NeutralUnit>, extract_pair_second<stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<NeutralUnit>>::iterator::value_type>>, 
                                    stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<NeutralUnit>>::iterator> 
          neutral_unit_iterator;
  typedef neutral_unit_iterator 
          neutral_unit_type_iterator;
  typedef boost::filter_iterator<IsBuilding<NeutralUnit>, neutral_unit_iterator>
          neutral_building_iterator;
  typedef boost::transform_iterator<composite_functor<extract_unit_const<NeutralUnit>, extract_pair_second_const<stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<NeutralUnit>>::const_iterator::value_type>>, 
                                    stdext::hash_multimap<BWAPI::UnitType, UnitWrapper<NeutralUnit>>::const_iterator> 
          neutral_unit_const_iterator;
  typedef neutral_unit_const_iterator 
          neutral_unit_type_const_iterator;
  typedef boost::filter_iterator<IsBuilding<NeutralUnit>, neutral_unit_const_iterator>
          neutral_building_const_iterator;
};
