#include "stdafx.h"
#include "PrimitiveLocation.h"
#include "../../../SparAIContainer.h"

void PrimitiveLocation::onUnitDiscover(BWAPI::Unit* unit)
{
  INVARIANT_UNIT_POINTERS_ASSUMPTION;

  PrimitiveLocation* oldLocation = Spar->getUnitManager().getUnitLocation(unit);
  UnitLocationData* unitL = reinterpret_cast<UnitLocationData*>(Spar->getUnitManager().getUnitData(unit).m_data);
  if (oldLocation == NULL || unitL == NULL)
  {
    this->addUnit(unit);
  }
  else
  {
    if (oldLocation != this)
    {
      // TODO: should be a transfer instead of add/remove (common parents should not raise a UnitAdded event)
      oldLocation->removeUnit(unit);
      this->addUnit(unit);
    }
    else
    {
      if (unitL->getPlayer() != unit->getPlayer())
      {
        this->removeUnit(unit);
        this->addUnit(unit);
      }
      else if (unitL->getType() != unit->getType())
      {
        // TODO: no UnitRemoved/UnitAdded events should be raised here
        this->removeUnit(unit);
        this->addUnit(unit);
      }
      else
      {
        this->updateUnit(unit);
      }
    }
  }
}

void PrimitiveLocation::onUnitDestroy(void* data)
{
  UnitLocationData* unit = reinterpret_cast<UnitLocationData*>(data);
  if (unit->getPlayer() == BWAPI::Broodwar->self())
  {
    onUnitDestroy(static_cast<OwnUnit*>(unit));
  }
  else if (unit->getPlayer()->isEnemy(BWAPI::Broodwar->self()))
  {
    onUnitDestroy(static_cast<EnemyUnit*>(unit));
  }
  else
  {
    assert(unit->getPlayer() == BWAPI::Broodwar->neutral());
    onUnitDestroy(static_cast<NeutralUnit*>(unit));
  }
}

void PrimitiveLocation::onUnitMorph(void* data)
{
  UnitLocationData* unit = reinterpret_cast<UnitLocationData*>(data);
  if (unit->getPlayer() == BWAPI::Broodwar->self())
  {
    onUnitMorph(static_cast<OwnUnit*>(unit));
  }
  else if (unit->getPlayer()->isEnemy(BWAPI::Broodwar->self()))
  {
    onUnitMorph(static_cast<EnemyUnit*>(unit));
  }
  else
  {
    assert(unit->getPlayer() == BWAPI::Broodwar->neutral());
    onUnitMorph(static_cast<NeutralUnit*>(unit));
  }
}

void PrimitiveLocation::onUnitRenegade(void* data)
{
  UnitLocationData* unit = reinterpret_cast<UnitLocationData*>(data);
  if (unit->getPlayer() == BWAPI::Broodwar->self())
  {
    onUnitRenegade(static_cast<OwnUnit*>(unit));
  }
  else if (unit->getPlayer()->isEnemy(BWAPI::Broodwar->self()))
  {
    onUnitRenegade(static_cast<EnemyUnit*>(unit));
  }
  else
  {
    assert(unit->getPlayer() == BWAPI::Broodwar->neutral());
    onUnitRenegade(static_cast<NeutralUnit*>(unit));
  }
}

void PrimitiveLocation::onUnitEvade(void* data)
{
  UnitLocationData* unit = reinterpret_cast<UnitLocationData*>(data);
  if (unit->getPlayer() == BWAPI::Broodwar->self())
  {
    onUnitEvade(static_cast<OwnUnit*>(unit));
  }
  else if (unit->getPlayer()->isEnemy(BWAPI::Broodwar->self()))
  {
    onUnitEvade(static_cast<EnemyUnit*>(unit));
  }
  else
  {
    assert(unit->getPlayer() == BWAPI::Broodwar->neutral());
    onUnitEvade(static_cast<NeutralUnit*>(unit));
  }
}

PrimitiveLocation* PrimitiveLocation::getHostLocation(BWAPI::Unit* unit)
{
  // Could optimize for buildings which cannot move (and cannot morph)

  return Spar->getPerceptualState().getHostLocation(unit);
}

void PrimitiveLocation::addUnit(BWAPI::Unit* unit)
{
  if (unit->getPlayer() == BWAPI::Broodwar->self())
  {
    // Should be equivalent to hooking onUnitCreate()
    addUnit<OwnUnit>(unit);
  }
  else if (unit->getPlayer()->isEnemy(BWAPI::Broodwar->self()))
  {
    addUnit<EnemyUnit>(unit);
  }
  else if (unit->getPlayer() == BWAPI::Broodwar->neutral())
  {
    addUnit<NeutralUnit>(unit);
  }
}

void PrimitiveLocation::removeUnit(BWAPI::Unit* unit)
{
  UnitLocationData* unitL = reinterpret_cast<UnitLocationData*>(Spar->getUnitManager().getUnitData(unit).m_data);
  if (unitL->getPlayer() == BWAPI::Broodwar->self())
  {
    // Should be equivalent to hooking onUnitCreate()
    removeUnit<OwnUnit>(static_cast<OwnUnit*>(unitL));
  }
  else if (unitL->getPlayer()->isEnemy(BWAPI::Broodwar->self()))
  {
    removeUnit<EnemyUnit>(static_cast<EnemyUnit*>(unitL));
  }
  else if (unitL->getPlayer() == BWAPI::Broodwar->neutral())
  {
    removeUnit<NeutralUnit>(static_cast<NeutralUnit*>(unitL));
  }
}

void PrimitiveLocation::updateUnit(BWAPI::Unit* unit)
{
  UnitLocationData* unitL = reinterpret_cast<UnitLocationData*>(Spar->getUnitManager().getUnitData(unit).m_data);
  BaseUnitWrapper* unitWrapper = reinterpret_cast<BaseUnitWrapper*>(unitL->m_primitiveLocationData);

  assert(unitWrapper->m_unitDestroyID.isValid());
  unitWrapper->m_unitEvadeID = Spar->getUnitManager().subscribeIndividualUnitEvade(unit, &m_unitEvadeEventHandler, unitL);
  unitWrapper->m_unitMorphID = Spar->getUnitManager().subscribeIndividualUnitMorph(unit, &m_unitMorphEventHandler, unitL);
  unitWrapper->m_unitRenegadeID = Spar->getUnitManager().subscribeIndividualUnitRenegade(unit, &m_unitRenegadeEventHandler, unitL);

  if (unitL->getPlayer() == BWAPI::Broodwar->self())
  {
    static_cast<OwnUnit*>(unitL)->markAsAccessible();
  }
  else if (unitL->getPlayer()->isEnemy(BWAPI::Broodwar->self()))
  {
    static_cast<EnemyUnit*>(unitL)->markAsAccessible();
  }
  else if (unitL->getPlayer() == BWAPI::Broodwar->neutral())
  {
    static_cast<NeutralUnit*>(unitL)->markAsAccessible();
  }
}
