#pragma once
#include "CompositeLocation.h"
#include "BaseLocation.h"
#include "RegionLocation.h"
#include <set>

class Map;
class Chokepoint;

class Region : public virtual CompositeLocation
{
public:
  template <class LocationTypes>
  Region(LocationTypes*, BWTA::Region* region, Map* parent)
    : CompositeLocation(parent)
    , StaticLocation(parent)
    , m_region(region)
    , m_regionLocation(LocationTypes::instantiateRegionLocation(this))
  {
    for(std::set<BWTA::BaseLocation*>::const_iterator it = region->getBaseLocations().begin(); it != region->getBaseLocations().end(); ++it)
    {
      BaseLocation* baseLocation = LocationTypes::instantiateBaseLocation(*it, this);
      m_baseLocations.insert(baseLocation);
      addSubLocation(baseLocation);
    }

    addSubLocation(m_regionLocation);
  }

  ~Region()
  {
    for (std::set<BaseLocation*>::iterator it = m_baseLocations.begin(); it != m_baseLocations.end(); ++it)
    {
      delete *it;
    }
    delete m_regionLocation;
  }

  Map* getParent();
  const Map* getParent() const;

  virtual BWAPI::Position getCurrentCenter() const
  {
    return m_region->getCenter();
  }

  virtual bool isInLocation(BWAPI::Position position) const
  {
    return BWTA::getRegion(BWAPI::TilePosition(position)) == m_region;
  }

  const std::set<std::pair<Chokepoint*, Region*>>& getChokepoints() const
  {
    return m_chokepoints;
  }

  const std::set<std::pair<Region*, Chokepoint*>>& getAdjacentRegions() const
  {
    return m_adjacentRegions;
  }

  const std::set<BaseLocation*>& getBaseLocations() const
  {
    return m_baseLocations;
  }

  const BWTA::Region* getBWTARegion() const
  {
    return m_region;
  }

  RegionLocation* getRegionLocation()
  {
    return m_regionLocation;
  }

protected:
  friend class Map;

  void addChokepoint(Chokepoint* chokepoint);

  virtual void output(std::ostream& out) const
  {
    out << "Region " << m_region->getCenter();
  }

  BWTA::Region* m_region;
  std::set<std::pair<Chokepoint*, Region*>> m_chokepoints;
  std::set<std::pair<Region*, Chokepoint*>> m_adjacentRegions;

  //
  // Composing locations
  //
  std::set<BaseLocation*> m_baseLocations;
  
  // Default primitive location regrouping the buildings in this region which are not
  // part of the mining locations (possibly no mining locations)
  RegionLocation* const m_regionLocation; 
};
