#pragma once
#include <BWAPI.h>
#include <iostream>

class UnitTypeTech;
class Techs;

class Tech
{
public:
  Tech()
    : m_isAcquired(false)
  {}
  virtual ~Tech() {}
  const std::set<Tech*>& getDependencies() const { return m_incomingDependencies; }
  bool isAcquired() const 
  {
    //assert(m_isAcquired == isAcquiredImpl());
#ifdef _DEBUG
    bool acquiredImpl = isAcquiredImpl();
    if (m_isAcquired != acquiredImpl)
    {
      SPAR_LOG(LogTypes::GENERAL_ERROR, "Tech \"%s\" is wrongly marked as %s", getObjectDescription(this).c_str(), (m_isAcquired ? "acquired" : "not acquired"));
    }
#endif
    return m_isAcquired;
  }
  bool allDependenciesAcquired() const;
  friend std::ostream& operator<<(std::ostream& out, const Tech& tech)
  {
    out << tech.m_name;
    return out;
  }
  void setTechAcquired();
protected:
  friend class UnitTypeTech;
  virtual bool isAcquiredImpl() const = 0;
  virtual void onIncomingDependencyAcquired(Tech* dependency) { unused(dependency); }
  virtual void onIncomingDependencyLost(Tech* dependency) { unused(dependency); }
  virtual void addDependency(BWAPI::UnitType unitType);
  virtual void addDependency(BWAPI::TechType techType);
  virtual void addDependency(BWAPI::UpgradeType upgradeType, int level);
  void addOutgoingDependency(Tech* tech) { m_outgoingDependencies.insert(tech); }
  std::set<Tech*> m_incomingDependencies;
  std::set<Tech*> m_outgoingDependencies;
  /**
   * @brief Cache: should always be equal to isAcquiredImpl()
   */
  bool m_isAcquired;
  std::string m_name;
};

// Could be separated in UnitTypeTech/BuildingTypeTech
class UnitTypeTech : public virtual Tech
{
public:
  UnitTypeTech(BWAPI::UnitType unitType);
  BWAPI::UnitType getType() const { return m_unitType; }
protected:
  friend class Techs;
  virtual void addDependency(BWAPI::UnitType unitType);
  void setTechLost();
  void onUnitCompleted()
  {
    if (!m_isAcquired)
      setTechAcquired();
  }
  void onUnitRemoved(size_t nbCompletedUnits)
  {
    if (m_isAcquired && nbCompletedUnits == 0)
    {
      setTechLost();
    }
  }
  virtual bool isAcquiredImpl() const;
  virtual void onIncomingDependencyAcquired(Tech* dependency);
  virtual void onIncomingDependencyLost(Tech* dependency);
  const BWAPI::UnitType m_unitType;
};

// You're welcome to change this hideous class name!
class TechTypeTech : public virtual Tech
{
public:
  TechTypeTech(BWAPI::TechType techType);
  BWAPI::TechType getTech() const { return m_techType; }
protected:
  friend class Techs;
  virtual bool isAcquiredImpl() const;
  const BWAPI::TechType m_techType;
};

class UpgradeTypeTech : public virtual Tech
{
public:
  UpgradeTypeTech(BWAPI::UpgradeType upgradeType, int level);
  UpgradeTypeTech(BWAPI::UpgradeType upgradeType);
  BWAPI::UpgradeType getUpgrade() const { return m_upgradeType; }
  int getLevel() const { return m_level; }
protected:
  friend class Techs;
  virtual bool isAcquiredImpl() const;
  const BWAPI::UpgradeType m_upgradeType;
  const int m_level;
};
