#pragma once
#include "UnitLocationData.h"

/**
 * @brief Own units information stored in locations; for own units, the unit pointer is actually enough.
 */
class OwnUnit : public UnitLocationData
{
public:
  OwnUnit(BWAPI::Unit* unit, PrimitiveLocation* location)
    : UnitLocationData(unit, location)
  {
    assert(unit->getPlayer() == BWAPI::Broodwar->self());
  }
  BWAPI::Unit* getUnit() const { return m_unit; }
  bool isAccessible() const { return true; }
  void markAsAccessible() {}
  void markAsInaccessible() {}

  BWAPI::UnitType getLastType() const { return m_unit->getType(); }
  BWAPI::UnitType getType() const { return m_unit->getType(); }
  BWAPI::TilePosition getTilePosition() const { return m_unit->getTilePosition(); }
  BWAPI::Position getPosition() const { return m_unit->getPosition(); }
  bool isCompleted() const { return m_unit->isCompleted(); }
  bool isBeingConstructed() const { return m_unit->isBeingConstructed(); }

  friend std::ostream& operator<<(std::ostream& out, const OwnUnit& ownUnit)
  {
    BWAPI::Unit* unit = ownUnit.getUnit();
    out << "Own unit (0x" << std::nouppercase << std::hex << unit << std::dec << ") " << unit->getType().getName() << ", hit points " << unit->getHitPoints();
    return out;
  }
};
