#pragma once
#include "UnitData.h"
#include "Utils/Event.h"
#include <BWAPI.h>
#include <list>

class PrimitiveLocation;

/**
 * @brief The unit manager is responsible for providing per-unit data.
 */
class UnitManager
{
public:
  ~UnitManager();

  void onStart() {}

  void onFrame() {}

  void onEnd() {}

  void onUnitDiscover(BWAPI::Unit*) {}
  void onUnitEvade(BWAPI::Unit* unit)
  {
    UnitData* unitData = reinterpret_cast<UnitData*>(unit->getClientInfo());
    if (unitData != NULL)
    {
      unitData->m_individualUnitEvadeEvent.raise();
    }
  }
  void onUnitShow(BWAPI::Unit*) {}
  void onUnitHide(BWAPI::Unit*) {}
  void onUnitCreate(BWAPI::Unit*) {}
  /**
   * @brief Notification when a unit is destroyed. Note that this is called AFTER all other components have
   *        been notified, allowing them use of the UnitData associated with the destroyed unit.
   * @param unit The newly destroyed unit.
   */
  void onUnitDestroy(BWAPI::Unit* unit);
  void onUnitMorph(BWAPI::Unit* unit)
  {
    UnitData* unitData = reinterpret_cast<UnitData*>(unit->getClientInfo());
    if (unitData != NULL)
    {
      unitData->m_individualUnitMorphEvent.raise();
    }
  }
  void onUnitRenegade(BWAPI::Unit* unit)
  {
    UnitData* unitData = reinterpret_cast<UnitData*>(unit->getClientInfo());
    if (unitData != NULL)
    {
      unitData->m_individualUnitRenegadeEvent.raise();
    }
  }

  UnitData& getUnitData(BWAPI::Unit* unit);

  PrimitiveLocation* getUnitLocation(BWAPI::Unit* unit)
  {
    return getUnitData(unit).m_location;
  }

  void setUnitLocation(BWAPI::Unit* unit, PrimitiveLocation* location, void* data)
  {
    UnitData& unitData = getUnitData(unit);
    unitData.m_location = location;
    unitData.m_data = data;
  }

  typedef UnitData::OnIndividualUnitDestroy OnIndividualUnitDestroy;
  OnIndividualUnitDestroy::SubscriberID subscribeIndividualUnitDestroy(BWAPI::Unit* unit, OnIndividualUnitDestroy::EventHandlerType* eventHandler, void* data)
  {
    return getUnitData(unit).m_individualUnitDestroyEvent.subscribe(eventHandler, data);
  }
  void unsubscribeIndividualUnitDestroy(BWAPI::Unit* unit, OnIndividualUnitDestroy::SubscriberID& id)
  {
    UnitData* unitData = reinterpret_cast<UnitData*>(unit->getClientInfo());
    if (unitData != NULL)
    {
      unitData->m_individualUnitDestroyEvent.unsubscribe(id);
    }
  }

  typedef UnitData::OnIndividualUnitEvade OnIndividualUnitEvade;
  OnIndividualUnitEvade::SubscriberID subscribeIndividualUnitEvade(BWAPI::Unit* unit, OnIndividualUnitEvade::EventHandlerType* eventHandler, void* data)
  {
    return getUnitData(unit).m_individualUnitEvadeEvent.subscribe(eventHandler, data);
  }
  void unsubscribeIndividualUnitEvade(BWAPI::Unit* unit, OnIndividualUnitEvade::SubscriberID& id)
  {
    UnitData* unitData = reinterpret_cast<UnitData*>(unit->getClientInfo());
    if (unitData != NULL)
    {
      unitData->m_individualUnitEvadeEvent.unsubscribe(id);
    }
  }

  typedef UnitData::OnIndividualUnitMorph OnIndividualUnitMorph;
  OnIndividualUnitMorph::SubscriberID subscribeIndividualUnitMorph(BWAPI::Unit* unit, OnIndividualUnitMorph::EventHandlerType* eventHandler, void* data)
  {
    return getUnitData(unit).m_individualUnitMorphEvent.subscribe(eventHandler, data);
  }
  void unsubscribeIndividualUnitMorph(BWAPI::Unit* unit, OnIndividualUnitMorph::SubscriberID& id)
  {
    UnitData* unitData = reinterpret_cast<UnitData*>(unit->getClientInfo());
    if (unitData != NULL)
    {
      unitData->m_individualUnitMorphEvent.unsubscribe(id);
    }
  }

  typedef UnitData::OnIndividualUnitRenegade OnIndividualUnitRenegade;
  OnIndividualUnitRenegade::SubscriberID subscribeIndividualUnitRenegade(BWAPI::Unit* unit, OnIndividualUnitRenegade::EventHandlerType* eventHandler, void* data)
  {
    return getUnitData(unit).m_individualUnitRenegadeEvent.subscribe(eventHandler, data);
  }
  void unsubscribeIndividualUnitRenegade(BWAPI::Unit* unit, OnIndividualUnitRenegade::SubscriberID& id)
  {
    UnitData* unitData = reinterpret_cast<UnitData*>(unit->getClientInfo());
    if (unitData != NULL)
    {
      unitData->m_individualUnitRenegadeEvent.unsubscribe(id);
    }
  }

protected:
  virtual UnitData* instantiateUnitData(BWAPI::Unit* unit) const = 0;

private:
  // Optimization: chain the UnitData to each other?
  std::list<UnitData*> m_unitDataList;
};