#pragma once
#include "../../Utils/Component.h"
#include "../Event.h"
#include <iostream>

/**
 * @brief Virtual interface for FSM states.
 */
class IFSMState : public Component
{
public:
  virtual ~IFSMState()
  {
  }

  /**
   * @brief This method is called when the FSM must be terminated due to an external event
   * (for example a unit FSM may be destroyed when its associated unit dies).
   * All registered events or dependent FSMs must be taken care of.
   */
  virtual void externalTerminate() = 0;

  /**
   * @return the name of this state.
   */
  virtual std::string getName() const = 0;

  DECLARE_EVENT(OnStateExit);
  OnStateExit& getStateExitEvent() { return m_stateExitEvent; }

  friend std::ostream& operator<<(std::ostream& out, const IFSMState& state)
  {
    out << state.getName();
    return out;
  }

protected:
  EVENT(OnStateExit) m_stateExitEvent;
};
