#pragma once
#include "../PersistentUnitGroup.h"
#include <BWAPI.h>

/**
 * @brief A group process is a process controlling a group of units.
 */
class GroupProcess : public virtual Process
{
public:
  GroupProcess(const Process* parent, const PersistentUnitGroup* group)
    : Process(parent)
    , m_group(group, *this)
    , m_unitAddedEventHandler(*this)
    , m_unitRemovedEventHandler(*this)
  {
  }

  ~GroupProcess()
  {
    m_group->getUnitAddedEvent().unsubscribe(m_unitAddedID);
    m_group->getUnitRemovedEvent().unsubscribe(m_unitRemovedID);
  }

  const PersistentUnitGroup* getGroup() const
  {
    return m_group;
  }
  
protected:
  virtual void onUnitAddedToGroup(void*, BWAPI::Unit* unit) = 0;
  virtual void onUnitRemovedFromGroup(void*, BWAPI::Unit* unit)= 0;

  virtual void executeImpl()
  {
    // WARNING: Process::executeImpl() is intended to be called by the derived class!

    m_unitAddedID = m_group->getUnitAddedEvent().subscribe(&m_unitAddedEventHandler, NULL);
    m_unitRemovedID = m_group->getUnitRemovedEvent().subscribe(&m_unitRemovedEventHandler, NULL);
  }

  virtual void terminateImpl()
  {
    // WARNING: Process::terminateImpl() is intended to be called by the derived class!

    m_group->getUnitAddedEvent().unsubscribe(m_unitAddedID);
    m_group->getUnitRemovedEvent().unsubscribe(m_unitRemovedID);
  }

private:
  EVENT_HANDLER1(GroupProcess, onUnitAddedToGroup, BWAPI::Unit*) m_unitAddedEventHandler;
  PersistentUnitGroup::OnUnitAdded::SubscriberID m_unitAddedID;
  EVENT_HANDLER1(GroupProcess, onUnitRemovedFromGroup, BWAPI::Unit*) m_unitRemovedEventHandler;
  PersistentUnitGroup::OnUnitRemoved::SubscriberID m_unitRemovedID;

  // Derived classes can access the (const) group via getGroup()
  /**
   * @brief The units currently assigned to this group process.
   */
  AutoPtr<const PersistentUnitGroup> m_group;
};