#include "stdafx.h"
#include "Logger.h"
#include <Config.h>

#include <sstream>
#undef min
#undef max
#include <io.h>
#include <fcntl.h>

/*
FOREGROUND_BLUE	Text color contains blue.
FOREGROUND_GREEN	Text color contains green.
FOREGROUND_RED	Text color contains red.
FOREGROUND_INTENSITY	Text color is intensified.
BACKGROUND_BLUE	Background color contains blue.
BACKGROUND_GREEN	Background color contains green.
BACKGROUND_RED	Background color contains red.
BACKGROUND_INTENSITY	Background color is intensified.
COMMON_LVB_LEADING_BYTE	Leading byte.
COMMON_LVB_TRAILING_BYTE	Trailing byte.
COMMON_LVB_GRID_HORIZONTAL	Top horizontal.
COMMON_LVB_GRID_LVERTICAL	Left vertical.
COMMON_LVB_GRID_RVERTICAL	Right vertical.
COMMON_LVB_REVERSE_VIDEO	Reverse foreground and background attributes.
COMMON_LVB_UNDERSCORE	Underscore.
*/

namespace LogColors
{
  static const WORD
    DARK_RED      = FOREGROUND_RED,
    DARK_GREEN    = FOREGROUND_GREEN,
    DARK_BLUE     = FOREGROUND_BLUE,
    LIGHT_RED     = FOREGROUND_INTENSITY | DARK_RED,
    LIGHT_GREEN   = FOREGROUND_INTENSITY | DARK_GREEN,
    LIGHT_BLUE    = FOREGROUND_INTENSITY | DARK_BLUE,

    DARK_YELLOW   = DARK_RED | DARK_GREEN,
    DARK_MAGENTA  = DARK_RED | DARK_BLUE,
    DARK_CYAN     = DARK_BLUE | DARK_GREEN,
    LIGHT_YELLOW  = FOREGROUND_INTENSITY | DARK_YELLOW,
    LIGHT_MAGENTA = FOREGROUND_INTENSITY | DARK_MAGENTA,
    LIGHT_CYAN    = FOREGROUND_INTENSITY | DARK_CYAN,

    BLACK         = 0,
    GREY          = DARK_RED | DARK_GREEN | DARK_BLUE,
    WHITE         = FOREGROUND_INTENSITY | GREY;
};

const WORD LogTypes::EventType::DEFAULT_COLOR = LogColors::GREY;
const unsigned int LogTypes::EventType::DEFAULT_HTML_COLOR = 0xcccccc;

const LogTypes::EventType LogTypes::ON_FRAME(0x1);
const LogTypes::EventType LogTypes::ON_MORPH(0x2);
const LogTypes::EventType LogTypes::ON_RENEGADE(0x4);
const LogTypes::EventType LogTypes::ON_DESTROY(0x8);
const LogTypes::EventType LogTypes::ON_DISCOVER(0x10);
const LogTypes::EventType LogTypes::ON_EVADE(0x20);
const LogTypes::EventType LogTypes::ON_SHOW(0x40);
const LogTypes::EventType LogTypes::ON_HIDE(0x80);
const LogTypes::EventType LogTypes::ON_CREATE(0x100);
const LogTypes::EventType LogTypes::GENERAL(0x200);
const LogTypes::EventType LogTypes::LAYER_1_DECISION(0x400, LogColors::LIGHT_BLUE, 0x3333cc);
const LogTypes::EventType LogTypes::LAYER_2_DECISION(0x800, LogColors::LIGHT_YELLOW, 0xffff66);
const LogTypes::EventType LogTypes::LAYER_3_DECISION(0x1000);
const LogTypes::EventType LogTypes::LAYER_4_DECISION(0x2000);
const LogTypes::EventType LogTypes::LAYER_1_SITUATION_ANALYSIS(0x4000);
const LogTypes::EventType LogTypes::LAYER_2_SITUATION_ANALYSIS(0x8000);
const LogTypes::EventType LogTypes::LAYER_3_SITUATION_ANALYSIS(0x10000);
const LogTypes::EventType LogTypes::LAYER_4_SITUATION_ANALYSIS(0x20000);
const LogTypes::EventType LogTypes::PERCEPTUAL_STATE(0x40000, LogColors::DARK_GREEN, 0x339933);
const LogTypes::EventType LogTypes::GENERAL_ERROR(0x80000, LogColors::LIGHT_RED, 0xcc3333);

const unsigned int LogTypes::CALLBACKS = ON_FRAME | ON_DESTROY | ON_MORPH | ON_RENEGADE | ON_DISCOVER | 
                                         ON_EVADE | ON_SHOW | ON_HIDE | ON_CREATE;
const unsigned int LogTypes::LAYERS = LAYER_1_DECISION | LAYER_2_DECISION | LAYER_3_DECISION | LAYER_4_DECISION |
                                      LAYER_1_SITUATION_ANALYSIS | LAYER_2_SITUATION_ANALYSIS | LAYER_3_SITUATION_ANALYSIS | LAYER_4_SITUATION_ANALYSIS;
const unsigned int LogTypes::ALL = CALLBACKS | LAYERS | GENERAL | PERCEPTUAL_STATE | GENERAL_ERROR;

#ifdef USE_SPAR_LOG

SparLogger SparLogger::s_logger(Config::getDefaultLoggerEvents());

SparLogger::SparLogger(int enabledLogger)
  : m_enabledLogger(enabledLogger)
{
  if(fopen_s(&m_sparLogFile, "bwapi-data/logs/SPAR.log.htm", "w") !=0)
    throw std::runtime_error(std::string("Cannot open file \"") + "bwapi-data/logs/SPAR.log" + "\".");
  fprintf(m_sparLogFile, "<body bgcolor=\"#000000\">\n");
  CreateConsoleLog("SPAR Logger");
}

SparLogger::~SparLogger()
{
  fprintf(m_sparLogFile, "</body>");
  fclose(m_sparLogFile);
}

void SparLogger::CreateConsoleLog(const char *winTitle)
{
  AllocConsole();
  SetConsoleTitleA(winTitle);
  int hConHandle;
  HANDLE lStdHandle;
  FILE *fp;	// redirect unbuffered STDOUT to the console	
  lStdHandle = GetStdHandle(STD_OUTPUT_HANDLE);
  hConHandle = _open_osfhandle((long)lStdHandle, _O_TEXT);
  fp = _fdopen( hConHandle, "w" );
  *stdout = *fp;
  setvbuf( stdout, NULL, _IONBF, 0 );	// redirect unbuffered STDIN to the console	
  lStdHandle = GetStdHandle(STD_INPUT_HANDLE);
  hConHandle = _open_osfhandle((long)lStdHandle, _O_TEXT);
  fp = _fdopen( hConHandle, "r" );
  *stdin = *fp;
  setvbuf( stdin, NULL, _IONBF, 0 );	// redirect unbuffered STDERR to the console	
  lStdHandle = GetStdHandle(STD_ERROR_HANDLE);
  hConHandle = _open_osfhandle((long)lStdHandle, _O_TEXT);
  fp = _fdopen( hConHandle, "w" );	
  *stderr = *fp;
  setvbuf( stderr, NULL, _IONBF, 0 );
}

#endif