#pragma once
#include <stdio.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef min
#undef max
#include <time.h>

struct LogTypes
{
  class EventType
  {
  public:
    const unsigned int code;
    const WORD color;
    static const WORD DEFAULT_COLOR;
    const unsigned int htmlColor;
    static const unsigned int DEFAULT_HTML_COLOR;
    operator unsigned int() const 
    {
      return code;
    }
  protected:
    friend struct LogTypes;
    EventType(unsigned int _code, WORD _color, unsigned int _htmlColor)
      : code(_code), color(_color), htmlColor(_htmlColor)
    {
    }
    EventType(unsigned int _code)
      : code(_code), color(DEFAULT_COLOR), htmlColor(DEFAULT_HTML_COLOR)
    {
    }
  };

  static const EventType
    ON_FRAME,
    ON_MORPH,
    ON_RENEGADE,
    ON_DESTROY,
    ON_DISCOVER,
    ON_EVADE,
    ON_SHOW,
    ON_HIDE,
    ON_CREATE,
    GENERAL,
    LAYER_1_DECISION,
    LAYER_2_DECISION,
    LAYER_3_DECISION,
    LAYER_4_DECISION,
    LAYER_1_SITUATION_ANALYSIS,
    LAYER_2_SITUATION_ANALYSIS,
    LAYER_3_SITUATION_ANALYSIS,
    LAYER_4_SITUATION_ANALYSIS,
    PERCEPTUAL_STATE,
    GENERAL_ERROR; // Can't name that ERROR due to stupid windows.h

  static const unsigned int
    CALLBACKS,
    LAYERS,
    ALL;
};

class SparLogger
{
private:
  FILE * m_sparLogFile;
  unsigned int m_enabledLogger;

  static SparLogger s_logger;
  SparLogger(int enabledLogger);
  ~SparLogger();
  void CreateConsoleLog(const char *winTitle);
public:
  static SparLogger* get()
  {
    return &s_logger;
  }
  FILE * getFile()
  {
    return m_sparLogFile;
  }
  void activate(unsigned int mask)
  {
    m_enabledLogger |= mask;
  }
  void deactivate(unsigned int mask)
  {
    m_enabledLogger &= ~mask;
  }
  bool isActive(unsigned int mask)
  {
    return (m_enabledLogger & mask) == mask;
  }
};

#ifdef USE_SPAR_LOG
  #define SPAR_LOGGER_ACTIVATE_EVENT(mask) SparLogger::get()->activate(mask)
  #define SPAR_LOGGER_DEACTIVATE_EVENT(mask) SparLogger::get()->deactivate(mask)
  #define SPAR_LOG(eventType, format, ...) \
    do{ \
      if(SparLogger::get()->isActive(eventType)) \
      { \
        time_t rawtime; \
        time(&rawtime); \
        struct tm timeinfo; \
        localtime_s(&timeinfo, &rawtime); \
        char buffer [80]; \
        strftime(buffer, 80, "%H:%M:%S", &timeinfo); \
        SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), LogTypes::EventType::DEFAULT_COLOR); \
        printf("%s (#%d) - ", buffer, BWAPI::Broodwar->getFrameCount()); \
        SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), eventType.color); \
        printf(format, __VA_ARGS__); \
        printf("\n"); \
        fprintf(SparLogger::get()->getFile(), "<font color=\"#%x\">%s (#%d) - </font><font color=\"#%x\">", \
                                                  LogTypes::EventType::DEFAULT_HTML_COLOR, \
                                                  buffer, \
                                                  BWAPI::Broodwar->getFrameCount(), \
                                                  eventType.htmlColor); \
        fprintf(SparLogger::get()->getFile(), format, __VA_ARGS__); \
        fprintf(SparLogger::get()->getFile(), "</font></br>\n"); \
      } \
    } while(false)
#else
  #define SPAR_LOGGER_ACTIVATE_EVENT(type) 
  #define SPAR_LOGGER_DEACTIVATE_EVENT(type) 
  #define SPAR_LOG(format, ...) 
#endif
