#pragma once

class Component;

/**
 * @brief A task is anything that can be run. Note that this class should be named simply "Task" but that name 
 *        would clash with BWSAL's tasks.
 */
class RunnableTask
{
public:
  /**
   * @brief Task constructor.
   * @param owner The owner of the task.
   */
  RunnableTask(const Component& owner)
#ifdef _DEBUG_COMPONENT
    : m_owner(owner)
#endif
  {
    unused(owner);
  }

  virtual ~RunnableTask() {}

  /**
   * @brief Non-virtual method to run the task.
   */
  void run()
  {
    runV();
  }

#ifdef _DEBUG_COMPONENT
  /**
   * @brief The owner of the task.
   */
  const Component& m_owner;
#endif

protected:
  /**
   * @brief Virtual method to run the task.
   */
  virtual void runV() = 0;

private:
  RunnableTask(const RunnableTask& task);
};