#include <Config.h>
#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
#include <windows.h>
#include "Scenarios/Scenarios.h"
#include "Instantiate.h"
#include <SparAIModule.h>
#include <SituationAnalysis/Layer1GameState/Layer1GameState.h>
#include <SituationAnalysis/Layer2Cueing/Layer2Cueing.h>
#include <SituationAnalysis/Layer3TacticalRecognition/Layer3TacticalRecognition.h>
#include <SituationAnalysis/Layer4StrategicRecognition/Layer4StrategicRecognition.h>
#include <DecisionMaking/Layer1Reactive/FSM/Layer1FSM.h>
#include <DecisionMaking/Layer2Actions/Managers/Layer2Managers.h>
#include <DecisionMaking/Layer3Tactical/Script/Layer3Script.h>
#include <DecisionMaking/Layer4Strategic/Heuristic/Layer4Heuristic.h>

BOOL APIENTRY DllMain( HINSTANCE hModule, 
                      DWORD  ul_reason_for_call, 
                      LPVOID lpReserved
                      )
{
  switch (ul_reason_for_call)
  {
  case DLL_PROCESS_ATTACH:
    BWAPI::BWAPI_init();
    break;
  case DLL_PROCESS_DETACH:
    // DO NOT DO ANYTHING IN HERE, BECAUSE APPARENTLY NOTHING WORKS
#ifdef _DEBUG_REFERENCE_COUNTED
    ReferenceCounted::signalMemoryLeaks();
#endif
    break;
  }
  return TRUE;
}

extern "C" __declspec(dllexport) BWAPI::AIModule* newAIModule(BWAPI::Game* game)
{
  BWAPI::Broodwar=game;
  SparAIContainer* module = Scenarios::instantiateSparAIForScenarios();
  if (module == NULL)
  {
    typedef Layer1GameState Layer1SituationAnalysis;
    typedef Layer2Cueing<Layer1GameState> Layer2SituationAnalysis;
    typedef Layer3TacticalRecognition<Layer2SituationAnalysis> Layer3SituationAnalysis;
    typedef Layer4StrategicRecognition<Layer3SituationAnalysis> Layer4SituationAnalysis;

    // Layer2Managers and Layer3Script should actually be templated
    typedef Layer1FSM Layer1DecisionMaking;
    typedef Layer2Managers Layer2DecisionMaking;
    typedef Layer3Script Layer3DecisionMaking;
    typedef Layer4Heuristic Layer4DecisionMaking;
  
    typedef SparAIModule<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                         Layer1DecisionMaking, Layer2DecisionMaking, Layer3DecisionMaking, Layer4DecisionMaking> SPAR;

    module = instantiate<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                         Layer1DecisionMaking, Layer2DecisionMaking, Layer3DecisionMaking, Layer4DecisionMaking>();

	if(!BWAPI::Broodwar->isMultiplayer())
		BWAPI::Broodwar->setLocalSpeed(module->getConfiguration().getLocalSpeedForSinglePlayerGames());
  }

  return module;
}
