#include "../Instantiate.h"
#include "Scenarios.h"
#include <Scenarios/ScenariosBehaviorExecutor.h>
#include <Scenarios/Layer1ScoutScenario.h>
#include <Scenarios/InactiveLayer1.h>
#include <Scenarios/InactiveLayer2.h>
#include <Scenarios/InactiveLayer3.h>
#include <Scenarios/InactiveLayer4.h>
#include <SparAIModule.h>
#include <SituationAnalysis/Layer1GameState/Layer1GameState.h>
#include <SituationAnalysis/Layer2Cueing/Layer2Cueing.h>
#include <SituationAnalysis/Layer3TacticalRecognition/Layer3TacticalRecognition.h>
#include <SituationAnalysis/Layer4StrategicRecognition/Layer4StrategicRecognition.h>
#include <DecisionMaking/Layer1Reactive/FSM/Old/AvoidEnemies.h>
#include <DecisionMaking/Layer1Reactive/FSM/Old/FinitePatrol.h>
#include <DecisionMaking/Layer1Reactive/FSM/Old/Wander.h>
#include <DecisionMaking/Layer1Reactive/FSM/Old/ObserverScout.h>
#include <DecisionMaking/Layer1Reactive/FSM/Old/EngageEnemies/GroupCarefulAttack/GroupCarefulAttack.h>
#include <DecisionMaking/Layer1Reactive/FSM/Behaviors/Scout.h>
#include <DecisionMaking/Layer1Reactive/FSM/Behaviors/EngageEnemies/GroupAttack.h>
#include <DecisionMaking/Layer1Reactive/FSM/Behaviors/ReaverDrop/ReaverControl.h>
#include <DecisionMaking/Layer1Reactive/FSM/Behaviors/ReaverDrop/ShuttleControl.h>
#include <DecisionMaking/Layer1Reactive/FSM/Behaviors/ReaverDrop/ReaverShuttleGroup.h>

SparAIContainer* Scenarios::instantiateAITestCallbacks()
{
  class CustomLayer1DecisionMaking : public InactiveLayer1<Layer1GameState>
  {
  public:
    CustomLayer1DecisionMaking(Layer1GameState& layer1GameState)
      : InactiveLayer1(layer1GameState)
      , m_lastFrame(-1)
      , m_eventHandler(*this)
      , m_enemyUnitEventHandler(*this)
    {
    }

    virtual void onStart()
    {
      InactiveLayer1::onStart();
      Spar->getPerceptualState().getUnitDiscoverEvent().subscribe(&m_eventHandler, NULL);
      Spar->getPerceptualState().getUnitEvadeEvent().subscribe(&m_eventHandler, NULL);
      Spar->getPerceptualState().getUnitShowEvent().subscribe(&m_eventHandler, NULL);
      Spar->getPerceptualState().getUnitHideEvent().subscribe(&m_eventHandler, NULL);
      Spar->getPerceptualState().getUnitCreateEvent().subscribe(&m_eventHandler, NULL);
      Spar->getPerceptualState().getUnitDestroyEvent().subscribe(&m_eventHandler, NULL);
      Spar->getPerceptualState().getUnitMorphEvent().subscribe(&m_eventHandler, NULL);
      Spar->getPerceptualState().getUnitRenegadeEvent().subscribe(&m_eventHandler, NULL);

      Spar->getPerceptualState().getMap()->getEnemyUnitAddedEvent().subscribe(&m_enemyUnitEventHandler, NULL);
      Spar->getPerceptualState().getMap()->getEnemyUnitRemovedEvent().subscribe(&m_enemyUnitEventHandler, NULL);
    }
    
    void onEvent(void*, BWAPI::Unit*)
    {
      m_lastFrame = BWAPI::Broodwar->getFrameCount();
    }
    void onEnemyUnitEvent(void*, EnemyUnit*)
    {
      m_lastFrame = BWAPI::Broodwar->getFrameCount();
    }

    virtual void onFrame()
    {
      InactiveLayer1::onFrame();
      if (m_lastFrame != -1)
      {
        m_lastFrame = -1;
        const Map* map = Spar->getPerceptualState().getMap();

        SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "List of own units:");
        std::pair<Map::own_unit_const_iterator, Map::own_unit_const_iterator> ownBuildings = map->getOwnUnits();
        for (Map::own_unit_const_iterator it = ownBuildings.first; it != ownBuildings.second; ++it)
        {
          const OwnUnit* ownUnit = *it;
          SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "  %s", getObjectDescription(ownUnit).c_str());
        }

        SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "List of enemy units:");
        std::pair<Map::enemy_unit_const_iterator, Map::enemy_unit_const_iterator> enemyBuildings = map->getEnemyUnits();
        for (Map::enemy_unit_const_iterator it = enemyBuildings.first; it != enemyBuildings.second; ++it)
        {
          const EnemyUnit* enemyUnit = *it;
          SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "  %s", getObjectDescription(enemyUnit).c_str());
        }

        SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "List of neutral units:");
        std::pair<Map::neutral_unit_const_iterator, Map::neutral_unit_const_iterator> neutralUnits = map->getNeutralUnits();
        for (Map::neutral_unit_const_iterator it = neutralUnits.first; it != neutralUnits.second; ++it)
        {
          const NeutralUnit* neutralUnit = *it;
          SPAR_LOG(LogTypes::PERCEPTUAL_STATE, "  %s", getObjectDescription(neutralUnit).c_str());
        }
      }
    }

  protected:
    int m_lastFrame;
    EVENT_HANDLER1(CustomLayer1DecisionMaking, onEvent, BWAPI::Unit*) m_eventHandler;
    EVENT_HANDLER1(CustomLayer1DecisionMaking, onEnemyUnitEvent, EnemyUnit*) m_enemyUnitEventHandler;
  };

  typedef Layer1GameState Layer1SituationAnalysis;
  typedef Layer2Cueing<Layer1GameState> Layer2SituationAnalysis;
  typedef Layer3TacticalRecognition<Layer2SituationAnalysis> Layer3SituationAnalysis;
  typedef Layer4StrategicRecognition<Layer3SituationAnalysis> Layer4SituationAnalysis;

  typedef CustomLayer1DecisionMaking InactiveLayer1DecisionMaking;
  typedef InactiveLayer2<Layer2SituationAnalysis, InactiveLayer1DecisionMaking> InactiveLayer2DecisionMaking;
  typedef InactiveLayer3<Layer3SituationAnalysis, InactiveLayer2DecisionMaking> InactiveLayer3DecisionMaking;
  typedef InactiveLayer4<Layer4SituationAnalysis, InactiveLayer3DecisionMaking> InactiveLayer4DecisionMaking;

  typedef SparAIModule<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                       InactiveLayer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking> SPAR;

  SPAR* module = instantiate<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                             InactiveLayer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking>();

  BWAPI::Broodwar->setLocalSpeed(30);

  return module;
}

SparAIContainer* Scenarios::instantiateAIModuleScenario1()
{
  typedef Layer1GameState Layer1SituationAnalysis;
  typedef Layer2Cueing<Layer1GameState> Layer2SituationAnalysis;
  typedef Layer3TacticalRecognition<Layer2SituationAnalysis> Layer3SituationAnalysis;
  typedef Layer4StrategicRecognition<Layer3SituationAnalysis> Layer4SituationAnalysis;

  typedef Layer1FSM Layer1DecisionMaking;
  typedef InactiveLayer2<Layer2SituationAnalysis, Layer1DecisionMaking> InactiveLayer2DecisionMaking;
  typedef InactiveLayer3<Layer3SituationAnalysis, InactiveLayer2DecisionMaking> InactiveLayer3DecisionMaking;
  typedef InactiveLayer4<Layer4SituationAnalysis, InactiveLayer3DecisionMaking> InactiveLayer4DecisionMaking;

  typedef SparAIModule<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                       Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking> SPAR;

  SPAR* module = instantiate<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                             Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking>();

  std::set<BWAPI::Unit*> units = BWAPI::Broodwar->self()->getUnits();
  // AvoidEnemies
  for (std::set<BWAPI::Unit*>::iterator it = units.begin();
       it != units.end();
       ++it)
  {
    BWAPI::Unit* unit = (*it);
    if (unit->getType() == BWAPI::UnitTypes::Terran_Medic)
    {
      ScenariosBehaviorExecutor::execute(module->getD1(), 
        new Wander(ScenariosBehaviorExecutor::getParent(), 100, unit));
    }
    ScenariosBehaviorExecutor::execute(module->getD1(), 
      new AvoidEnemies(ScenariosBehaviorExecutor::getParent(), 1000, unit));
  }
  return module;
}

SparAIContainer* Scenarios::instantiateAIModuleScenario2()
{
  typedef Layer1GameState Layer1SituationAnalysis;
  typedef Layer2Cueing<Layer1GameState> Layer2SituationAnalysis;
  typedef Layer3TacticalRecognition<Layer2SituationAnalysis> Layer3SituationAnalysis;
  typedef Layer4StrategicRecognition<Layer3SituationAnalysis> Layer4SituationAnalysis;

  typedef Layer1FSM Layer1DecisionMaking;
  typedef InactiveLayer2<Layer2SituationAnalysis, Layer1DecisionMaking> InactiveLayer2DecisionMaking;
  typedef InactiveLayer3<Layer3SituationAnalysis, InactiveLayer2DecisionMaking> InactiveLayer3DecisionMaking;
  typedef InactiveLayer4<Layer4SituationAnalysis, InactiveLayer3DecisionMaking> InactiveLayer4DecisionMaking;

  typedef SparAIModule<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                       Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking> SPAR;

  SPAR* module = instantiate<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                             Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking>();
  Spar = module;
  std::set<BWAPI::Unit*> units = BWAPI::Broodwar->self()->getUnits();
  // FinitePatrol
  PersistentUnitGroup* group = new PersistentUnitGroup("Initial group", units.begin(), units.end());
  //std::list<BWAPI::Position> checkPoints;
  //checkPoints.push_back(BWAPI::Position(100, 100));
  //checkPoints.push_back(BWAPI::Position(500, 100));
  ScenariosBehaviorExecutor::execute(module->getD1(), 
    //new FinitePatrol(ScenariosBehaviorExecutor::getParent(), 150, group, checkPoints));
    new GroupAttack(ScenariosBehaviorExecutor::getParent(), 150, group, new CoordinatesLocation(BWAPI::Position(100, 100))));

  return module;
}

SparAIContainer* Scenarios::instantiateAIModuleScenarioReaver()
{
  typedef Layer1GameState Layer1SituationAnalysis;
  typedef Layer2Cueing<Layer1GameState> Layer2SituationAnalysis;
  typedef Layer3TacticalRecognition<Layer2SituationAnalysis> Layer3SituationAnalysis;
  typedef Layer4StrategicRecognition<Layer3SituationAnalysis> Layer4SituationAnalysis;

  typedef Layer1FSM Layer1DecisionMaking;
  typedef InactiveLayer2<Layer2SituationAnalysis, Layer1DecisionMaking> InactiveLayer2DecisionMaking;
  typedef InactiveLayer3<Layer3SituationAnalysis, InactiveLayer2DecisionMaking> InactiveLayer3DecisionMaking;
  typedef InactiveLayer4<Layer4SituationAnalysis, InactiveLayer3DecisionMaking> InactiveLayer4DecisionMaking;

  typedef SparAIModule<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                       Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking> SPAR;

  SPAR* module = instantiate<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                             Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking>();

  std::set<BWAPI::Unit*> units = BWAPI::Broodwar->self()->getUnits();
  BWAPI::Unit* reaver[15];
  int k = 0;
  // Reaver
  for (std::set<BWAPI::Unit*>::iterator it = units.begin();
       it != units.end();
       ++it)
  {
    BWAPI::Unit* unit = (*it);
    if (unit->getType() == BWAPI::UnitTypes::Protoss_Reaver){
      ScenariosBehaviorExecutor::execute(module->getD1(), 
        new ReaverControl(ScenariosBehaviorExecutor::getParent(), 100, unit));
      reaver[k] = unit;
      k++;
    }
  }
  k=0;
  // Shuttle
  for (std::set<BWAPI::Unit*>::iterator it = units.begin();
       it != units.end();
       ++it)
  {
    BWAPI::Unit* unit = (*it);
    if (unit->getType() == BWAPI::UnitTypes::Protoss_Shuttle){
      ScenariosBehaviorExecutor::execute(module->getD1(), 
        new ShuttleControl(ScenariosBehaviorExecutor::getParent(), 100, unit,reaver[k]));
      k++;
    }
  }
  return module;
}

SparAIContainer* Scenarios::instantiateAIModuleScoutScenario()
{
  typedef Layer1GameState Layer1SituationAnalysis;
  typedef Layer2Cueing<Layer1GameState> Layer2SituationAnalysis;
  typedef Layer3TacticalRecognition<Layer2SituationAnalysis> Layer3SituationAnalysis;
  typedef Layer4StrategicRecognition<Layer3SituationAnalysis> Layer4SituationAnalysis;

  typedef Layer1ScoutScenario Layer1DecisionMaking;
  typedef InactiveLayer2<Layer2SituationAnalysis, Layer1DecisionMaking> InactiveLayer2DecisionMaking;
  typedef InactiveLayer3<Layer3SituationAnalysis, InactiveLayer2DecisionMaking> InactiveLayer3DecisionMaking;
  typedef InactiveLayer4<Layer4SituationAnalysis, InactiveLayer3DecisionMaking> InactiveLayer4DecisionMaking;

  typedef SparAIModule<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                       Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking> SPAR;

  SPAR* module = instantiate<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                             Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking>();
  Spar = module;
  return module;
}

SparAIContainer* Scenarios::instantiateAIModuleObserverScenario()
{
  typedef Layer1GameState Layer1SituationAnalysis;
  typedef Layer2Cueing<Layer1GameState> Layer2SituationAnalysis;
  typedef Layer3TacticalRecognition<Layer2SituationAnalysis> Layer3SituationAnalysis;
  typedef Layer4StrategicRecognition<Layer3SituationAnalysis> Layer4SituationAnalysis;

  typedef Layer1FSM Layer1DecisionMaking;
  typedef InactiveLayer2<Layer2SituationAnalysis, Layer1DecisionMaking> InactiveLayer2DecisionMaking;
  typedef InactiveLayer3<Layer3SituationAnalysis, InactiveLayer2DecisionMaking> InactiveLayer3DecisionMaking;
  typedef InactiveLayer4<Layer4SituationAnalysis, InactiveLayer3DecisionMaking> InactiveLayer4DecisionMaking;

  typedef SparAIModule<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                       Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking> SPAR;

  SPAR* module = instantiate<Layer1SituationAnalysis, Layer2SituationAnalysis, Layer3SituationAnalysis, Layer4SituationAnalysis,
                             Layer1DecisionMaking, InactiveLayer2DecisionMaking, InactiveLayer3DecisionMaking, InactiveLayer4DecisionMaking>();
  Spar = module;
  std::set<BWAPI::Unit*> units = BWAPI::Broodwar->self()->getUnits();
  std::set<BWAPI::TilePosition> positions = BWAPI::Broodwar->getStartLocations();
  std::set<BWAPI::TilePosition>::iterator posIt = positions.begin();
  ++posIt;
  BWAPI::Position pos((*posIt).x() * TILE_SIZE, (*posIt).y() * TILE_SIZE);

  // Scout
  for (std::set<BWAPI::Unit*>::iterator it = units.begin();
       it != units.end();
       ++it)
  {
  //std::set<BWAPI::Unit*>::iterator it = units.begin();
    
    BWAPI::Unit* unit = (*it);
    
    //ScenariosBehaviorExecutor::execute(module->getD1(), new Wander(100, unit));
    if (unit->getType() == BWAPI::UnitTypes::Protoss_Observer){
      ScenariosBehaviorExecutor::execute(module->getD1(), 
        new ObserverScout(ScenariosBehaviorExecutor::getParent(), 100, pos, unit));
    }
  }

  return module;
}