#include "stdafx.h"
#include "mainwindow.h"
#include <sstream>

bool MainWindow::s_initialized(false);
int MainWindow::UpdateCurrentProcessesEventType;
int MainWindow::UpdatePerceptualStateEventType;
int MainWindow::UpdateLayer3EventType;

void MainWindow::initialize()
{
  assert(!s_initialized);
  s_initialized = true;

  UpdateCurrentProcessesEventType = QEvent::registerEventType();
  UpdatePerceptualStateEventType = QEvent::registerEventType();
  UpdateLayer3EventType = QEvent::registerEventType();
}

MainWindow::MainWindow(std::set<BWAPI::Unit*>& units, boost::mutex& unitsMutex)
  : ui(new Ui::MainWindow)
  , m_units(units)
  , m_unitsMutex(unitsMutex)
{
  assert(s_initialized);

  ui->setupUi(this);

  ui->chkMapAnalysis->setChecked(Spar->getConfiguration().displayMap());
  ui->chkEnemyInfluence->setChecked(Spar->getConfiguration().displayEnemyInfluenceMap());
  ui->chkSelfInfluence->setChecked(Spar->getConfiguration().displaySelfInfluenceMap());
  ui->chkVisibilityInfluence->setChecked(Spar->getConfiguration().displayVisibilityInfluenceMap());
  ui->chkStructureInfluence->setChecked(Spar->getConfiguration().displayStructureInfluenceMap());
  ui->chkOwnUnitsActions->setChecked(Spar->getConfiguration().displayOwnUnitsActions());

  QObject::connect(ui->chkMapAnalysis, SIGNAL(stateChanged(int)),
                   this, SLOT(chkMapAnalysisValueChanged(int)));
  QObject::connect(ui->chkEnemyInfluence, SIGNAL(stateChanged(int)),
                   this, SLOT(chkEnemyInfluenceValueChanged(int)));
  QObject::connect(ui->chkSelfInfluence, SIGNAL(stateChanged(int)),
                   this, SLOT(chkSelfInfluenceValueChanged(int)));
  QObject::connect(ui->chkVisibilityInfluence, SIGNAL(stateChanged(int)),
                   this, SLOT(chkVisibilityInfluenceValueChanged(int)));
  QObject::connect(ui->chkStructureInfluence, SIGNAL(stateChanged(int)),
                   this, SLOT(chkStructureInfluenceValueChanged(int)));
  QObject::connect(ui->chkOwnUnitsActions, SIGNAL(stateChanged(int)),
                   this, SLOT(chkOwnUnitsActionsValueChanged(int)));

  const std::set<BWAPI::UnitType>& unitTypes = BWAPI::UnitTypes::allUnitTypes();
  for (std::set<BWAPI::UnitType>::const_iterator it = unitTypes.begin(); it != unitTypes.end(); ++it)
  {
    BWAPI::UnitType type = *it;
    std::stringstream ss;
    ss << type.getID() << ". " << type.getName();
    ui->lstUnitTypes->addItem(QString::fromStdString(ss.str()));
  }

  const std::set<BWAPI::UpgradeType>& upgradeTypes = BWAPI::UpgradeTypes::allUpgradeTypes();
  for (std::set<BWAPI::UpgradeType>::const_iterator it = upgradeTypes.begin(); it != upgradeTypes.end(); ++it)
  {
    BWAPI::UpgradeType type = *it;
    std::stringstream ss;
    ss << type.getID() << ". " << type.getName();
    ui->lstUpgradeTypes->addItem(QString::fromStdString(ss.str()));
  }

  const std::set<BWAPI::TechType>& techTypes = BWAPI::TechTypes::allTechTypes();
  for (std::set<BWAPI::TechType>::const_iterator it = techTypes.begin(); it != techTypes.end(); ++it)
  {
    BWAPI::TechType type = *it;
    std::stringstream ss;
    ss << type.getID() << ". " << type.getName();
    ui->lstTechTypes->addItem(QString::fromStdString(ss.str()));
  }
}

MainWindow::~MainWindow()
{
  delete ui;
}

void MainWindow::updateCurrentProcessesUnits()
{
  // Takes place on the GUI thread

  //boost::mutex::scoped_lock(m_unitsMutex);
  m_unitsMutex.lock();

  m_units.clear();
  QList<QTreeWidgetItem*> selectedItems = ui->trvProcesses->selectedItems();
  for (QList<QTreeWidgetItem*>::const_iterator it = selectedItems.begin(); it != selectedItems.end(); ++it)
  {
    QTreeWidgetItemCustom<const Process*>* item = dynamic_cast<QTreeWidgetItemCustom<const Process*>*>(*it);
    if (item != NULL)
    {
      // TODO: UNSAFE!!! THE PROCESS MAY HAVE BEEN DELETED ONCE THIS THREAD GETS HERE
      const Process* process = item->getCustomData();
      std::set<BWAPI::Unit*> units = process->getCommandedUnits();
      m_units.insert(units.begin(), units.end());
    }
  }

  m_unitsMutex.unlock();
}

std::set<const void*> getSelection(const QTreeWidget* tree)
{
  std::set<const void*> selection;
  QList<QTreeWidgetItem*> selectedItems = tree->selectedItems();
  for (QList<QTreeWidgetItem*>::const_iterator it = selectedItems.begin(); it != selectedItems.end(); ++it)
  {
    QTreeWidgetItemKey* item = dynamic_cast<QTreeWidgetItemKey*>(*it);
    if (item != NULL)
      selection.insert(item->getKey());
  }
  return selection;
}

void select(QTreeWidgetItem* item, const std::set<const void*>& selection)
{
  for (int i = 0; i < item->childCount(); ++i)
  {
    QTreeWidgetItem* child = item->child(i);
    QTreeWidgetItemKey* customChild = dynamic_cast<QTreeWidgetItemKey*>(child);
    if (customChild != NULL)
    {
      if (selection.find(customChild->getKey()) != selection.end())
        customChild->setSelected(true);
    }
    select(child, selection);
  }
}

bool MainWindow::event(QEvent* ev)
{
  // At the moment the controls are directly cleared then refreshed... obviously a better solution would be to
  // modify the model instead.

  if (ev->type() == UpdateCurrentProcessesEventType)
  {
    // TODO: allow double-click to center screen on process units

    ev->accept();
    UpdateCurrentProcessesEvent* processesEvent = static_cast<UpdateCurrentProcessesEvent*>(ev);

    std::set<const void*> selection = getSelection(ui->trvProcesses);

    ui->trvProcesses->clear();
    processesEvent->getRoot()->displayIn(new QTreeWidgetItemCustom<const Process*>(ui->trvProcesses));

    select(ui->trvProcesses->topLevelItem(0), selection);

    updateCurrentProcessesUnits();
  }
  else if (ev->type() == UpdatePerceptualStateEventType)
  {
    // TODO: allow double-click to center screen on location/unit

    ev->accept();
    UpdatePerceptualStateEvent* perceptualStateEvent = static_cast<UpdatePerceptualStateEvent*>(ev);

    std::set<const void*> selection = getSelection(ui->trvLocations);
    ui->trvLocations->clear();
    perceptualStateEvent->getRoot()->displayIn(new QTreeWidgetItemKey(ui->trvLocations));
    select(ui->trvLocations->topLevelItem(0), selection);

    const std::vector<TechWrapper>& techs = perceptualStateEvent->getTechs();
    ui->lstTechs->clear();
    for (std::vector<TechWrapper>::const_iterator it = techs.begin(); it != techs.end(); ++it)
    {
      const TechWrapper& techWrapper = *it;
      QListWidgetItem* item = new QListWidgetItem();
      techWrapper.displayIn(item);
      ui->lstTechs->addItem(item);
    }

    const std::map<BWAPI::UnitType, size_t>& enemyArmyComposition = perceptualStateEvent->getEnemyArmyComposition();
    ui->lstEnemyArmyComposition->clear();
    for (std::map<BWAPI::UnitType, size_t>::const_iterator it = enemyArmyComposition.begin(); it != enemyArmyComposition.end(); ++it)
    {
      std::stringstream ss;
      ss << it->second << " " << it->first.getName();
      ui->lstEnemyArmyComposition->addItem(QString::fromStdString(ss.str()));
    }
  }
  else if (ev->type() == UpdateLayer3EventType)
  {
    ev->accept();
    UpdateLayer3Event* layer3Event = static_cast<UpdateLayer3Event*>(ev);

    ui->lstTargetArmyComposition->clear();
    const std::vector<QString>& targetArmy = layer3Event->getTargetArmyComposition();
    for (std::vector<QString>::const_iterator it = targetArmy.begin(); it != targetArmy.end(); ++it)
    {
      ui->lstTargetArmyComposition->addItem(*it);
    }

    ui->lstTargetTechs->clear();
    const std::vector<QString>& targetTechs = layer3Event->getTargetTechs();
    for (std::vector<QString>::const_iterator it = targetTechs.begin(); it != targetTechs.end(); ++it)
    {
      ui->lstTargetTechs->addItem(*it);
    }

    ui->lstCurrentTechActions->clear();
    const std::vector<QString>& currentTechActions = layer3Event->getCurrentTechActions();
    for (std::vector<QString>::const_iterator it = currentTechActions.begin(); it != currentTechActions.end(); ++it)
    {
      ui->lstCurrentTechActions->addItem(*it);
    }
  }
  QMainWindow::event(ev);

  return ev->isAccepted();
}