#pragma once
#include "Utils.h"
#include <functional>

//------------------------------------------------------------------------------
// Adaptor functor that dereferences a pointer
//------------------------------------------------------------------------------
template<class T>
struct dereference : std::unary_function<T*,T&>
{
  result_type operator()(argument_type& arg) const { return *arg; }
};

//------------------------------------------------------------------------------
// Predicate that checks whether a value is between two others or not
//------------------------------------------------------------------------------
template<class T, bool Inclusive=false>
struct is_between : std::unary_function<T,bool>
{
  typename static_if_else<
    Inclusive,
    std::less_equal<T>,
    std::less<T>
  >::type comparator;

  is_between(const T& v1, const T& v2) : a(v1), b(v2) { }

  const bool operator()(const T& v) const
  {
    return (comparator(a, v) && comparator(v, b))
        || (comparator(b, v) && comparator(v, a));
  }

protected:
  const T a, b;
};

template <class Pair>
struct extract_pair_first : public std::unary_function<Pair, typename Pair::first_type&>
{
  typename Pair::first_type& operator()(Pair& p) const
  {
    return p.first;
  }
};
template <class Pair>
struct extract_pair_first_const : public std::unary_function<Pair, typename const Pair::first_type&>
{
  extract_pair_first_const() {}
  extract_pair_first_const(const extract_pair_first<Pair>&) {}

  typename const Pair::first_type& operator()(const Pair& p) const
  {
    return p.first;
  }
};

template <class Pair>
struct extract_pair_second : public std::unary_function<Pair, typename Pair::second_type&>
{
  typename Pair::second_type& operator()(Pair& p) const
  {
    return p.second;
  }
};
template <class Pair>
struct extract_pair_second_const : public std::unary_function<Pair, typename const Pair::second_type&>
{
  extract_pair_second_const() {}
  extract_pair_second_const(const extract_pair_second<Pair>&) {}

  typename const Pair::second_type& operator()(const Pair& p) const
  {
    return p.second;
  }
};

template <class Functor1, class Functor2>
struct composite_functor : public std::unary_function<typename Functor2::argument_type, typename Functor1::result_type>
{
  composite_functor(Functor1 f1 = Functor1(), Functor2 f2 = Functor2())
    : m_f1(f1)
    , m_f2(f2)
  {}
  template <class OtherFunctor1, class OtherFunctor2>
  composite_functor(const composite_functor<OtherFunctor1, OtherFunctor2>& other)
    : m_f1(other.m_f1)
    , m_f2(other.m_f2)
  {}
  typename Functor1::result_type operator()(typename Functor2::argument_type arg) const
  {
    return m_f1(m_f2(arg));
  }

  Functor1 m_f1;
  Functor2 m_f2;
};