#include "ReferenceCounted.h"

#if defined(_DEBUG) && defined(_DEBUG_REFERENCE_COUNTED)

#include <boost/iterator/transform_iterator.hpp>
#include <sstream>
#include <set>
#include <algorithm>
#include <cassert>

const ReferenceCounted* ReferenceCounted::s_head = NULL;

struct make_pair : public std::unary_function<const void*, std::pair<const void*, const type_info*>>
{
  std::pair<const void*, const type_info*> operator()(const void* p) const
  {
    return std::make_pair(p, &typeid(p));
  }
};

struct Comparison
{
  bool operator()(const std::pair<const void*, const type_info*>& p1, const std::pair<const void*, const type_info*>& p2) const
  {
    return p1.first < p2.first;
  }
};

size_t ReferenceCounted::signalMemoryLeaks()
{
  std::cout << std::endl << "Checking for memory leaks..." << std::endl;
  size_t refs = 0;
  std::set<const void*> pendingObjects;
  std::set<std::pair<const void*, const type_info*>, Comparison> acquirers;
  const ReferenceCounted *object = ReferenceCounted::s_head;
  while (object != NULL)
  {
    pendingObjects.insert(object);
    ++refs;
    std::string description;
    try {
      description = object->toString();
    }
    catch (...) {
      std::stringstream ss;
      ss << typeid(*object).name() << " " << std::hex << (void*)object;
      description = ss.str();
    }

    std::cout << "'" << description << "' " << std::hex << static_cast<const void*>(object) << " has " << std::dec << object->m_nbRef << " pending reference(s)." << std::endl;
    for (std::map<const void*, boost::tuple<const type_info*, size_t, bool>>::const_iterator it = object->m_acquirers.begin();
         it != object->m_acquirers.end();
         ++it)
    {
      acquirers.insert(std::make_pair(it->first, it->second.get<0>()));

      std::cout << "    '" << it->second.get<0>()->name() << " " << std::hex << it->first << "' " << it->second.get<1>() << " ref(s)";
      if (!it->second.get<2>())
        std::cout << " (NOT REFERENCE COUNTED)";
      std::cout << std::endl;
    }
    object = object->m_previous;
  }
  assert(pendingObjects.size() == refs);

  // If there are cycles, we're fucked

  std::set<std::pair<const void*, const type_info*>, Comparison> rootAcquirers;
  std::set_difference(acquirers.begin(), 
                      acquirers.end(), 
                      boost::transform_iterator<make_pair, std::set<const void*>::const_iterator>(pendingObjects.begin()),
                      boost::transform_iterator<make_pair, std::set<const void*>::const_iterator>(pendingObjects.end()),
                      std::inserter(rootAcquirers, rootAcquirers.begin()), 
                      Comparison());
  if (rootAcquirers.size() > 0)
  {
    std::cout << std::endl << "Root acquirers..." << std::endl;
    for (std::set<std::pair<const void*, const type_info*>, Comparison>::const_iterator it = rootAcquirers.begin();
         it != rootAcquirers.end();
         ++it)
    {
      assert(pendingObjects.find(it->first) == pendingObjects.end());
      std::cout << "'" << it->second->name() << " " << std::hex << it->first << "'" << std::endl;
    }
  }

  return refs;
}


#endif